/***************************************************************************
****************************************************************************

This is a part of the xModule Source code.

Copyright (C) 2003-2008 The LS-Universe Team

    Homepage: http://www.ls-universe.info

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************
****************************************************************************/

#include "stdafx.h"
#include "Popup.h"

#include "VerInfo.h"

#include "GhostPopup.h"
#include "GhostButton.h"

#pragma comment(lib, "xPaintClass-1.0.lib")

//xStatsClass
ProcessStatsRequestFunc m_pProcessStatsRequest = NULL;

string str_LoadedxStatsClass = "";

//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// ModulesCallback
//
BOOL __stdcall ModulesCallback(LPCSTR pszPath, DWORD dwFlags, LPARAM lParam)
{
	CStr strCheck = pszPath;
	strCheck.Lower();
	int iPos = strCheck.Find("xstatsclass-");
	if ( iPos != -1 )
	{
		str_LoadedxStatsClass = strCheck.Right(strCheck.Length()-iPos);
		return FALSE;
	}

    return TRUE;
}

CompletePaintClasses *pDefaultPaintClasses = NULL;

GhostPopupList *plistDefaultGhostPopup = NULL;
GhostButtonList *plistDefaultGhostButton = NULL;

//DropShadow under XP
#ifndef CS_DROPSHADOW
	#define CS_DROPSHADOW 0x00020000
#endif

HINSTANCE hInstance;
HWND messageHandler;

UINT (WINAPI *pPrivateExtractIcons)(LPCTSTR,INT,INT,INT,HICON*,UINT*,UINT,UINT) = NULL;

PopupList listPopups;
PopupList listVisiblePopups;

StringList listLoadedPopupCFG;

int lsMessages[] = {
	LM_GETREVID,
	LM_REFRESH,
	0
};

bool bInternalPopupToggle = false;

int iBitsPerPixel = QueryPixelDepth();

string dragndropSource = "-1";
string dragndropDestination = "";

int iRealMonitors = GetSystemMetrics(SM_CMONITORS);

int iScreenX = GetSystemMetrics(SM_XVIRTUALSCREEN);
int iScreenY = GetSystemMetrics(SM_YVIRTUALSCREEN);

int iScreenWidth = GetSystemMetrics(SM_CXVIRTUALSCREEN);
int iScreenHeight = GetSystemMetrics(SM_CYVIRTUALSCREEN);

void StartUp();
void PreParseStepRC(StringList &theList);
string PrepareAnchor(POINT &pt, const char *pszArguments);
void PreCalcPosition(int &x, int &y, int width, int height, LPCSTR anchor);

void BangRescanPopup(HWND hwndCaller, const char *pszName, const char *pszArguments);
void BangPopup(HWND hwndCaller, const char *pszName, const char *pszArguments);
void BangDynamicPopup(HWND hwndCaller, const char *pszName, const char *pszArguments);
void BangLoadPopupCFG(HWND hwndCaller, const char *pszName, const char *pszArguments);

void BangTasks(HWND hwndCaller, const char *pszName, const char *pszArguments);
void BangControlPanel(HWND hwndCaller, const char *pszName, const char *pszArguments);
void BangMyComputer(HWND hwndCaller, const char *pszName, const char *pszArguments);
void BangNetwork(HWND hwndCaller, const char *pszName, const char *pszArguments);
void BangPrinters(HWND hwndCaller, const char *pszName, const char *pszArguments);
void BangRecentDocuments(HWND hwndCaller, const char *pszName, const char *pszArguments);
void BangRecycleBin(HWND hwndCaller, const char *pszName, const char *pszArguments);

void BangReposition(HWND hwndCaller, const char *pszName, const char *pszArguments);
void BangPin(HWND hwndCaller, const char *pszName, const char *pszArguments);
void BangUnPin(HWND hwndCaller, const char *pszName, const char *pszArguments);
void BangTogglePin(HWND hwndCaller, const char *pszName, const char *pszArguments);
void BangToggleShade(HWND hwndCaller, const char *pszName, const char *pszArguments);
void BangToggle(HWND hwndCaller, const char *pszName, const char *pszArguments);

LRESULT WINAPI MessageHandlerProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	switch(message)
	{
		case LM_GETREVID:
		{
			UINT uLength;
			StringCchPrintf((char*)lParam, 64, "%s.dll: %s", V_NAME, V_VERSION);
			
			if (SUCCEEDED(StringCchLength((char*)lParam, 64, &uLength)))
				return uLength;

			lParam = NULL;
			return 0;
		}

		case LM_REFRESH:
		{	
			LSExecute(NULL, "!popuprescan", SW_SHOWNORMAL);

			return 0;
		}

		case WM_TIMER:
		{
			KillTimer(messageHandler, NULL);
			StartUp();
		}
	}

	return DefWindowProc(hWnd, message, wParam, lParam);
}

// Why are we advoiding the crt startup code? The start up code is not large
// enough to cause bloat (except possibly a long system call, and I suspect that
// would need done eventually anyway) (default _DllMainCRTStartup is in
// (MSVCDIR\crt\src\dllcrt0.c) The elimination of the call to _CRT_INIT is 
// also causing warnings with MSVC7
BOOL __stdcall _DllMainCRTStartup(HINSTANCE hInst, DWORD fdwReason, LPVOID lpvRes)
{
	// We don't need thread notifications for what we're doing.  Thus, get
	// rid of them, thereby eliminating some of the overhead of this DLL
	DisableThreadLibraryCalls(hInst);
	return TRUE;
}

int initModuleEx(HWND hParent, HINSTANCE hInstance, const char *lsPath)
{
	WNDCLASSEX wc;

	wc.cbSize = sizeof(WNDCLASSEX);
	wc.style = CS_GLOBALCLASS | CS_DBLCLKS;
	wc.lpfnWndProc = Popup::windowProcedure;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = sizeof(Popup *);
	wc.hInstance = hInstance;
	wc.hbrBackground = 0;
	wc.hCursor = LoadCursor(0, IDC_ARROW);
	wc.hIcon = 0;
	wc.lpszMenuName = 0;
	wc.lpszClassName = "xPopup";
	wc.hIconSm = 0;

	RegisterClassEx(&wc);

	if(IsOS(OS_XP))
	{
		wc.cbSize = sizeof(WNDCLASSEX);
		wc.style = CS_GLOBALCLASS | CS_DBLCLKS | CS_DROPSHADOW;
		wc.lpfnWndProc = Popup::windowProcedure;
		wc.cbClsExtra = 0;
		wc.cbWndExtra = sizeof(Popup *);
		wc.hInstance = hInstance;
		wc.hbrBackground = 0;
		wc.hCursor = LoadCursor(0, IDC_ARROW);
		wc.hIcon = 0;
		wc.lpszMenuName = 0;
		wc.lpszClassName = "xPopupXP";
		wc.hIconSm = 0;

		RegisterClassEx(&wc);
	}

	wc.cbSize = sizeof(WNDCLASSEX);
	wc.style = CS_GLOBALCLASS;
	wc.lpfnWndProc = MessageHandlerProc;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = 0;
	wc.hInstance = hInstance;
	wc.hbrBackground = 0;
	wc.hCursor = 0;
	wc.hIcon = 0;
	wc.lpszMenuName = 0;
	wc.lpszClassName = "xPopupMessageHandler";
	wc.hIconSm = 0;

	RegisterClassEx(&wc);

	wc.cbSize = sizeof(WNDCLASSEX);
	wc.style = CS_GLOBALCLASS;
	wc.lpfnWndProc = DefWindowProc;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = 0;
	wc.hInstance = hInstance;
	wc.hbrBackground = 0;
	wc.hCursor = LoadCursor(0, IDC_ARROW);
	wc.hIcon = 0;
	wc.lpszMenuName = 0;
	wc.lpszClassName = "xPopupOwner";
	wc.hIconSm = 0;

	RegisterClassEx(&wc);

	messageHandler = CreateWindowEx(WS_EX_TOOLWINDOW,
		"xPopupMessageHandler",
		0,
		0,
		0, 0, 0, 0, 
		HWND_MESSAGE,
		0,
		hInstance,
		0);

	if (!messageHandler)
		return 1;
	
	SendMessage(GetLitestepWnd(),
		LM_REGISTERMESSAGE,
		(WPARAM) messageHandler,
		(LPARAM) lsMessages);

	::hInstance = hInstance;

	HRESULT hr = EnumLSData(2 /*ELD_MODULES*/, (FARPROC)ModulesCallback, (LPARAM) NULL);

	if ( !m_pProcessStatsRequest )
	{
		m_pProcessStatsRequest = (ProcessStatsRequestFunc) GetProcAddress(GetModuleHandle(str_LoadedxStatsClass.c_str()), "processStatsRequest");
	} 
	
	if (!pPrivateExtractIcons)
	{
		pPrivateExtractIcons = (UINT (WINAPI *)(LPCTSTR,INT,INT,INT,HICON*,UINT*,UINT,UINT)) GetProcAddress(GetModuleHandle(_T("USER32.DLL")), "PrivateExtractIconsA");
	}

	listPopups.clear();
	listVisiblePopups.clear();
	listLoadedPopupCFG.clear();

	AddBangCommandEx("!Popup", BangPopup);
	AddBangCommandEx("!PopupDynamicFolder", BangDynamicPopup);
	AddBangCommandEx("!PopupLoadCFG", BangLoadPopupCFG);
	AddBangCommandEx("!PopupRescan", BangRescanPopup);

	AddBangCommandEx("!PopupTasks", BangTasks);
	AddBangCommandEx("!PopupControlPanel", BangControlPanel);
    AddBangCommandEx("!PopupMyComputer", BangMyComputer);
    AddBangCommandEx("!PopupNetwork", BangNetwork);
    AddBangCommandEx("!PopupPrinters", BangPrinters);
    AddBangCommandEx("!PopupRecentDocuments", BangRecentDocuments);
    AddBangCommandEx("!PopupRecycleBin", BangRecycleBin);

	bool bUseAlphaMap = false;
	if ( IsOS(OS_2KXP) )
		bUseAlphaMap = Getx2RCBool("xPopup", "AlphaMap");

	bool bUseFolderDefault = Getx2RCBool("xPopup", "ActiveFolderUseFolderDefault");

	//Create "xPopup" Default Painting Classes
	pDefaultPaintClasses = new CompletePaintClasses();
	
	pDefaultPaintClasses->pPopupTexture->configure("xPopup", "", bUseAlphaMap);

	//------------------
	//Texture
	//------------------
	pDefaultPaintClasses->pEntryTexture->configure("xPopup", "Entry", bUseAlphaMap);
	pDefaultPaintClasses->pActiveEntryTexture->configure("xPopup", "ActiveEntry", bUseAlphaMap, pDefaultPaintClasses->pEntryTexture);
	pDefaultPaintClasses->pFolderTexture->configure("xPopup", "Folder", bUseAlphaMap, pDefaultPaintClasses->pEntryTexture);
	pDefaultPaintClasses->pActiveFolderTexture->configure("xPopup", "ActiveFolder", bUseAlphaMap, (bUseFolderDefault ? pDefaultPaintClasses->pFolderTexture : pDefaultPaintClasses->pActiveEntryTexture));
	
	pDefaultPaintClasses->pInfoTexture->configure("xPopup", "Info", bUseAlphaMap);
	pDefaultPaintClasses->pActiveInfoTexture->configure("xPopup", "ActiveInfo", bUseAlphaMap, pDefaultPaintClasses->pInfoTexture);
	
	pDefaultPaintClasses->pTitleTexture->configure("xPopup", "Title", bUseAlphaMap);
	
	pDefaultPaintClasses->pBottomTexture->configure("xPopup", "Bottom", bUseAlphaMap);

	pDefaultPaintClasses->pSeparatorTexture->configure("xPopup", "Separator", bUseAlphaMap);

	pDefaultPaintClasses->pArrowTexture->configure("xPopup", "Arrow", bUseAlphaMap);
	pDefaultPaintClasses->pActiveArrowTexture->configure("xPopup", "ActiveArrow", bUseAlphaMap, pDefaultPaintClasses->pArrowTexture);

	//------------------
	//Icon
	//------------------
	PaintIcon *defaulticon = Create_xIconClass();
	defaulticon->configure("xPopup", "", true, bUseAlphaMap);

	pDefaultPaintClasses->pEntryIcon->configure("xPopup", "Entry", true, bUseAlphaMap, defaulticon);
	pDefaultPaintClasses->pActiveEntryIcon->configure("xPopup", "ActiveEntry", true, bUseAlphaMap, pDefaultPaintClasses->pEntryIcon);
	pDefaultPaintClasses->pFolderIcon->configure("xPopup", "Folder", true, bUseAlphaMap, pDefaultPaintClasses->pEntryIcon);
	pDefaultPaintClasses->pActiveFolderIcon->configure("xPopup", "ActiveFolder", true, bUseAlphaMap, (bUseFolderDefault ? pDefaultPaintClasses->pFolderIcon : pDefaultPaintClasses->pActiveEntryIcon));
	
	Destroy_xIconClass(defaulticon);

	pDefaultPaintClasses->pInfoIcon->configure("xPopup", "Info", true, bUseAlphaMap);
	pDefaultPaintClasses->pActiveInfoIcon->configure("xPopup", "ActiveInfo", true, bUseAlphaMap, pDefaultPaintClasses->pInfoIcon);

	if ( Getx2RCBool("xPopup", "TitleShowIcon") )
	{
		pDefaultPaintClasses->pTitleIcon->configure("xPopup", "Title", true, bUseAlphaMap);
	}
	else
	{
		Destroy_xIconClass(pDefaultPaintClasses->pTitleIcon);
		pDefaultPaintClasses->pTitleIcon = NULL;
	}
	if ( Getx2RCBool("xPopup", "PinnedShowIcon") )
	{
		pDefaultPaintClasses->pPinnedIcon->configure("xPopup", "Pinned", true, bUseAlphaMap);
	}
	else
	{
		Destroy_xIconClass(pDefaultPaintClasses->pPinnedIcon);
		pDefaultPaintClasses->pPinnedIcon = NULL;
	}

	//------------------
	//Text
	//------------------
	PaintText *defaulttext = Create_xTextClass();
	defaulttext->configure("xPopup", "", bUseAlphaMap);

	pDefaultPaintClasses->pEntryText->configure("xPopup", "Entry", bUseAlphaMap, defaulttext);
	pDefaultPaintClasses->pActiveEntryText->configure("xPopup", "ActiveEntry", bUseAlphaMap, pDefaultPaintClasses->pEntryText);
	pDefaultPaintClasses->pFolderText->configure("xPopup", "Folder", bUseAlphaMap, pDefaultPaintClasses->pEntryText);
	pDefaultPaintClasses->pActiveFolderText->configure("xPopup", "ActiveFolder", bUseAlphaMap, (bUseFolderDefault ? pDefaultPaintClasses->pFolderText : pDefaultPaintClasses->pActiveEntryText));
	
	Destroy_xTextClass(defaulttext);

	pDefaultPaintClasses->pInfoText->configure("xPopup", "Info", bUseAlphaMap);
	pDefaultPaintClasses->pActiveInfoText->configure("xPopup", "ActiveInfo", bUseAlphaMap, pDefaultPaintClasses->pInfoText);
	
	pDefaultPaintClasses->pTitleText->configure("xPopup", "Title", bUseAlphaMap);

	if ( Getx2RCBool("xPopup", "UseTextArrow") )
	{
		pDefaultPaintClasses->pArrowText->configure("xPopup", "Arrow", bUseAlphaMap);
		pDefaultPaintClasses->pActiveArrowText->configure("xPopup", "ActiveArrow", bUseAlphaMap, pDefaultPaintClasses->pArrowText);
	}
	else
	{
		Destroy_xTextClass(pDefaultPaintClasses->pArrowText);
		pDefaultPaintClasses->pArrowText = NULL;
		Destroy_xTextClass(pDefaultPaintClasses->pActiveArrowText);
		pDefaultPaintClasses->pActiveArrowText = NULL;
	}

	//Create Default OverlayPopup Painting
	//Initialize Basic Helper Strings
	int numTokens = 0;
	string one = "";
	string two = "";
	string three = "";
	string four = "";
	string five = "";
	string six = "";
	string seven = "";

	plistDefaultGhostPopup = new GhostPopupList();
	plistDefaultGhostButton = new GhostButtonList();

	StringList OverlayPaintingList;
	OverlayPaintingList.clear();
	
	OverlayPaintingList = GetxRCLineList( ("*xPopupOverlayPopup") );
	
	if ( !OverlayPaintingList.empty() )
	{
		for(StringListIterator sit = OverlayPaintingList.begin(); sit != OverlayPaintingList.end(); sit++)
		{
			numTokens = SplitString((*sit), one, two, three, four, five, six, seven);
			
			if ( numTokens >= 5 )
			{
				GhostPopup* newpopup = new GhostPopup();
			
				newpopup->strOrgX = one;
				newpopup->strOrgY = two;

				newpopup->strOrgWidth = three;
				newpopup->strOrgHeight = four;
	
				if ( numTokens >= 6 && IsOS( OS_2KXP ) && (strchr(six.c_str(), 'a') != 0 || strchr(six.c_str(), 'A') != 0) )
					newpopup->pPopupTexture->configure(five.c_str(), "", true);
				else
					newpopup->pPopupTexture->configure(five.c_str(), "", false);
			
				plistDefaultGhostPopup->insert(plistDefaultGhostPopup->end(), newpopup);
			}
		}
	}

	OverlayPaintingList.clear();
	//Create Default OverlayButton Painting
	OverlayPaintingList = GetxRCLineList( ("*xPopupOverlayEntry") );
	
	if ( !OverlayPaintingList.empty() )
	{
		for(StringListIterator sit = OverlayPaintingList.begin(); sit != OverlayPaintingList.end(); sit++)
		{
			//Use Split Helper Strings
			numTokens = SplitString((*sit), one, two, three, four, five, six, seven);
			
			if ( numTokens >= 5 )
			{
				GhostButton* newbutton = new GhostButton();
			
				newbutton->strOrgX = one;
				newbutton->strOrgY = two;

				newbutton->strOrgWidth = three;
				newbutton->strOrgHeight = four;

				if ( numTokens >= 6 && IsOS( OS_2KXP ) && (strchr(six.c_str(), 'a') != 0 || strchr(six.c_str(), 'A') != 0) )
				{
					newbutton->pEntryTexture->configure(five.c_str(), "Entry", true);
					newbutton->pActiveEntryTexture->configure(five.c_str(), "ActiveEntry", true, newbutton->pEntryTexture);
					newbutton->pFolderTexture->configure(five.c_str(), "Folder", true, newbutton->pEntryTexture);
					newbutton->pActiveFolderTexture->configure(five.c_str(), "ActiveFolder", true, (bUseFolderDefault ? newbutton->pFolderTexture : newbutton->pActiveEntryTexture));
					
					newbutton->pInfoTexture->configure(five.c_str(), "Info", true);
					newbutton->pActiveInfoTexture->configure(five.c_str(), "ActiveInfo", true, newbutton->pInfoTexture);

					newbutton->pTitleTexture->configure(five.c_str(), "Title", true);
				
					newbutton->pBottomTexture->configure(five.c_str(), "Bottom", true);
				
					newbutton->pSeparatorTexture->configure(five.c_str(), "Separator", true);
				}
				else
				{
					newbutton->pEntryTexture->configure(five.c_str(), "Entry", false);
					newbutton->pActiveEntryTexture->configure(five.c_str(), "ActiveEntry", false, newbutton->pEntryTexture);
					newbutton->pFolderTexture->configure(five.c_str(), "Folder", false, newbutton->pEntryTexture);
					newbutton->pActiveFolderTexture->configure(five.c_str(), "ActiveFolder", false, (bUseFolderDefault ? newbutton->pFolderTexture : newbutton->pActiveEntryTexture));
					
					newbutton->pInfoTexture->configure(five.c_str(), "Info", false);
					newbutton->pActiveInfoTexture->configure(five.c_str(), "ActiveInfo", false, newbutton->pInfoTexture);

					newbutton->pTitleTexture->configure(five.c_str(), "Title", false);
				
					newbutton->pBottomTexture->configure(five.c_str(), "Bottom", false);
				
					newbutton->pSeparatorTexture->configure(five.c_str(), "Separator", false);
				}

				plistDefaultGhostButton->insert(plistDefaultGhostButton->end(), newbutton);
			}
		}
	}

	PrefixedLSSetVariable("xPopup", "ExportedHoverCaption", "");
	PrefixedLSSetVariable("xPopup", "ExportedHoverAction", "");

	//Parse Step
	SetTimer(messageHandler, NULL, 100, 0);

	return 0;
}

void StartUp()
{
	StringList listPinnedPopups;
	listPinnedPopups.clear();

	//Use listPinnedPopups as empty dummy list
	PreParseStepRC(listPinnedPopups);

	GetxRCList( "*xPopupStartPinned", listPinnedPopups );

	char szFirst[MAX_LINE_LENGTH] = { 0 };
	char szSecond[MAX_LINE_LENGTH] = { 0 };
	char szThird[MAX_LINE_LENGTH] = { 0 };
	char szFourth[MAX_LINE_LENGTH] = { 0 };

	char *buffer[] = { szFirst, szSecond , szThird, szFourth };

	int x = 0;
	int y = 0;
	for (StringListIterator sit = listPinnedPopups.begin(); sit != listPinnedPopups.end(); ++sit)
	{
		LCTokenize((*sit).c_str(), buffer, 4, NULL);

		x = xParseCoord(szSecond, 0, iScreenWidth, false);
		y = xParseCoord(szThird, 0, iScreenHeight, true);

		for(PopupListIterator plit = listPopups.begin(); plit != listPopups.end(); plit++)
		{
			if (_stricmp((*plit)->BangToCall.c_str(), szFirst) == NULL)
			{
				//Calculate Position due to arguments
				PreCalcPosition(x, y, (*plit)->getWidth(), (*plit)->getHeight(), szFourth);

				(*plit)->reposition(3, x, y, (*plit)->getWidth(), (*plit)->getHeight());
				(*plit)->repaint(true);
				(*plit)->bPinnedPopupMode = true;

				if ( (*plit)->bPinnedPopupNotOnTop )
					(*plit)->setAlwaysOnTop(false);
				else
					(*plit)->setAlwaysOnTop(true);

				(*plit)->show();
				break;
			}
		}		
	}

	listPinnedPopups.clear();
}

void PreParseStepRC(StringList &theList)
{
	StringList settingsList;
	settingsList.clear();

	bool bCustomLoadCFG = false;

	if ( !theList.empty() )
	{
		settingsList = theList;
		bCustomLoadCFG = true;
	}
	else
	{
		GetxRCList( "*Popup", settingsList );
	}

	char szFirst[MAX_LINE_LENGTH] = { 0 };
	char szSecond[MAX_LINE_LENGTH] = { 0 };
	char szThird[MAX_LINE_LENGTH] = { 0 };
	char szFourth[MAX_LINE_LENGTH] = { 0 };

	char *buffer[] = { szFirst, szSecond , szThird, szFourth };

	//Initialize
	Popup* pAddNewPopup = NULL;
	bool bInsertToPopup = false;
	StringList listSentToPopup;
	listSentToPopup.clear();
	int iTokens = 0;
	string strAddNewBang = "";

	for (StringListIterator sit = settingsList.begin(); sit != settingsList.end(); ++sit)
	{
		iTokens = LCTokenize((*sit).c_str(), buffer, 3, szFourth);

		if(iTokens >= 2 && (_stricmp(szFirst, "!new") == NULL || _stricmp(szSecond, "!new") == NULL))
		{	
			//Search for Popup Start with !New and Create New Popup
			Popup *popup = NULL;
			if (_stricmp(szFirst, "!New") == NULL)
			{
				if ( _stricmp(szThird, "") == NULL )
				{
					popup = new Popup("xpopup", "", pDefaultPaintClasses, plistDefaultGhostPopup, plistDefaultGhostButton);
				}
				else
				{
					popup = new Popup(szThird, "");
				}

				popup->BangToCall = szSecond;

				//Remove if already defined
				if ( bCustomLoadCFG )
				{
					for(PopupListIterator plit = listPopups.begin(); plit != listPopups.end(); plit++)
					{
						if (_stricmp((*plit)->BangToCall.c_str(), popup->BangToCall.c_str()) == NULL)
						{
							(*plit)->hide(true/*noFade*/);
							delete *plit;
							listPopups.erase(plit);
							break;
						}
					}
				}
			
				AddBangCommandEx(szSecond, BangPopup);

				//Reposition Bang
				strAddNewBang = szSecond;
				strAddNewBang.append("Reposition");
				AddBangCommandEx(strAddNewBang.c_str(), BangReposition);

				//Pin Bang == Show
				strAddNewBang = szSecond;
				strAddNewBang.append("Pin");
				AddBangCommandEx(strAddNewBang.c_str(), BangPin);

				//UnPin Bang == Hide
				strAddNewBang = szSecond;
				strAddNewBang.append("UnPin");
				AddBangCommandEx(strAddNewBang.c_str(), BangUnPin);

				//Toggle bPinnedPopupMode == Toggle
				strAddNewBang = szSecond;
				strAddNewBang.append("TogglePin");
				AddBangCommandEx(strAddNewBang.c_str(), BangTogglePin);

				//Toggle ShadeMode == Toggle
				strAddNewBang = szSecond;
				strAddNewBang.append("ToggleShade");
				AddBangCommandEx(strAddNewBang.c_str(), BangToggleShade);

				//Toggle Visibility == Toggle
				strAddNewBang = szSecond;
				strAddNewBang.append("Toggle");
				AddBangCommandEx(strAddNewBang.c_str(), BangToggle);
			}
			else if (_stricmp(szSecond, "!new") == NULL)
			{
				if ( _stricmp(szFourth, "") == NULL )
					popup = new Popup("xpopup", szFirst, pDefaultPaintClasses, plistDefaultGhostPopup, plistDefaultGhostButton);
				else
					popup = new Popup(szFourth, szFirst);

				popup->BangToCall = szThird;

				//Remove if already defined
				if ( bCustomLoadCFG )
				{		
					for(PopupListIterator plit = listPopups.begin(); plit != listPopups.end(); plit++)
					{
						if (_stricmp((*plit)->BangToCall.c_str(), popup->BangToCall.c_str()) == NULL)
						{			
							(*plit)->hide(true/*noFade*/);
							delete *plit;
							listPopups.erase(plit);
							break;
						}
					}
				}

				AddBangCommandEx(szThird, BangPopup);
		
				//Reposition Bang
				strAddNewBang = szThird;
				strAddNewBang.append("Reposition");
				AddBangCommandEx(strAddNewBang.c_str(), BangReposition);

				//Pin Bang == Show
				strAddNewBang = szThird;
				strAddNewBang.append("Pin");
				AddBangCommandEx(strAddNewBang.c_str(), BangPin);

				//UnPin Bang == Hide
				strAddNewBang = szThird;
				strAddNewBang.append("UnPin");
				AddBangCommandEx(strAddNewBang.c_str(), BangUnPin);

				//Toggle bPinnedPopupMode == Toggle
				strAddNewBang = szThird;
				strAddNewBang.append("TogglePin");
				AddBangCommandEx(strAddNewBang.c_str(), BangTogglePin);

				//Toggle ShadeMode == Toggle
				strAddNewBang = szThird;
				strAddNewBang.append("ToggleShade");
				AddBangCommandEx(strAddNewBang.c_str(), BangToggleShade);

				//Toggle Visibility == Toggle
				strAddNewBang = szThird;
				strAddNewBang.append("Toggle");
				AddBangCommandEx(strAddNewBang.c_str(), BangToggle);
			}
		
			if (popup)
			{
				listPopups.insert(listPopups.end(), popup);
				
				pAddNewPopup = popup;

				//Allow following settings to be tracked
				bInsertToPopup = true;
			}
		}
		else if (iTokens >= 1 && (_stricmp(szFirst, "~new") == NULL || _stricmp(szSecond, "~new") == NULL))
		{
			//Close Popup Definition, give Settings and Load
			if (pAddNewPopup)
			{
				pAddNewPopup->load(hInstance);
				pAddNewPopup->popupcontentList = listSentToPopup;
				pAddNewPopup->ParseStepRC();
			}
			pAddNewPopup = NULL;
			bInsertToPopup = false;
			listSentToPopup.clear();
		}
		else if (bInsertToPopup)
		{
			//TODO
			//Add to specific Popup Settings list
			if (strstr(szFirst, ".icon=") == NULL && strstr(szFirst, ".Icon=") == NULL && strstr(szFirst, ".ICON=") == NULL)
			{
				listSentToPopup.insert(listSentToPopup.end(), (".icon=\".noexist\" " + (*sit)));
			}
			else
			{
				listSentToPopup.insert(listSentToPopup.end(), (*sit));
			}
		}
	}

	listSentToPopup.clear();
}

string PrepareAnchor(POINT &pt, const char *pszArguments)
{
	char x[32] = {0};
	char y[32] = {0};
	char anchor1[64] = {0};
	char anchor2[64] = {0};
	string combinedanchor = "topleft";
	char *posbuffers[] = {x, y, anchor1, anchor2};
	
	int numTokens = LCTokenize( pszArguments, posbuffers, 4, NULL);

	if ( numTokens > 0 )
	{
		if ( _stricmp(x, "topleft") == NULL || _stricmp(x, "topcenter") == NULL || _stricmp(x, "topright") == NULL
				|| _stricmp(x, "centerleft") == NULL || _stricmp(x, "centercenter") == NULL || _stricmp(x, "centerright") == NULL
				|| _stricmp(x, "bottomleft") == NULL || _stricmp(x, "bottomcenter") == NULL || _stricmp(x, "bottomright") == NULL )
		{
			combinedanchor = x;
			return combinedanchor;
		}
		else if ( _stricmp(x, "bottom") == NULL || _stricmp(x, "top") == NULL || _stricmp(x, "center") == NULL )
		{
			combinedanchor = x;
			if ( _stricmp(y, "left") == NULL || _stricmp(y, "right") == NULL || _stricmp(y, "center") == NULL )
				combinedanchor.append(y);
			else
				combinedanchor.append("left");

			return combinedanchor;
		}
		else if ( _stricmp(x, "left") == NULL || _stricmp(x, "right") == NULL )
		{
			if ( _stricmp(y, "bottom") == NULL || _stricmp(y, "top") == NULL || _stricmp(y, "center") == NULL )
				combinedanchor = y;
			else
				combinedanchor = "top";
			
			combinedanchor.append(x);

			return combinedanchor;
		}
		else
		{
			pt.x = xParseCoord(x, pt.x, iScreenWidth, false);
			pt.y = xParseCoord(y, pt.y, iScreenHeight, true);

			if ( _stricmp(anchor1, "topleft") == NULL || _stricmp(anchor1, "topcenter") == NULL || _stricmp(anchor1, "topright") == NULL
				|| _stricmp(anchor1, "centerleft") == NULL || _stricmp(anchor1, "centercenter") == NULL || _stricmp(anchor1, "centerright") == NULL
				|| _stricmp(anchor1, "bottomleft") == NULL || _stricmp(anchor1, "bottomcenter") == NULL || _stricmp(anchor1, "bottomright") == NULL )
			{
				combinedanchor = anchor1;
				return combinedanchor;
			}
			else if ( _stricmp(anchor1, "bottom") == NULL || _stricmp(anchor1, "top") == NULL || _stricmp(anchor1, "center") == NULL )
			{
				combinedanchor = anchor1;
				if ( _stricmp(anchor2, "left") == NULL || _stricmp(anchor2, "right") == NULL || _stricmp(anchor2, "center") == NULL )
					combinedanchor.append(y);
				else
					combinedanchor.append("left");

				return combinedanchor;
			}
			else if ( _stricmp(anchor1, "left") == NULL || _stricmp(anchor1, "right") == NULL )
			{
				if ( _stricmp(anchor2, "bottom") == NULL || _stricmp(anchor2, "top") == NULL || _stricmp(anchor2, "center") == NULL )
					combinedanchor = anchor2;
				else
					combinedanchor = "top";
				
				combinedanchor.append(anchor1);

				return combinedanchor;
			}
		}
	}

	return combinedanchor;
}

void PreCalcPosition(int &x, int &y, int width, int height, LPCSTR anchor)
{
	int screenX = 0;
	int screenY = 0;
	int screenWidth = iScreenWidth;
	int screenHeight = iScreenHeight;

	if ( iRealMonitors > 1 )
	{
		POINT pt = {x, y};
		HMONITOR hMonitor = MonitorFromPoint(pt, MONITOR_DEFAULTTONEAREST);
				
		MONITORINFO mi = { sizeof(MONITORINFO) };
		GetMonitorInfo(hMonitor, &mi);

		screenX = mi.rcMonitor.left;
		screenY = mi.rcMonitor.top;

		screenWidth = mi.rcMonitor.right-mi.rcMonitor.left;
		screenHeight = mi.rcMonitor.bottom-mi.rcMonitor.top;
	}

	// TopLeft is Default
	if (_stricmp(anchor, "topleft") == NULL)
	{
		if (x+width > screenWidth)
			x = screenWidth-width;

		if (y+height > screenHeight)
			y = screenHeight-height;
	}
	else if (_stricmp(anchor, "topcenter") == NULL)
	{
		if (x-width/2 < screenX)
			x = screenX;
		else if (x+width/2 > screenWidth)
			x = screenWidth-width;
		else
			x = x-width/2;

		if (y+height > screenHeight)
			y = screenHeight-height;
	}
	else if (_stricmp(anchor, "topright") == NULL)
	{
		if (x-width < screenX)
			x = screenX;
		else
			x = x-width;

		if (y+height > screenHeight)
			y = screenHeight-height;
	}
	else if (_stricmp(anchor, "centerleft") == NULL)
	{
		if (x+width > screenWidth)
			x = screenWidth-width;

		if (y-height/2 < screenY)
			y = screenY;
		else if (y+height/2 > screenHeight)
			y = screenHeight-height;
		else 
			y = y-height/2;
	}
	else if (_stricmp(anchor, "centercenter") == NULL)
	{
		if (x-width/2 < screenX)
			x = screenX;
		else if (x+width/2 > screenWidth)
			x = screenWidth-width;
		else
			x = x-width/2;

		if (y-height/2 < screenY)
			y = screenY;
		else if (y+height/2 > screenHeight)
			y = screenHeight-height;
		else 
			y = y-height/2;
	}
	else if (_stricmp(anchor, "centerright") == NULL)
	{
		if (x-width < screenX)
			x = screenX;
		else
			x = x-width;

		if (y-height/2 < screenY)
			y = screenY;
		else if (y+height/2 > screenHeight)
			y = screenHeight-height;
		else 
			y = y-height/2;
	}
	else if(_stricmp(anchor, "bottomleft") == NULL)
	{
		if (x+width > screenWidth)
			x = screenWidth-width;

		if (y-height < screenY)
			y = screenY;
		else
			y = y-height;
	}
	else if (_stricmp(anchor, "bottomcenter") == NULL)
	{
		if (x-width/2 < screenX)
			x = screenX;
		else if (x+width/2 > screenWidth)
			x = screenWidth-width;
		else
			x = x-width/2;

		if (y-height < screenY)
			y = screenY;
		else
			y = y-height;
	}
	else if(_stricmp(anchor, "bottomright") == NULL)
	{
		if (x-width < screenX)
			x = screenX;
		else
			x = x-width;

		if (y-height < screenY)
			y = screenY;
		else
			y = y-height;
	}
	// TopLeft is Default
	else
	{
		if (x+width > screenWidth)
			x = screenWidth-width;

		if (y+height > screenHeight)
			y = screenHeight-height;
	}
}

void quitModule(HINSTANCE hInstance)
{
	listLoadedPopupCFG.clear();

	listVisiblePopups.clear();

	for(PopupListIterator plit = listPopups.begin(); plit != listPopups.end(); plit++)
	{
		(*plit)->hide(true/*noFade*/);
		delete *plit;
		*plit = NULL;
	}

	listPopups.clear();

	RemoveBangCommand("!Popup");
	RemoveBangCommand("!PopupDynamicFolder");
	RemoveBangCommand("!PopupLoadCFG");
	RemoveBangCommand("!PopupRescan");

	RemoveBangCommand("!PopupTasks");
	RemoveBangCommand("!PopupControlPanel");
    RemoveBangCommand("!PopupMyComputer");
    RemoveBangCommand("!PopupNetwork");
    RemoveBangCommand("!PopupPrinters");
    RemoveBangCommand("!PopupRecentDocuments");
    RemoveBangCommand("!PopupRecycleBin");

	if (pDefaultPaintClasses)
		delete pDefaultPaintClasses;

	pDefaultPaintClasses = NULL;

	for(GhostPopupListIterator gtlit = plistDefaultGhostPopup->begin(); gtlit != plistDefaultGhostPopup->end(); gtlit++)
	{
		delete *gtlit;
		*gtlit = NULL;
	}

	plistDefaultGhostPopup->clear();
	delete plistDefaultGhostPopup;

	plistDefaultGhostPopup = NULL;

	for(GhostButtonListIterator glit = plistDefaultGhostButton->begin(); glit != plistDefaultGhostButton->end(); glit++)
	{
		delete *glit;
		*glit = NULL;
	}

	plistDefaultGhostButton->clear();
	delete plistDefaultGhostButton;

	plistDefaultGhostButton = NULL;

	SendMessage(GetLitestepWnd(),
		LM_UNREGISTERMESSAGE,
		(WPARAM) messageHandler,
		(LPARAM) lsMessages);

	DestroyWindow(messageHandler);

	UnregisterClass("xPopup", hInstance);
	if(IsOS(OS_XP))
		UnregisterClass("xPopupXP", hInstance);
	UnregisterClass("xPopupMessageHandler", hInstance);
	UnregisterClass("xPopupOwner", hInstance);
}

void BangPopup(HWND hwndCaller, const char *pszName, const char *pszArguments)
{
	POINT ptCursorPos;
	GetCursorPos(&ptCursorPos);

	CString anchor = PrepareAnchor(ptCursorPos, pszArguments).c_str();
	anchor.MakeLower();

	bool bOnePopupShown = false;

	for(PopupListIterator plit = listPopups.begin(); plit != listPopups.end(); plit++)
	{
		if ( !bOnePopupShown && (_stricmp( (*plit)->BangToCall.c_str(), pszName) == NULL || _stricmp( pszName, "!popup") == NULL) )
		{
			//Hide, if already visible
			if ( (*plit)->isVisible() && bInternalPopupToggle )
			{
				(*plit)->hide();
				bOnePopupShown = true;
				continue;
			}

			int screenWidth = iScreenWidth;
			int screenHeight = iScreenHeight;

			if ( iRealMonitors > 1 )
			{
				POINT pt = {ptCursorPos.x, ptCursorPos.y};
				HMONITOR hMonitor = MonitorFromPoint(pt, MONITOR_DEFAULTTONEAREST);
						
				MONITORINFO mi = { sizeof(MONITORINFO) };
				GetMonitorInfo(hMonitor, &mi);

				screenWidth = mi.rcMonitor.right-mi.rcMonitor.left;
				screenHeight = mi.rcMonitor.bottom-mi.rcMonitor.top;
			}

			if (ptCursorPos.y+(*plit)->getHeight()/2 > screenHeight)
			{
				anchor.Replace("top", "bottom");
				if (anchor.Find("center") == 0)
				{
					anchor = anchor.Right(anchor.GetLength()-6);
					anchor.Insert(0, "bottom");
				}
			}
			else if (ptCursorPos.y+(*plit)->getHeight() > screenHeight && ptCursorPos.y+(*plit)->getHeight()/2 <= screenHeight)
			{
				anchor.Replace("top", "center");
			}
									
			if (ptCursorPos.x+(*plit)->getWidth()/2 > screenWidth)
			{
				anchor.Replace("left", "right");
				if (anchor.Find("center", 1) == 3 || anchor.Find("center", 1) == 6)
				{
					anchor = anchor.Left(anchor.Find("center", 1));
					anchor += "right";
				}
			}
			else if (ptCursorPos.x+(*plit)->getWidth() > screenWidth && ptCursorPos.x+(*plit)->getWidth()/2 <= screenWidth)
			{
				anchor.Replace("left", "center");
			}

			//Calculate Position due to arguments
			int tmpx = ptCursorPos.x;
			int tmpy = ptCursorPos.y;
			PreCalcPosition(tmpx, tmpy, (*plit)->getWidth(), (*plit)->getHeight(), anchor);
			ptCursorPos.x = tmpx;
			ptCursorPos.y = tmpy;

			KillTimer((*plit)->gethWnd(), 1);
			(*plit)->reposition(3, ptCursorPos.x, ptCursorPos.y, (*plit)->getWidth(), (*plit)->getHeight());
			(*plit)->repaint(true);
			if ((*plit)->bPinnedPopupMode)
			{
				(*plit)->bPinnedPopupMode = false;
				(*plit)->setAlwaysOnTop(true);
			}
	
			(*plit)->show();
			
			bOnePopupShown = true;
		}
		else if ( (*plit)->isVisible() )
		{
			(*plit)->hide();
		}
	}

	bInternalPopupToggle = false;
}

void BangDynamicPopup(HWND hwndCaller, const char *pszName, const char *pszArguments)
{
	char path[MAX_LINE_LENGTH] = {0};
	char rest[MAX_LINE_LENGTH] = {0};
	char *posbuffers[] = {path};
	
	CString anchor = "";
	
	LCTokenize( pszArguments, posbuffers, 1, rest);

	POINT ptCursorPos;
	GetCursorPos(&ptCursorPos);

	if ( _stricmp(pszName, "controlpanel") == NULL || _stricmp(pszName, "mycomputer") == NULL ||
			_stricmp(pszName, "network") == NULL || _stricmp(pszName, "printers") == NULL ||
			_stricmp(pszName, "recentdocuments") == NULL || _stricmp(pszName, "recyclebin") == NULL ||
			_stricmp(pszName, "tasks") == NULL )
	{
		anchor = PrepareAnchor(ptCursorPos, pszArguments).c_str();
	}
	else
	{
		anchor = PrepareAnchor(ptCursorPos, rest).c_str();
	}

	anchor.MakeLower();

	Popup* popup = NULL;
	int alreadyCreated = 0;
	for(PopupListIterator plit = listPopups.begin(); plit != listPopups.end(); plit++)
	{
		if ( alreadyCreated == 0 && _stricmp( (*plit)->BangToCall.c_str(), path ) == NULL )
		{
			alreadyCreated = 1;
			popup = *plit;
		}
		else if ( alreadyCreated == 0 && _stricmp( (*plit)->BangToCall.c_str(), pszName ) == NULL )
		{
			alreadyCreated = 2;
			popup = *plit;
		}
		else if ( (*plit)->isVisible() )
		{
			(*plit)->hide();
		}
	}
	if (alreadyCreated == 0)
	{
		popup = new Popup("xpopup", "", pDefaultPaintClasses, plistDefaultGhostPopup, plistDefaultGhostButton);

		listPopups.insert(listPopups.end(), popup);
		
		if ( _stricmp(pszName, "controlpanel") == NULL || _stricmp(pszName, "mycomputer") == NULL ||
				_stricmp(pszName, "network") == NULL	|| _stricmp(pszName, "printers") == NULL ||
				_stricmp(pszName, "recentdocuments") == NULL || _stricmp(pszName, "recyclebin") == NULL ||
				_stricmp(pszName, "tasks") == NULL )
			popup->BangToCall = pszName;
		else
			popup->BangToCall = path;
			
		popup->load(hInstance);
	}
	
	if (popup != NULL)
	{
		if ( _stricmp(pszName, "controlpanel") == NULL || _stricmp(pszName, "mycomputer") == NULL ||
				_stricmp(pszName, "network") == NULL	|| _stricmp(pszName, "printers") == NULL ||
				_stricmp(pszName, "recentdocuments") == NULL || _stricmp(pszName, "recyclebin") == NULL ||
				_stricmp(pszName, "tasks") == NULL )
		{
			int clsid = 0;

			if ( _stricmp(pszName, "controlpanel") == NULL )
			{
				clsid = CSIDL_CONTROLS;
				popup->captionname = "ControlPanel";
			}
			else if ( _stricmp(pszName, "mycomputer") == NULL )
			{
				clsid = CSIDL_DRIVES;
				popup->captionname = "MyComputer";
			}
			else if ( _stricmp(pszName, "network") == NULL )
			{
				clsid = CSIDL_NETWORK;
				popup->captionname = "Network";
			}
			else if ( _stricmp(pszName, "printers") == NULL )
			{
				clsid = CSIDL_PRINTERS;
				popup->captionname = "Printers";
			}
			else if ( _stricmp(pszName, "recentdocuments") == NULL )
			{
				clsid = CSIDL_RECENT;
				popup->captionname = "RecentDocuments";
			}
			else if ( _stricmp(pszName, "recyclebin") == NULL )
			{
				clsid = CSIDL_BITBUCKET;
				popup->captionname = "RecycleBin";
			}
			else if ( _stricmp(pszName, "tasks") == NULL )
			{
				clsid = -1;
				popup->captionname = "Tasks";
			}

			if (clsid == 0)
				return;

			if ( clsid == -1 )
				popup->ScanTasks();
			else
				popup->ScanShellFolder(clsid);
		}
		else
		{
			CString easyWorker = path;
			if (easyWorker.Right(1) == '\\')
				easyWorker = easyWorker.Left(easyWorker.GetLength()-1);
			easyWorker = easyWorker.Right(easyWorker.GetLength()-easyWorker.ReverseFind('\\')-1);
			popup->captionname = (string) easyWorker;
			popup->ScanSubMenu((string) path, -2);
		}

		int screenWidth = iScreenWidth;
		int screenHeight = iScreenHeight;

		if ( iRealMonitors > 1 )
		{
			POINT pt = {ptCursorPos.x, ptCursorPos.y};
			HMONITOR hMonitor = MonitorFromPoint(pt, MONITOR_DEFAULTTONEAREST);
						
			MONITORINFO mi = { sizeof(MONITORINFO) };
			GetMonitorInfo(hMonitor, &mi);

			screenWidth = mi.rcMonitor.right-mi.rcMonitor.left;
			screenHeight = mi.rcMonitor.bottom-mi.rcMonitor.top;
		}

		if (ptCursorPos.y+popup->getHeight()/2 > screenHeight)
		{
			anchor.Replace("top", "bottom");
			if (anchor.Find("center") == 0)
			{
				anchor = anchor.Right(anchor.GetLength()-6);
				anchor.Insert(0, "bottom");
			}
		}
		else if (ptCursorPos.y+popup->getHeight() > screenHeight && ptCursorPos.y+popup->getHeight()/2 <= screenHeight)
		{
			anchor.Replace("top", "center");
		}
								
		if (ptCursorPos.x+popup->getWidth()/2 > screenWidth)
		{
			anchor.Replace("left", "right");
			if (anchor.Find("center", 1) == 3 || anchor.Find("center", 1) == 6)
			{
				anchor = anchor.Left(anchor.Find("center", 1));
				anchor += "right";
			}
		}
		else if (ptCursorPos.x+popup->getWidth() > screenWidth && ptCursorPos.x+popup->getWidth()/2 <= screenWidth)
		{
			anchor.Replace("left", "center");
		}

		//Calculate Position due to arguments
		int tmpx = ptCursorPos.x;
		int tmpy = ptCursorPos.y;
		PreCalcPosition(tmpx, tmpy, popup->getWidth(), popup->getHeight(), anchor);
		ptCursorPos.x = tmpx;
		ptCursorPos.y = tmpy;

		popup->reposition(1, ptCursorPos.x, ptCursorPos.y, 0, 0);
		popup->show();
	}
}

void BangLoadPopupCFG(HWND hwndCaller, const char *pszName, const char *pszArguments)
{
	char prefix[64] = {0};
	char path[MAX_LINE_LENGTH] = {0};
	char *posbuffers[] = {prefix, path};

	if ( _stricmp(pszName, "internal") != NULL )
		listLoadedPopupCFG.insert(listLoadedPopupCFG.end(), (string) pszArguments);

	LCTokenize( pszArguments, posbuffers, 2, NULL);

	StringList cfgList;
	GetxRCList( prefix, cfgList, path );

	if ( cfgList.size() >= 3 )
		PreParseStepRC(cfgList);
}

void BangTasks(HWND hwndCaller, const char *pszName, const char *pszArguments)
{
	BangDynamicPopup(hwndCaller, "Tasks", pszArguments);
}

void BangControlPanel(HWND hwndCaller, const char *pszName, const char *pszArguments)
{
	BangDynamicPopup(hwndCaller, "ControlPanel", pszArguments);
}

void BangMyComputer(HWND hwndCaller, const char *pszName, const char *pszArguments)
{
	BangDynamicPopup(hwndCaller, "MyComputer", pszArguments);
}

void BangNetwork(HWND hwndCaller, const char *pszName, const char *pszArguments)
{
	BangDynamicPopup(hwndCaller, "Network", pszArguments);
}

void BangPrinters(HWND hwndCaller, const char *pszName, const char *pszArguments)
{
	BangDynamicPopup(hwndCaller, "Printers", pszArguments);
}

void BangRecentDocuments(HWND hwndCaller, const char *pszName, const char *pszArguments)
{
	BangDynamicPopup(hwndCaller, "RecentDocuments", pszArguments);
}

void BangRecycleBin(HWND hwndCaller, const char *pszName, const char *pszArguments)
{
	BangDynamicPopup(hwndCaller, "RecycleBin", pszArguments);
}

void BangReposition(HWND hwndCaller, const char *pszName, const char *pszArguments)
{
	int x;
	int y;
	char A[16], B[16], C[16], D[16], E[32];
	char *buffers[] = {A, B, C, D, E};

	int tokens = LCTokenize(pszArguments, buffers, 5, NULL);

	if ( tokens >= 2 )
	{
		for(PopupListIterator plit = listPopups.begin(); plit != listPopups.end(); plit++)
		{
			if ( _stricmp( ((*plit)->BangToCall + "reposition").c_str(), pszName) == NULL )
			{	
				if ( _stricmp(A, "currentx") == NULL)
					x = (*plit)->getX();
				else
					x = xParseCoord(A, 0, iScreenWidth, false);
				if ( _stricmp(B, "currenty") == NULL)
					y = (*plit)->getY();
				else
					y = xParseCoord(B, 0, iScreenHeight, true);

				if (tokens == 2)
				{
					(*plit)->reposition(1, x, y, 0, 0);
				}
				else if (tokens == 3)
				{
					PreCalcPosition(x, y, (*plit)->getWidth(), (*plit)->getHeight(), C);

					(*plit)->reposition(1, x, y, 0, 0);
				}
				else if (tokens == 4)
				{
					(*plit)->reposition(3, x, y, atoi( C ), atoi( D ), 1);
					(*plit)->repaint(true);
				}
				else if (tokens == 5)
				{
					PreCalcPosition(x, y, atoi( C ), atoi( D ), E);
					
					(*plit)->reposition(3, x, y, atoi( C ), atoi( D ), 1);
					(*plit)->repaint(true);
				}

				break;
			}
		}
	}
}

void BangPin(HWND hwndCaller, const char *pszName, const char *pszArguments)
{
	for(PopupListIterator plit = listPopups.begin(); plit != listPopups.end(); plit++)
	{
		if ( _stricmp( ((*plit)->BangToCall + "pin").c_str(), pszName) == NULL )
		{
			(*plit)->bPinnedPopupMode = true;
			if ( (*plit)->bPinnedPopupNotOnTop )
				(*plit)->setAlwaysOnTop(false);
			else
				(*plit)->setAlwaysOnTop(true);
			(*plit)->show();
			break;
		}
	}
}

void BangUnPin(HWND hwndCaller, const char *pszName, const char *pszArguments)
{
	for(PopupListIterator plit = listPopups.begin(); plit != listPopups.end(); plit++)
	{
		if ( _stricmp( ((*plit)->BangToCall + "unpin").c_str(), pszName) == NULL )
		{
			(*plit)->bPinnedPopupMode = false;
			(*plit)->setAlwaysOnTop(true);
			(*plit)->hide();
			break;
		}
	}
}

void BangTogglePin(HWND hwndCaller, const char *pszName, const char *pszArguments)
{
	for(PopupListIterator plit = listPopups.begin(); plit != listPopups.end(); plit++)
	{
		if ( _stricmp( ((*plit)->BangToCall + "togglepin").c_str(), pszName) == NULL )
		{
			if ((*plit)->bPinnedPopupMode)
			{
				(*plit)->bPinnedPopupMode = false;
				(*plit)->setAlwaysOnTop(true);
				(*plit)->hide();
			}
			else
			{
				(*plit)->bPinnedPopupMode = true;
				if ( (*plit)->bPinnedPopupNotOnTop )
					(*plit)->setAlwaysOnTop(false);
				else
					(*plit)->setAlwaysOnTop(true);
				(*plit)->show();
			}
			break;
		}
	}
}

void BangToggleShade(HWND hwndCaller, const char *pszName, const char *pszArguments)
{
	for(PopupListIterator plit = listPopups.begin(); plit != listPopups.end(); plit++)
	{
		if ( _stricmp( ((*plit)->BangToCall + "toggleshade").c_str(), pszName) == NULL )
		{
			(*plit)->ToggleShade();

			break;
		}
	}
}

void BangToggle(HWND hwndCaller, const char *pszName, const char *pszArguments)
{
	CString convert = pszName;
	convert = convert.Left(convert.GetLength()-6);

	bInternalPopupToggle = true;

	BangPopup(hwndCaller, convert, pszArguments);
}

void BangRescanPopup(HWND hwndCaller, const char *pszName, const char *pszArguments)
{
	if ( _stricmp(pszArguments, "reload") == NULL )
		LSExecute(NULL, "!reload", SW_SHOWNORMAL);

	listVisiblePopups.clear();

	for(PopupListIterator plit = listPopups.begin(); plit != listPopups.end(); plit++)
	{
		(*plit)->hide(true/*noFade*/);
		delete *plit;
		*plit = NULL;
	}

	listPopups.clear();

	if (pDefaultPaintClasses)
		delete pDefaultPaintClasses;

	pDefaultPaintClasses = NULL;

	for(GhostPopupListIterator gtlit = plistDefaultGhostPopup->begin(); gtlit != plistDefaultGhostPopup->end(); gtlit++)
	{
		delete *gtlit;
		*gtlit = NULL;
	}

	plistDefaultGhostPopup->clear();
	delete plistDefaultGhostPopup;

	plistDefaultGhostPopup = NULL;

	for(GhostButtonListIterator glit = plistDefaultGhostButton->begin(); glit != plistDefaultGhostButton->end(); glit++)
	{
		delete *glit;
		*glit = NULL;
	}

	plistDefaultGhostButton->clear();
	delete plistDefaultGhostButton;

	plistDefaultGhostButton = NULL;

	bool bUseAlphaMap = false;
	if ( IsOS(OS_2KXP) )
		bUseAlphaMap = Getx2RCBool("xPopup", "AlphaMap");

	bool bUseFolderDefault = Getx2RCBool("xPopup", "ActiveFolderUseFolderDefault");

	//Create "xPopup" Default Painting Classes
	pDefaultPaintClasses = new CompletePaintClasses();
	
	pDefaultPaintClasses->pPopupTexture->configure("xPopup", "", bUseAlphaMap);

	//------------------
	//Texture
	//------------------
	pDefaultPaintClasses->pEntryTexture->configure("xPopup", "Entry", bUseAlphaMap);
	pDefaultPaintClasses->pActiveEntryTexture->configure("xPopup", "ActiveEntry", bUseAlphaMap, pDefaultPaintClasses->pEntryTexture);
	pDefaultPaintClasses->pFolderTexture->configure("xPopup", "Folder", bUseAlphaMap, pDefaultPaintClasses->pEntryTexture);
	pDefaultPaintClasses->pActiveFolderTexture->configure("xPopup", "ActiveFolder", bUseAlphaMap, (bUseFolderDefault ? pDefaultPaintClasses->pFolderTexture : pDefaultPaintClasses->pActiveEntryTexture));
	
	pDefaultPaintClasses->pInfoTexture->configure("xPopup", "Info", bUseAlphaMap);
	pDefaultPaintClasses->pActiveInfoTexture->configure("xPopup", "ActiveInfo", bUseAlphaMap, pDefaultPaintClasses->pInfoTexture);
	
	pDefaultPaintClasses->pTitleTexture->configure("xPopup", "Title", bUseAlphaMap);
	
	pDefaultPaintClasses->pBottomTexture->configure("xPopup", "Bottom", bUseAlphaMap);

	pDefaultPaintClasses->pSeparatorTexture->configure("xPopup", "Separator", bUseAlphaMap);

	pDefaultPaintClasses->pArrowTexture->configure("xPopup", "Arrow", bUseAlphaMap);
	pDefaultPaintClasses->pActiveArrowTexture->configure("xPopup", "ActiveArrow", bUseAlphaMap, pDefaultPaintClasses->pArrowTexture);

	//------------------
	//Icon
	//------------------
	PaintIcon *defaulticon = Create_xIconClass();
	defaulticon->configure("xPopup", "", true, bUseAlphaMap);
	
	pDefaultPaintClasses->pEntryIcon->configure("xPopup", "Entry", true, bUseAlphaMap, defaulticon);
	pDefaultPaintClasses->pActiveEntryIcon->configure("xPopup", "ActiveEntry", true, bUseAlphaMap, pDefaultPaintClasses->pEntryIcon);
	pDefaultPaintClasses->pFolderIcon->configure("xPopup", "Folder", true, bUseAlphaMap, pDefaultPaintClasses->pEntryIcon);
	pDefaultPaintClasses->pActiveFolderIcon->configure("xPopup", "ActiveFolder", true, bUseAlphaMap, (bUseFolderDefault ? pDefaultPaintClasses->pFolderIcon : pDefaultPaintClasses->pActiveEntryIcon));
	
	Destroy_xIconClass(defaulticon);

	pDefaultPaintClasses->pInfoIcon->configure("xPopup", "Info", true, bUseAlphaMap);
	pDefaultPaintClasses->pActiveInfoIcon->configure("xPopup", "ActiveInfo", true, bUseAlphaMap, pDefaultPaintClasses->pInfoIcon);

	if ( Getx2RCBool("xPopup", "TitleShowIcon") )
	{
		pDefaultPaintClasses->pTitleIcon->configure("xPopup", "Title", true, bUseAlphaMap);
	}
	else
	{
		Destroy_xIconClass(pDefaultPaintClasses->pTitleIcon);
		pDefaultPaintClasses->pTitleIcon = NULL;
	}
	if ( Getx2RCBool("xPopup", "PinnedShowIcon") )
	{
		pDefaultPaintClasses->pPinnedIcon->configure("xPopup", "Pinned", true, bUseAlphaMap);
	}
	else
	{
		Destroy_xIconClass(pDefaultPaintClasses->pPinnedIcon);
		pDefaultPaintClasses->pPinnedIcon = NULL;
	}

	//------------------
	//Text
	//------------------
	PaintText *defaulttext = Create_xTextClass();
	defaulttext->configure("xPopup", "", bUseAlphaMap);

	pDefaultPaintClasses->pEntryText->configure("xPopup", "Entry", bUseAlphaMap, defaulttext);
	pDefaultPaintClasses->pActiveEntryText->configure("xPopup", "ActiveEntry", bUseAlphaMap, pDefaultPaintClasses->pEntryText);
	pDefaultPaintClasses->pFolderText->configure("xPopup", "Folder", bUseAlphaMap, pDefaultPaintClasses->pEntryText);
	pDefaultPaintClasses->pActiveFolderText->configure("xPopup", "ActiveFolder", bUseAlphaMap, (bUseFolderDefault ? pDefaultPaintClasses->pFolderText : pDefaultPaintClasses->pActiveEntryText));
	
	Destroy_xTextClass(defaulttext);

	pDefaultPaintClasses->pInfoText->configure("xPopup", "Info", bUseAlphaMap);
	pDefaultPaintClasses->pActiveInfoText->configure("xPopup", "ActiveInfo", bUseAlphaMap, pDefaultPaintClasses->pInfoText);
	
	pDefaultPaintClasses->pTitleText->configure("xPopup", "Title", bUseAlphaMap);

	if ( Getx2RCBool("xPopup", "UseTextArrow") )
	{
		pDefaultPaintClasses->pArrowText->configure("xPopup", "Arrow", bUseAlphaMap);
		pDefaultPaintClasses->pActiveArrowText->configure("xPopup", "ActiveArrow", bUseAlphaMap, pDefaultPaintClasses->pArrowText);
	}
	else
	{
		Destroy_xTextClass(pDefaultPaintClasses->pArrowText);
		pDefaultPaintClasses->pArrowText = NULL;
		Destroy_xTextClass(pDefaultPaintClasses->pActiveArrowText);
		pDefaultPaintClasses->pActiveArrowText = NULL;
	}

	//Create Default OverlayPopup Painting
	//Initialize Basic Helper Strings
	int numTokens = 0;
	string one = "";
	string two = "";
	string three = "";
	string four = "";
	string five = "";
	string six = "";
	string seven = "";

	plistDefaultGhostPopup = new GhostPopupList();
	plistDefaultGhostButton = new GhostButtonList();

	StringList OverlayPaintingList = GetxRCLineList( ("*xPopupOverlayPopup") );
	
	if ( !OverlayPaintingList.empty() )
	{
		for(StringListIterator sit = OverlayPaintingList.begin(); sit != OverlayPaintingList.end(); sit++)
		{
			numTokens = SplitString((*sit), one, two, three, four, five, six, seven);
			
			if ( numTokens >= 5 )
			{
				GhostPopup* newpopup = new GhostPopup();
			
				newpopup->strOrgX = one;
				newpopup->strOrgY = two;

				newpopup->strOrgWidth = three;
				newpopup->strOrgHeight = four;
	
				if ( numTokens >= 6 && IsOS( OS_2KXP ) && (strchr(six.c_str(), 'a') != 0 || strchr(six.c_str(), 'A') != 0) )
					newpopup->pPopupTexture->configure(five.c_str(), "", true);
				else
					newpopup->pPopupTexture->configure(five.c_str(), "", false);
			
				plistDefaultGhostPopup->insert(plistDefaultGhostPopup->end(), newpopup);
			}
		}
	}

	OverlayPaintingList.clear();
	//Create Default OverlayButton Painting
	OverlayPaintingList = GetxRCLineList( ("*xPopupOverlayEntry") );
	
	if ( !OverlayPaintingList.empty() )
	{
		for(StringListIterator sit = OverlayPaintingList.begin(); sit != OverlayPaintingList.end(); sit++)
		{
			//Use Split Helper Strings
			numTokens = SplitString((*sit), one, two, three, four, five, six, seven);
			
			if ( numTokens >= 5 )
			{
				GhostButton* newbutton = new GhostButton();
			
				newbutton->strOrgX = one;
				newbutton->strOrgY = two;

				newbutton->strOrgWidth = three;
				newbutton->strOrgHeight = four;

				if ( numTokens >= 6 && IsOS( OS_2KXP ) && (strchr(six.c_str(), 'a') != 0 || strchr(six.c_str(), 'A') != 0) )
				{
					newbutton->pEntryTexture->configure(five.c_str(), "Entry", true);
					newbutton->pActiveEntryTexture->configure(five.c_str(), "ActiveEntry", true, newbutton->pEntryTexture);
					newbutton->pFolderTexture->configure(five.c_str(), "Folder", true, newbutton->pEntryTexture);
					newbutton->pActiveFolderTexture->configure(five.c_str(), "ActiveFolder", true, (bUseFolderDefault ? newbutton->pFolderTexture : newbutton->pActiveEntryTexture));
					
					newbutton->pInfoTexture->configure(five.c_str(), "Info", true);
					newbutton->pActiveInfoTexture->configure(five.c_str(), "ActiveInfo", true, newbutton->pInfoTexture);

					newbutton->pTitleTexture->configure(five.c_str(), "Title", true);
				
					newbutton->pBottomTexture->configure(five.c_str(), "Bottom", true);
				
					newbutton->pSeparatorTexture->configure(five.c_str(), "Separator", true);
				}
				else
				{
					newbutton->pEntryTexture->configure(five.c_str(), "Entry", false);
					newbutton->pActiveEntryTexture->configure(five.c_str(), "ActiveEntry", false, newbutton->pEntryTexture);
					newbutton->pFolderTexture->configure(five.c_str(), "Folder", false, newbutton->pEntryTexture);
					newbutton->pActiveFolderTexture->configure(five.c_str(), "ActiveFolder", false, (bUseFolderDefault ? newbutton->pFolderTexture : newbutton->pActiveEntryTexture));
					
					newbutton->pInfoTexture->configure(five.c_str(), "Info", false);
					newbutton->pActiveInfoTexture->configure(five.c_str(), "ActiveInfo", false, newbutton->pInfoTexture);

					newbutton->pTitleTexture->configure(five.c_str(), "Title", false);
				
					newbutton->pBottomTexture->configure(five.c_str(), "Bottom", false);
				
					newbutton->pSeparatorTexture->configure(five.c_str(), "Separator", false);
				}

				plistDefaultGhostButton->insert(plistDefaultGhostButton->end(), newbutton);
			}
		}
	}

	StartUp();

	for(StringListIterator sit = listLoadedPopupCFG.begin(); sit != listLoadedPopupCFG.end(); sit++)
		BangLoadPopupCFG(NULL, "internal", (*sit).c_str());
}