/***************************************************************************
****************************************************************************

This is a part of the xModule Source code.

Copyright (C) 2003-2008 The LS-Universe Team

    Homepage: http://www.ls-universe.info

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************
****************************************************************************/

#include "stdafx.h"
#include "Popup.h"

#include "ContextMenu.h"
#include "NotifyDirCheck.h"

#include "GhostButton.h"
#include "GhostPopup.h"

#include <shlwapi.h>
#include <algorithm>

#include <process.h>

//////////////////////////////////////////////////////////////////////
// Defines
//////////////////////////////////////////////////////////////////////

extern PopupList listVisiblePopups;

extern int iRealMonitors;

extern int iBitsPerPixel;

extern string dragndropSource;
extern string dragndropDestination;

#define GET_X_LPARAM(lp) ((int)(short)LOWORD(lp))
#define GET_Y_LPARAM(lp) ((int)(short)HIWORD(lp))

#define WM_MOUSEWHEEL 0x020A

  #define WM_XBUTTONDOWN 0x020B
  #define WM_XBUTTONUP 0x020C
  //#define WM_XBUTTONDBLCLK 0x020D

  #define MK_XBUTTON1 0x0020
  #define MK_XBUTTON2 0x0040
  #define XBUTTON1 0x0001
  #define XBUTTON2 0x0002

//Alpha Transparency
typedef DWORD (WINAPI *PSLWA)(HWND, DWORD, BYTE, DWORD);
static PSLWA pSetLayeredWindowAttributes = NULL;

typedef BOOL (WINAPI *PULW)(HWND hwnd, HDC hdcDst, POINT *pptDst, SIZE *psize, HDC hdcSrc, POINT *pptSrc, COLORREF crKey, BLENDFUNCTION *pblend, DWORD dwFlags);
static PULW pUpdateLayeredWindow = NULL;

//extern VOID (WINAPI *pSwitchToThisWindow)(HWND,BOOL);

#define TIMER_MOUSETRACK	1
#define TIMER_ALPHAFADEIN	2
#define TIMER_ALPHAFADEOUT	3
#define TIMER_UPDATE		4
#define TIMER_ACTIVE		5
#define TIMER_HOVER			6
#define TIMER_CHORDING		7
#define TIMER_AUTOUNLOCK	8

#define TIMER_MULTISCROLLUP		10
#define TIMER_MULTISCROLLDOWN	11

extern ProcessStatsRequestFunc m_pProcessStatsRequest;

extern CompletePaintClasses *pDefaultPaintClasses;

extern GhostPopupList *plistDefaultGhostPopup;
extern GhostButtonList *plistDefaultGhostButton;

extern int iScreenX;
extern int iScreenY;

extern int iScreenWidth;
extern int iScreenHeight;

#define LM_HIDEPOPUP				9183
#define LM_BRINGTOFRONT				8891

//////////////////////////////////////////////////////////////////////
// Konstruktion/Destruktion
//////////////////////////////////////////////////////////////////////

Popup::Popup(const string &name, string subname, CompletePaintClasses *pParentPaintClasses, GhostPopupList *plistParentGhostPopup, GhostButtonList *plistParentGhostButton, int oldsubFolderLevel, string groupName) :
	actualx(0),
	actualy(0),
	actualwidth(0),
	actualheight(0),

	bAlphaInitialized(false),

	hWnd(NULL),
	hWndOwner(NULL),
	hInstance(NULL),

	pTooltip(NULL),
	hWndToolTip(NULL),

	bResetBackground(false),
	bLockKeyboard(false),

	iTooltipMode(0),

	iSnappedCode(0),

	bCustomPaintingClasses(false),
	pPaintClasses(NULL),

	isTaskPopup(false),
	hasHookedWindows(false),
	hasInfoEntry(false),

	hasDynamicContent(false),

	bMoveButtonDown(false),

	dropShadow(false),

	bOpenToTheLeft(false),

	pParentPopup(NULL),
	pParentEntry(NULL),

	startentry(NULL),
	lastactiveentry(NULL),
	hoverentry(NULL),

	lastkeyentry(-1),

	bPinnedPopupMode(false),

	defaultDC(NULL),
	defaultBitmap(NULL),
	olddefaultBMP(NULL),

	isFading(false),
	fadeBackupDC(NULL),
	fadeBackupBitmap(NULL),
	oldfadeBackupBMP(NULL),

	fakealphaMap(false),
	alphaMap(false),

	subFolderLevel(-1),

	iCurrentPart(0),
	iMaxParts(0),
	iMaxMenuBreaks(0),

	dynamicPopup(false),
	rootScanPath(""),
	rootUniversalAction(""),

	showExtensions(false),

	//Bang which calls this Popup (only Top Level Popups have this)
	BangToCall(""),

	execSoundPlayed(false),

	UpdateAlpha(0),
	alphaFadeStep(4),
	alphaFadeDelay(10),

	isMultiScrollTimerRunning(0),

	mouseInside(false),
	mousePressed(false),
	mouseMoved(false),

	manualMove(true),
	
	visible(false),

	inChordingTimeOut(false),

	shaded(-1),

	alwaysOnTop(false),

	str_completeCheckStringBackup(""),

	tmpActionForTimer(""),
	tmpPidlForTimer(NULL),

	strtitleLClickAction(""),
	strtitleMClickAction(""),
	strtitleRClickAction(""),

	strbottomLClickAction(""),
	strbottomMClickAction(""),
	strbottomRClickAction("")
{
	//Specifies the Cfg Prefix
	strName = name;

	if ( groupName.empty() )
	{
		if ( _stricmp(strName.c_str(), "xPopup") != NULL && !GetxRCBool(strName.c_str(), "AddToGroup") )
			PrefixedLSSetVariable(strName, "AddToGroup", "xPopup");
	}
	else
	{
		if ( _stricmp(strName.c_str(), groupName.c_str()) != NULL && !GetxRCBool(strName.c_str(), "AddToGroup") )
			PrefixedLSSetVariable(strName, "AddToGroup", groupName);
	}

	subpopupList.clear();
	entrylist.clear();

	//Specifies the internal name (Caption/Title) plus "sub_" Prefix
	this->subname = subname;

		CString help = subname.c_str();
		help.Replace("sub_", "");

		CString help1 = help.Left(1);
		help1.MakeUpper();
	
	//Specifies the Caption/Title
	this->captionname = (help1 + help.Right(help.GetLength()-1));

	this->subFolderLevel = oldsubFolderLevel+1;

	oldmovetracker.x = 0;
	oldmovetracker.y = 0;

	//Only Create Painting Classes, if it's a TopLevel Popup
	//-> Nothing is sent to the "SubPopup"
	if (pParentPaintClasses == NULL && plistParentGhostPopup == NULL && plistParentGhostButton == NULL)
	{
		bCustomPaintingClasses = true;

		bool bUseAlphaMap = false;
		if ( IsOS(OS_2KXP) )
			bUseAlphaMap = Getx2RCBool(strName.c_str(), "AlphaMap");

		bool bUseFolderDefault = Getx2RCBool(strName.c_str(), "ActiveFolderUseFolderDefault");

		//Create name Default Painting Classes
		pPaintClasses = new CompletePaintClasses();
		
		pPaintClasses->pPopupTexture->configure(strName.c_str(), "", bUseAlphaMap);

		//------------------
		//Texture
		//------------------
		pPaintClasses->pEntryTexture->configure(strName.c_str(), "Entry", bUseAlphaMap);
		pPaintClasses->pActiveEntryTexture->configure(strName.c_str(), "ActiveEntry", bUseAlphaMap, pPaintClasses->pEntryTexture);
		pPaintClasses->pFolderTexture->configure(strName.c_str(), "Folder", bUseAlphaMap, pPaintClasses->pEntryTexture);
		pPaintClasses->pActiveFolderTexture->configure(strName.c_str(), "ActiveFolder", bUseAlphaMap, (bUseFolderDefault ? pPaintClasses->pFolderTexture : pPaintClasses->pActiveEntryTexture));
		
		pPaintClasses->pInfoTexture->configure(strName.c_str(), "Info", bUseAlphaMap);
		pPaintClasses->pActiveInfoTexture->configure(strName.c_str(), "ActiveInfo", bUseAlphaMap, pPaintClasses->pInfoTexture);
		
		pPaintClasses->pTitleTexture->configure(strName.c_str(), "Title", bUseAlphaMap);
		
		pPaintClasses->pBottomTexture->configure(strName.c_str(), "Bottom", bUseAlphaMap);

		pPaintClasses->pSeparatorTexture->configure(strName.c_str(), "Separator", bUseAlphaMap);

		pPaintClasses->pArrowTexture->configure(strName.c_str(), "Arrow", bUseAlphaMap);
		pPaintClasses->pActiveArrowTexture->configure(strName.c_str(), "ActiveArrow", bUseAlphaMap, pPaintClasses->pArrowTexture);

		//------------------
		//Icon
		//------------------
		PaintIcon *defaulticon = Create_xIconClass();
		defaulticon->configure(strName.c_str(), "", true, bUseAlphaMap);

		pPaintClasses->pEntryIcon->configure(strName.c_str(), "Entry", true, bUseAlphaMap, defaulticon);
		pPaintClasses->pActiveEntryIcon->configure(strName.c_str(), "ActiveEntry", true, bUseAlphaMap, pPaintClasses->pEntryIcon);
		pPaintClasses->pFolderIcon->configure(strName.c_str(), "Folder", true, bUseAlphaMap, pPaintClasses->pEntryIcon);
		pPaintClasses->pActiveFolderIcon->configure(strName.c_str(), "ActiveFolder", true, bUseAlphaMap, (bUseFolderDefault ? pPaintClasses->pFolderIcon : pPaintClasses->pActiveEntryIcon));
		
		Destroy_xIconClass(defaulticon);

		pPaintClasses->pInfoIcon->configure(strName.c_str(), "Info", true, bUseAlphaMap);
		pPaintClasses->pActiveInfoIcon->configure(strName.c_str(), "ActiveInfo", true, bUseAlphaMap, pPaintClasses->pInfoIcon);

		if ( Getx2RCBool(strName.c_str(), "TitleShowIcon") )
		{
			pPaintClasses->pTitleIcon->configure(strName.c_str(), "Title", true, bUseAlphaMap);
		}
		else
		{
			Destroy_xIconClass(pPaintClasses->pTitleIcon);
			pPaintClasses->pTitleIcon = NULL;
		}
		if ( Getx2RCBool(strName.c_str(), "PinnedShowIcon") )
		{
			pPaintClasses->pPinnedIcon->configure(strName.c_str(), "Pinned", true, bUseAlphaMap);
		}
		else
		{
			Destroy_xIconClass(pPaintClasses->pPinnedIcon);
			pPaintClasses->pPinnedIcon = NULL;
		}

		//------------------
		//Text
		//------------------
		PaintText *defaulttext = Create_xTextClass();
		defaulttext->configure(strName.c_str(), "", bUseAlphaMap);

		pPaintClasses->pEntryText->configure(strName.c_str(), "Entry", bUseAlphaMap, defaulttext);
		pPaintClasses->pActiveEntryText->configure(strName.c_str(), "ActiveEntry", bUseAlphaMap, pPaintClasses->pEntryText);
		pPaintClasses->pFolderText->configure(strName.c_str(), "Folder", bUseAlphaMap, pPaintClasses->pEntryText);
		pPaintClasses->pActiveFolderText->configure(strName.c_str(), "ActiveFolder", bUseAlphaMap, (bUseFolderDefault ? pPaintClasses->pFolderText : pPaintClasses->pActiveEntryText));
		
		Destroy_xTextClass(defaulttext);

		pPaintClasses->pInfoText->configure(strName.c_str(), "Info", bUseAlphaMap);
		pPaintClasses->pActiveInfoText->configure(strName.c_str(), "ActiveInfo", bUseAlphaMap, pPaintClasses->pInfoText);
		
		pPaintClasses->pTitleText->configure(strName.c_str(), "Title", bUseAlphaMap);

		if ( Getx2RCBool(strName.c_str(), "UseTextArrow") )
		{
			pPaintClasses->pArrowText->configure(strName.c_str(), "Arrow", bUseAlphaMap);
			pPaintClasses->pActiveArrowText->configure(strName.c_str(), "ActiveArrow", bUseAlphaMap, pPaintClasses->pArrowText);
		}
		else
		{
			Destroy_xTextClass(pPaintClasses->pArrowText);
			pPaintClasses->pArrowText = NULL;
			Destroy_xTextClass(pPaintClasses->pActiveArrowText);
			pPaintClasses->pActiveArrowText = NULL;
		}

		//Create Default OverlayPopup Painting
		//Initialize Basic Helper Strings
		int numTokens = 0;
		string one = "";
		string two = "";
		string three = "";
		string four = "";
		string five = "";
		string six = "";
		string seven = "";

		plistGhostPopup = new GhostPopupList();
		plistGhostButton = new GhostButtonList();

		StringList OverlayPaintingList = GetxRCLineList( ("*" + strName + "OverlayPopup") );
		
		if ( !OverlayPaintingList.empty() )
		{
			for(StringListIterator sit = OverlayPaintingList.begin(); sit != OverlayPaintingList.end(); sit++)
			{
				numTokens = SplitString((*sit), one, two, three, four, five, six, seven);
				
				if ( numTokens >= 5 )
				{
					GhostPopup* newpopup = new GhostPopup();
				
					newpopup->strOrgX = one;
					newpopup->strOrgY = two;

					newpopup->strOrgWidth = three;
					newpopup->strOrgHeight = four;

					if ( numTokens >= 6 && IsOS( OS_2KXP ) )
					{
						if ( (strchr(six.c_str(), 'a') != 0 || strchr(six.c_str(), 'A') != 0) )
							newpopup->pPopupTexture->configure(five.c_str(), "", true);
						else
							newpopup->pPopupTexture->configure(five.c_str(), "", false);
					}
			
					plistGhostPopup->insert(plistGhostPopup->end(), newpopup);
				}
			}
		}

		OverlayPaintingList.clear();
		//Create Default OverlayButton Painting
		OverlayPaintingList = GetxRCLineList( ("*" + strName + "OverlayEntry") );
		
		if ( !OverlayPaintingList.empty() )
		{
			for(StringListIterator sit = OverlayPaintingList.begin(); sit != OverlayPaintingList.end(); sit++)
			{
				//Use Split Helper Strings
				numTokens = SplitString((*sit), one, two, three, four, five, six, seven);
				
				if ( numTokens >= 5 )
				{
					GhostButton* newbutton = new GhostButton();
				
					newbutton->strOrgX = one;
					newbutton->strOrgY = two;

					newbutton->strOrgWidth = three;
					newbutton->strOrgHeight = four;
				
					if ( numTokens >= 6 && IsOS( OS_2KXP ) && (strchr(six.c_str(), 'a') != 0 || strchr(six.c_str(), 'A') != 0) )
					{
						newbutton->pEntryTexture->configure(five.c_str(), "Entry", true);
						newbutton->pActiveEntryTexture->configure(five.c_str(), "ActiveEntry", true, newbutton->pEntryTexture);
						newbutton->pFolderTexture->configure(five.c_str(), "Folder", true, newbutton->pEntryTexture);
						newbutton->pActiveFolderTexture->configure(five.c_str(), "ActiveFolder", true, (bUseFolderDefault ? newbutton->pFolderTexture : newbutton->pActiveEntryTexture));
						
						newbutton->pInfoTexture->configure(five.c_str(), "Info", true);
						newbutton->pActiveInfoTexture->configure(five.c_str(), "ActiveInfo", true, newbutton->pInfoTexture);

						newbutton->pTitleTexture->configure(five.c_str(), "Title", true);
					
						newbutton->pBottomTexture->configure(five.c_str(), "Bottom", true);
					
						newbutton->pSeparatorTexture->configure(five.c_str(), "Separator", true);
					}
					else
					{
						newbutton->pEntryTexture->configure(five.c_str(), "Entry", false);
						newbutton->pActiveEntryTexture->configure(five.c_str(), "ActiveEntry", false, newbutton->pEntryTexture);
						newbutton->pFolderTexture->configure(five.c_str(), "Folder", false, newbutton->pEntryTexture);
						newbutton->pActiveFolderTexture->configure(five.c_str(), "ActiveFolder", false, (bUseFolderDefault ? newbutton->pFolderTexture : newbutton->pActiveEntryTexture));
						
						newbutton->pInfoTexture->configure(five.c_str(), "Info", false);
						newbutton->pActiveInfoTexture->configure(five.c_str(), "ActiveInfo", false, newbutton->pInfoTexture);

						newbutton->pTitleTexture->configure(five.c_str(), "Title", false);
					
						newbutton->pBottomTexture->configure(five.c_str(), "Bottom", false);
					
						newbutton->pSeparatorTexture->configure(five.c_str(), "Separator", false);
					}

					plistGhostButton->insert(plistGhostButton->end(), newbutton);
				}
			}
		}
	}
	else
	{
		pPaintClasses = pParentPaintClasses;

		plistGhostPopup = plistParentGhostPopup;
		plistGhostButton = plistParentGhostButton;
	}
}

Popup::~Popup()
{
	for(CheckDirectoryListIterator dcit = checkdirectorylist.begin(); dcit != checkdirectorylist.end(); dcit++)
	{
		(*dcit)->Stop();
		delete *dcit;
		*dcit = NULL;
	}

	checkdirectorylist.clear();

	if (pTooltip)
	{
		Destroy_xTooltipClass(pTooltip);
		pTooltip = NULL;
	}

	if ( !BangToCall.empty() /*-> TopLevel Popup*/ )
	{
		// unregister LS messages
        int msgs[] = { LM_HIDEPOPUP, 0 };

		SendMessage(GetLitestepWnd(), LM_UNREGISTERMESSAGE,
            (WPARAM)hWnd, (LPARAM) msgs);

		RemoveBangCommand(BangToCall.c_str());
		
		string strRemoveBang = "";

		//Reposition Bang
		strRemoveBang = BangToCall;
		strRemoveBang.append("Reposition");
		RemoveBangCommand(strRemoveBang.c_str());
		
		//Pin Bang == Show
		strRemoveBang = BangToCall;
		strRemoveBang.append("Pin");
		RemoveBangCommand(strRemoveBang.c_str());

		//UnPin Bang == Hide
		strRemoveBang = BangToCall;
		strRemoveBang.append("UnPin");
		RemoveBangCommand(strRemoveBang.c_str());

		//Toggle bPinnedPopupMode == Toggle
		strRemoveBang = BangToCall;
		strRemoveBang.append("TogglePin");
		RemoveBangCommand(strRemoveBang.c_str());

		//Toggle ShadeMode == Toggle
		strRemoveBang = BangToCall;
		strRemoveBang.append("ToggleShade");
		RemoveBangCommand(strRemoveBang.c_str());

		//Toggle Visiblity == Toggle
		strRemoveBang = BangToCall;
		strRemoveBang.append("Toggle");
		RemoveBangCommand(strRemoveBang.c_str());
	}

	for(PopupListIterator plit = subpopupList.begin(); plit != subpopupList.end(); plit++)
	{
		(*plit)->hide(true/*noFade*/);
		delete *plit;
		*plit = NULL;
	}

	subpopupList.clear();

	for (EntryListIterator eit = entrylist.begin(); eit != entrylist.end(); ++eit)
	{
		delete *eit;
		*eit = NULL;
	}

	entrylist.clear();

	tasklist.clear();

	// shutdown timer
	if(IsWindow(hWnd))
	{
		KillTimer(hWnd, TIMER_MOUSETRACK);
		KillTimer(hWnd, TIMER_ALPHAFADEIN);
		KillTimer(hWnd, TIMER_ALPHAFADEOUT);
		KillTimer(hWnd, TIMER_UPDATE);
		KillTimer(hWnd, TIMER_ACTIVE);
		KillTimer(hWnd, TIMER_HOVER);
		KillTimer(hWnd, TIMER_CHORDING);
		KillTimer(hWnd, TIMER_AUTOUNLOCK);
		KillTimer(hWnd, TIMER_MULTISCROLLUP);
		KillTimer(hWnd, TIMER_MULTISCROLLDOWN);

		DestroyWindow(hWnd);
	}

	if ( !BangToCall.empty() /*-> TopLevel Popup*/ && hWndOwner != NULL )
		DestroyWindow(hWndOwner);

	ResetBackground();
	
	if (bCustomPaintingClasses) 
	{	
		if (pPaintClasses && pPaintClasses != pDefaultPaintClasses)
		{
			delete pPaintClasses;

			pPaintClasses = NULL;
		}

		if ( plistGhostPopup && plistGhostPopup != plistDefaultGhostPopup )
		{
			for(GhostPopupListIterator gtlit = plistGhostPopup->begin(); gtlit != plistGhostPopup->end(); gtlit++)
			{
				delete *gtlit;
				*gtlit = NULL;
			}

			plistGhostPopup->clear();
			delete plistGhostPopup;

			plistGhostPopup = NULL;
		}

		if ( plistGhostButton && plistGhostButton != plistDefaultGhostButton )
		{
			for(GhostButtonListIterator glit = plistGhostButton->begin(); glit != plistGhostButton->end(); glit++)
			{
				delete *glit;
				*glit = NULL;
			}

			plistGhostButton->clear();
			delete plistGhostButton;

			plistGhostButton = NULL;
		}
	}
}

void Popup::load(HINSTANCE hInstance)
{
	this->hInstance = hInstance;
	
	if (IsWindow(hWnd))	// window already exists...
		return;

	if ( IsOS(OS_XP) && Getx2RCBool(strName.c_str(), "DropShadow") )
		dropShadow = true;

	if ( !BangToCall.empty() /*-> TopLevel Popup*/ )
	{
		hWndOwner = CreateWindowEx(WS_EX_TOOLWINDOW,
								"xPopupOwner",
								BangToCall.c_str(),
								0,
								0, 0, 0, 0, 
								HWND_MESSAGE,
								0,
								hInstance,
								0);
	}
	else if ( pParentPopup != NULL )
	{
		hWndOwner = pParentPopup->gethWndOwner();
	}

	hWnd = CreateWindowEx( WS_EX_TOOLWINDOW,
							( dropShadow ? "xPopupXP" : "xPopup" ),
							(!BangToCall.empty() ? BangToCall.c_str() : captionname.c_str()),
							WS_POPUP | WS_CLIPCHILDREN,
							actualx, actualy,		
							actualwidth, actualheight,
							hWndOwner, 
							0,
							hInstance,
							this);
		
	if (hWnd)
	{
		SetWindowLong(hWnd, GWL_USERDATA, MAGIC_DWORD);

		// register LS messages
		if ( !BangToCall.empty() )
		{
			int msgs[] = { LM_HIDEPOPUP, 0 };

			SendMessage(GetLitestepWnd(), LM_REGISTERMESSAGE, (WPARAM)hWnd, (LPARAM) msgs);
		}

		reconfigure();
		
		setAlwaysOnTop(true);
	}
	else
	{
		delete this;
	}
}

void Popup::reconfigure()
{
	//Initialize Basic Helper Strings
	int numTokens = 0;
	string one = "";
	string two = "";
	string three = "";
	string four = "";
	string five = "";
	string six = "";
	string seven = "";	

	const NameValuePair captionPositionValues[] = {
			{ ".none", 0 },
			{ ".top", 1 },
			{ ".bottom", 2 },
			{ ".custom", 3 },
			{ "none", 0 },
			{ "top", 1 },
			{ "bottom", 2 },
			{ "custom", 3 },
			{ 0, 0 }
		};

	const NameValuePair quicklaunchValues[] = {
			{ ".none", 0 },
			{ ".horizontal", 1 },
			{ ".vertical", 2 },
			{ ".horizontal+text", -1 },
			{ ".vertical+text", -2 },
			{ "none", 0 },
			{ "horizontal", 1 },
			{ "vertical", 2 },
			{ "horizontal+text", -1 },
			{ "vertical+text", -2 },
			{ 0, 0 }
		};

	const NameValuePair tooltipValues[] = {
			{ ".none", 0 },
			{ ".all", 1 },
			{ ".defined", 2 },
			{ "none", 0 },
			{ "all", 1 },
			{ "defined", 2 },
			{ 0, 0 }
		};

	const NameValuePair alphaFadeValues[] = {
			{ ".none", 0 },
			{ "false", 0 },
			{ ".allpopups", 1 },
			{ "true", 1 },
			{ ".mainpopups", 2 },
			{ "none", 0 },
			{ "allpopups", 1 },
			{ "mainpopups", 2 },
			{ 0, 0}
		};

	const NameValuePair clickactionValues[] = {
			{ ".none", 0 },
			{ ".pin", 1 },
			{ ".unpin", 2 },
			{ ".togglepin", 3 },
			{ ".toggleshade", 4 },
			{ ".toggleontop", 5 },
			{ ".move", 6 },
			{ ".close", 7 },
			{ ".pageup", 8 },
			{ ".pagedown", 9 },
			{ ".execute", 10 },
			{ "none", 0 },
			{ "pin", 1 },
			{ "unpin", 2 },
			{ "togglepin", 3 },
			{ "toggleshade", 4 },
			{ "toggleontop", 5 },
			{ "move", 6 },
			{ "close", 7 },
			{ "pageup", 8 },
			{ "pagedown", 9 },
			{ "execute", 10 },
			{ 0, 0}
		};

		const NameValuePair hiddenstateValues[] = {
			{ ".none", 0 },
			{ ".folders", 1 },
			{ ".files", 2 },
			{ ".all", 3 },
			{ "none", 0 },
			{ "folders", 1 },
			{ "files", 2 },
			{ "all", 3 },
			{ 0, 0}
		};

	GetxRCList( "*xPopupCustomSubFolder", customSubFolderList );

	//---------------------------
	//Get DefaultIcons
	//---------------------------
	bShowIcon = Getx2RCBool(strName.c_str(), "ShowIcon", true);
	bShowText = Getx2RCBool(strName.c_str(), "ShowText", true);

	DefaultIcon = iGetx2RCString(strName, "DefaultIcon", "");		
	DefaultFolderIcon = iGetx2RCString(strName, "FolderIcon", "");
	
	strPinnedIconPath = iGetx2RCString(strName, "PinnedIconPath", "");
	
	strArrowText = iGetx2RCString(strName, "ArrowText", "");
	strSelArrowText = iGetx2RCString(strName, "ActiveArrowText", strArrowText);
	iArrowSpacing = Getx2RCInt(strName.c_str(), "ArrowOffset", 4);
	bArrowOnLeft = Getx2RCBool(strName.c_str(), "ArrowOnLeft");
	iSelArrowSpacing = Getx2RCInt(strName.c_str(), "ActiveArrowOffset", iArrowSpacing);
	bSelArrowOnLeft = Getx2RCBool(strName.c_str(), "ActiveArrowOnLeft", bArrowOnLeft);

	int iCrucialIconSize = Getx2RCInt(strName.c_str(), "LargeIconChangeAt", 32, 17);

	bBigEntryIcon = (pPaintClasses->pEntryIcon->getIconSize() >= iCrucialIconSize ? true : false);
	bBigActiveEntryIcon = (pPaintClasses->pActiveEntryIcon->getIconSize() >= iCrucialIconSize ? true : false);
	bBigFolderIcon = (pPaintClasses->pFolderIcon->getIconSize() >= iCrucialIconSize ? true : false);
	bBigActiveFolderIcon = (pPaintClasses->pActiveFolderIcon->getIconSize() >= iCrucialIconSize ? true : false);
	
	bBigInfoIcon = (pPaintClasses->pInfoIcon->getIconSize() >= iCrucialIconSize ? true : false);
	bBigActiveInfoIcon = (pPaintClasses->pActiveInfoIcon->getIconSize() >= iCrucialIconSize ? true : false);

	if (pPaintClasses->pTitleIcon)
		bBigTitleIcon = (pPaintClasses->pTitleIcon->getIconSize() >= iCrucialIconSize ? true : false);
	else 
		bBigTitleIcon = false;

	if (pPaintClasses->pPinnedIcon)
		bBigPinnedIcon = (pPaintClasses->pPinnedIcon->getIconSize() >= iCrucialIconSize ? true : false);
	else 
		bBigPinnedIcon = false;

	iCacheLevel = Getx2RCInt(strName.c_str(), "CacheLevel", 0, 0, 2)-1;

	//Use Split Helper Strings
	numTokens = SplitString(iGetx2RCLine(strName.c_str(), "SnapTo", "10 all"), one, two, three, four, five, six, seven);

	if ( numTokens >= 1 )
	{
		iSnapDistance = atoi( one.c_str() );

		if ( iSnapDistance == 0 )
		{
			iSnapTo = 0;
		}
		else
		{
			if (numTokens == 1)
				iSnapTo = 1;
			else if (numTokens >= 2)
			{
				if (_stricmp(two.c_str(), "all") == NULL)
					iSnapTo = 1;
				else if (_stricmp(two.c_str(), "popups") == NULL)
					iSnapTo = 2;
				else if (_stricmp(two.c_str(), "desktop") == NULL)
					iSnapTo = 3;
			}
		}
	}

	customWorkArea.left = iScreenX;
	customWorkArea.right = iScreenWidth;
	customWorkArea.top = iScreenY;
	customWorkArea.bottom = iScreenHeight;

	string workArea = iGetx2RCLine(strName, "WorkArea", "");
	if ( !workArea.empty() )
	{
		numTokens = SplitString(workArea, one, two, three, four, five, six, seven);

		if ( numTokens == 4 )
		{
			customWorkArea.left = xParseCoord(one.c_str(), customWorkArea.left, iScreenWidth, false);
			customWorkArea.right = xParseCoord(two.c_str(), customWorkArea.right, iScreenWidth, false);
			customWorkArea.top = xParseCoord(three.c_str(), customWorkArea.top, iScreenHeight, true);
			customWorkArea.bottom = xParseCoord(four.c_str(), customWorkArea.bottom, iScreenHeight, true);
		}
	}

	restrictToWorkArea = Getx2RCBool(strName.c_str(), "RestrictToWorkArea");

	showCaption = Getx2RCNamedValue(strName.c_str(), "ShowCaption", captionPositionValues, 1);
	showExtensions = Getx2RCBool(strName.c_str(), "ShowExtensions");
	if ( showExtensions )
		noshowExtensions = split(iGetx2RCString(strName, "ShowExtensions", ""), "|");

	bPinnedPopupNotOnTop = Getx2RCBool(strName.c_str(), "PinnedPopupNotOnTop");
	HideAllowed = Getx2RCBool(strName.c_str(), "ManualUnPin", true);
	MoveAllowed = Getx2RCBool(strName.c_str(), "Moveable", true);
	ShadeAllowed = Getx2RCBool(strName.c_str(), "Shadeable", true);
	AutoPin = Getx2RCBool(strName.c_str(), "PinOnMove", true);

	folderOpenOnClick = Getx2RCBool(strName.c_str(), "FolderOpenOnClick");

	dragndropEnabled = Getx2RCBool(strName.c_str(), "EnableDragnDrop");

	iShowHiddenFiles = Getx2RCNamedValue(strName.c_str(), "IgnoreHiddenAttribute", hiddenstateValues, 0);

	bOpenToTheLeft = Getx2RCBool(strName.c_str(), "OpenToLeft");

	noLoadingPopup = Getx2RCBool(strName.c_str(), "NoLoadingPopup");

	iTooltipMode = Getx2RCNamedValue(strName.c_str(), "TooltipMode", tooltipValues, 0);

	if ( pParentPopup == NULL )
		iQuicklaunchMode = Getx2RCNamedValue(strName.c_str(), "QuicklaunchMode", quicklaunchValues, 0);
	else
		iQuicklaunchMode = pParentPopup->iQuicklaunchMode;

	quicklaunchItemSize = Getx2RCInt(strName.c_str(), "QuicklaunchItemSize", 32, 16, 64);
	quicklaunchWrapCount = Getx2RCInt(strName.c_str(), "QuicklaunchWrapCount", 100, 1);
	quicklaunchOpenTo = Getx2RCNamedValue(strName.c_str(), "QuicklaunchOpenTo", captionPositionValues, 1);

	openSound = iGetx2RCString(strName, "OpenSound", "");
	closeSound = iGetx2RCString(strName, "CloseSound", "");
	subSound = iGetx2RCString(strName, "SwitchSound", "");
	execSound = iGetx2RCString(strName, "ActionSound", "");

	dropCommand = iGetx2RCLine(strName, "OnDrop", "");
	openCommand = iGetx2RCLine(strName, "OnOpen", "");
	closeCommand = iGetx2RCLine(strName, "OnClose", "");
	shadeCommand = iGetx2RCLine(strName, "OnShade", "");
	unshadeCommand = iGetx2RCLine(strName, "OnUnShade", "");

	folderDoubleClickAction = iGetx2RCLine(strName, "OnFolderDoubleClick", "explorer.exe /e,");

	closeAfterAction = Getx2RCBool(strName.c_str(), "CloseAfterAction", true);

	titleLClickAction = Getx2RCNamedValue(strName.c_str(), "OnTitleLeftClick", clickactionValues, 6);
	if ( titleLClickAction == 10 )
	{
		CString strTemp = iGetx2RCLine(strName.c_str(), "OnTitleLeftClick", "").c_str();
		strTemp.Replace("\t", " ");
		strTemp.TrimLeft();
		strTemp.TrimRight();
		strtitleLClickAction = strTemp.Right(strTemp.GetLength() - strTemp.Find(" "));
	}
	titleMClickAction = Getx2RCNamedValue(strName.c_str(), "OnTitleMiddleClick", clickactionValues, 0);
	if ( titleMClickAction == 10 )
	{
		CString strTemp = iGetx2RCLine(strName.c_str(), "OnTitleMiddleClick", "").c_str();
		strTemp.Replace("\t", " ");
		strTemp.TrimLeft();
		strTemp.TrimRight();
		strtitleMClickAction = strTemp.Right(strTemp.GetLength() - strTemp.Find(" "));
	}
	titleRClickAction = Getx2RCNamedValue(strName.c_str(), "OnTitleRightClick", clickactionValues, 4);
	if ( titleRClickAction == 10 )
	{
		CString strTemp = iGetx2RCLine(strName.c_str(), "OnTitleRightClick", "").c_str();
		strTemp.Replace("\t", " ");
		strTemp.TrimLeft();
		strTemp.TrimRight();
		strtitleRClickAction = strTemp.Right(strTemp.GetLength() - strTemp.Find(" "));
	}

	bottomLClickAction = Getx2RCNamedValue(strName.c_str(), "OnBottomLeftClick", clickactionValues, 0);
	if ( bottomLClickAction == 10 )
	{
		CString strTemp = iGetx2RCLine(strName.c_str(), "OnBottomLeftClick", "").c_str();
		strTemp.Replace("\t", " ");
		strTemp.TrimLeft();
		strTemp.TrimRight();
		strbottomLClickAction = strTemp.Right(strTemp.GetLength() - strTemp.Find(" "));
	}
	bottomMClickAction = Getx2RCNamedValue(strName.c_str(), "OnBottomMiddleClick", clickactionValues, 0);
	if ( bottomMClickAction == 10 )
	{
		CString strTemp = iGetx2RCLine(strName.c_str(), "OnBottomMiddleClick", "").c_str();
		strTemp.Replace("\t", " ");
		strTemp.TrimLeft();
		strTemp.TrimRight();
		strbottomMClickAction = strTemp.Right(strTemp.GetLength() - strTemp.Find(" "));
	}
	bottomRClickAction = Getx2RCNamedValue(strName.c_str(), "OnBottomRightClick", clickactionValues, 0);
	if ( bottomRClickAction == 10 )
	{
		CString strTemp = iGetx2RCLine(strName.c_str(), "OnBottomRightClick", "").c_str();
		strTemp.Replace("\t", " ");
		strTemp.TrimLeft();
		strTemp.TrimRight();
		strbottomRClickAction = strTemp.Right(strTemp.GetLength() - strTemp.Find(" "));
	}

	//File Actions
	if (dropCommand.empty())
		DragAcceptFiles(hWnd, false);
	else
		DragAcceptFiles(hWnd, true);

	leftBorder = Getx2RCInt(strName.c_str(), "LeftBorder", 0, 0);
	topBorder = Getx2RCInt(strName.c_str(), "TopBorder", 0, 0);
	rightBorder = Getx2RCInt(strName.c_str(), "RightBorder", 0, 0);
	bottomBorder = Getx2RCInt(strName.c_str(), "BottomBorder", 0, 0);

	overlapX = Getx2RCInt(strName.c_str(), "OverlapX", 0);
	overlapY = Getx2RCInt(strName.c_str(), "OverlapY", -topBorder);

	entryHeight = Getx2RCInt(strName.c_str(), "EntryHeight", 18, 5);
	folderHeight = Getx2RCInt(strName.c_str(), "FolderHeight", entryHeight, 5);
	infoHeight = Getx2RCInt(strName.c_str(), "InfoHeight", entryHeight, 5);
	separatorHeight = Getx2RCInt(strName.c_str(), "SeparatorHeight", 5, 1);

	autoSeparator = Getx2RCInt(strName.c_str(), "AutoSeparator", 0, 0, 3);
	if ( autoSeparator == 0 )
	{
		if ( Getx2RCBool(strName.c_str(), "AutoSeparator") )
			autoSeparator = 1;
	}
	autoMenuBreak = Getx2RCBool(strName.c_str(), "AutoMenuBreak");

	maxWidth = Getx2RCInt(strName.c_str(), "MaxWidth", iScreenWidth, 0, (customWorkArea.right-customWorkArea.left) );
	maxHeight = Getx2RCInt(strName.c_str(), "MaxHeight", iScreenHeight, 0, (customWorkArea.bottom-customWorkArea.top) );
	minWidth = Getx2RCInt(strName.c_str(), "MinWidth", 0, 0, iScreenWidth);
	minHeight = Getx2RCInt(strName.c_str(), "MinHeight", /*max(*/topBorder+bottomBorder/*, popuptexture->getMinNeededHeight())*/, /*max(*/topBorder+bottomBorder/*, popuptexture->getMinNeededHeight())*/, iScreenHeight);
	orgminHeight = minHeight;

	maxEntryWidth = Getx2RCInt(strName.c_str(), "MaxEntryWidth", iScreenWidth, 0, iScreenWidth);
	minEntryWidth = Getx2RCInt(strName.c_str(), "MinEntryWidth", 0, 0, iScreenWidth);

	xSpacing = Getx2RCInt(strName.c_str(), "xSpacing", 0, 0);
	ySpacing = Getx2RCInt(strName.c_str(), "ySpacing", 0, 0);

	separatorySpacing = Getx2RCInt(strName.c_str(), "SeparatorySpacing", 0, -ySpacing);

	autoHide = Getx2RCBool(strName.c_str(), "AutoHide");

	bExportHoverState = Getx2RCBool(strName.c_str(), "ExportHoverState");

	hideDelay = Getx2RCInt(strName.c_str(), "HideDelay", 250, 100);
	trackDelay = Getx2RCInt(strName.c_str(), "TrackingDelay", 50, 10);
	openDelay = Getx2RCInt(strName.c_str(), "FolderOpenDelay", 50, 1);
	tasksUpdate = Getx2RCInt(strName.c_str(), "TasksUpdateInterval", 5000, 1000);
	multiscrollDelay = Getx2RCInt(strName.c_str(), "MultiPartScrollDelay", 2000, 250);

	bTrackingIgnoreTrueTransparency = Getx2RCBool(strName.c_str(), "TrackingIgnoreTrueTransparency");

	if ( IsOS(OS_2KXP) )
	{
		alphaTransparency = Getx2RCInt(strName.c_str(), "AlphaTransparency", 255, 0, 255);

		alphaMap = Getx2RCBool(strName.c_str(), "AlphaMap");

		//---------------------------------------------------------------
		//---------------------------------------------------------------
		string config = iGetx2RCLine(strName, "CustomAlphaFade", "");

		numTokens = SplitString(config, one, two, three, four, five, six, seven);
	
		if ( config.empty() )
		{
			alphaFade = false;
		}
		else if (numTokens == 1)
		{
			alphaFadeStep = atoi( one.c_str() );
			if (alphaFadeStep < 1 || alphaFadeStep > 255)
				alphaFadeStep = 2;
			alphaFade = true;
		}
		else if (numTokens == 2)
		{
			alphaFadeStep = atoi( one.c_str() );
			if (alphaFadeStep < 1 || alphaFadeStep > 255)
				alphaFadeStep = 2;
			alphaFadeDelay = atoi( two.c_str() );
			if (alphaFadeDelay < 1 || alphaFadeDelay > 500)
				alphaFadeDelay = 10;
			alphaFade = true;
		}

		int value = Getx2RCNamedValue(strName.c_str(), "AlphaFade", alphaFadeValues, -1);

		if (value == -1)
		{
			alphaFade = Getx2RCBool(strName.c_str(), "AlphaFade", alphaFade);
		}
		else if ( value == 2 )
		{
			if ( BangToCall.empty() )
				alphaFade = false;
			else
				alphaFade = true;
		}
		else if ( value == 1 )
		{
			alphaFade = true;
		}

		if ( !alphaMap && (alphaFade || alphaTransparency != 255 || bTrackingIgnoreTrueTransparency) )
			fakealphaMap = true;
		else 
			fakealphaMap = false;
		//---------------------------------------------------------------
	}
	else if ( !IsOS(OS_2KXP) )
	{
		alphaTransparency = 255;
		alphaFade = false;
		alphaMap = false;
		fakealphaMap = false;
	}

	//---------------------------------------------------------------
	//---------------------------------------------------------------
	if ( IsOS(OS_2KXP) && (alphaTransparency != 255 || alphaFade || alphaMap || fakealphaMap || bTrackingIgnoreTrueTransparency) )
	{
		ModifyStyleEx(hWnd, 0, WS_EX_LAYERED);

		if (alphaFade)
			UpdateWindowTransparent(hWnd, 0);
		else
			UpdateWindowTransparent(hWnd, alphaTransparency);
	}

	if ( iQuicklaunchMode != 0 || iTooltipMode != 0 )
	{
		if ( !pTooltip )
		{
			pTooltip = Create_xTooltipClass(hWnd, hInstance);
			hWndToolTip = pTooltip->configure(strName.c_str(), "");

			pTooltip->apply("");
		}
	}

	updateRegion();
}

//////////////////////////////////////////////////////////////////////
// Styles
//////////////////////////////////////////////////////////////////////

void Popup::setAlwaysOnTop(bool newalwaysOnTop)
{
	alwaysOnTop = newalwaysOnTop;
		
	SetWindowPos(hWnd, newalwaysOnTop ? HWND_TOPMOST : HWND_NOTOPMOST,
		0, 0, 0, 0,
		SWP_NOACTIVATE | SWP_NOMOVE | SWP_NOSIZE );
}

void Popup::OverlayButtonPainter(HDC hDC, int x, int y, int width, int height, int buttonType)
{
	if ( !plistGhostButton )
		return;

	if ( !plistGhostButton->empty() )
	{
		for(GhostButtonListIterator gblit = plistGhostButton->begin(); gblit != plistGhostButton->end(); gblit++)
		{
			if ( *gblit != NULL )
			{
				int gbx = x + xParseCoord((*gblit)->strOrgX.c_str(), 0, width, false);
				int gby = y + xParseCoord((*gblit)->strOrgY.c_str(), 0, height, true);
				int gbwidth = xParseDimen((*gblit)->strOrgWidth.c_str(), width, width, false);
				int gbheight = xParseDimen((*gblit)->strOrgHeight.c_str(), height, height, true);

				if ( buttonType == 1 && (*gblit)->pEntryTexture )
					(*gblit)->pEntryTexture->apply(hDC, gbx, gby, gbwidth, gbheight);
				else if ( buttonType == 2 && (*gblit)->pActiveEntryTexture )
					(*gblit)->pActiveEntryTexture->apply(hDC, gbx, gby, gbwidth, gbheight);
				else if ( buttonType == 3 && (*gblit)->pFolderTexture )
					(*gblit)->pFolderTexture->apply(hDC, gbx, gby, gbwidth, gbheight);
				else if ( buttonType == 4 && (*gblit)->pActiveFolderTexture )
					(*gblit)->pActiveFolderTexture->apply(hDC, gbx, gby, gbwidth, gbheight);
				else if ( buttonType == 5 && (*gblit)->pInfoTexture )
					(*gblit)->pInfoTexture->apply(hDC, gbx, gby, gbwidth, gbheight);
				else if ( buttonType == 6 && (*gblit)->pActiveInfoTexture )
					(*gblit)->pActiveInfoTexture->apply(hDC, gbx, gby, gbwidth, gbheight);
				else if ( buttonType == 10 && (*gblit)->pSeparatorTexture )
					(*gblit)->pSeparatorTexture->apply(hDC, gbx, gby, gbwidth, gbheight);
			}
			else
			{
				break;
			}
		}
	}
}

void Popup::OverlayPopupPainter(HDC hDC)
{
	if ( !plistGhostPopup )
		return;

	if ( !plistGhostPopup->empty() )
	{
		for(GhostPopupListIterator gplit = plistGhostPopup->begin(); gplit != plistGhostPopup->end(); gplit++)
		{
			if ( *gplit != NULL && (*gplit)->pPopupTexture )
			{
				int gtx = xParseCoord((*gplit)->strOrgX.c_str(), 0, actualwidth, false);
				int gty = xParseCoord((*gplit)->strOrgY.c_str(), 0, actualheight, true);
				int gtwidth = xParseDimen((*gplit)->strOrgWidth.c_str(), actualwidth, actualwidth, false);
				int gtheight = xParseDimen((*gplit)->strOrgHeight.c_str(), actualheight, actualheight, true);

				(*gplit)->pPopupTexture->apply(hDC, gtx, gty, gtwidth, gtheight );
			}
			else
			{
				break;
			}
		}
	}
}

void Popup::UpdateWindowTransparent(HWND hWnd, int factor)
{
	// First, see if we can get the API call we need. If we've tried
    // once, we don't need to try again. 
	if (!bAlphaInitialized)
	{
		HMODULE hDLL = LoadLibrary ("user32");

		pSetLayeredWindowAttributes = 
					(PSLWA) GetProcAddress(hDLL, "SetLayeredWindowAttributes");

		pUpdateLayeredWindow = 
					(PULW) GetProcAddress(hDLL, "UpdateLayeredWindow");

		bAlphaInitialized = true;

		ResetBackground();
   
		if (pSetLayeredWindowAttributes == NULL || pUpdateLayeredWindow == NULL)
		{
		    ModifyStyleEx(hWnd, WS_EX_LAYERED, 0 );

			alphaTransparency = 255;
			alphaFade = false;
			alphaMap = false;
			fakealphaMap = false;

			return;
		}
	}

	if ( bResetBackground )
	{
		bResetBackground = false;
		ResetBackground();
	}

	RECT rectPosition;
	GetWindowRect(hWnd, &rectPosition);

	int width = (rectPosition.right-rectPosition.left);
	int height = (rectPosition.bottom-rectPosition.top);
	
	if ( width <= 0 || height <= 0 || !pPaintClasses )
		return;
	
	// Now, we need to set the 'layered window attributes'. This
	// is where the alpha values get set. 
  	if (alphaMap || fakealphaMap)
	{
		//------------------------------------------------------------------
		// double buffer for flicker-free paint
		HDC bufferDC = CreateCompatibleDC(NULL);
	
		BITMAPINFO alpha;
		ZeroMemory( &alpha.bmiHeader, sizeof(BITMAPINFOHEADER) );
		alpha.bmiHeader.biWidth=width;      // Set size you need
		alpha.bmiHeader.biHeight=height;    // Set size you need
		alpha.bmiHeader.biPlanes=1;
		alpha.bmiHeader.biBitCount=32;            // Can be 8, 16, 32 bpp or other number
		alpha.bmiHeader.biSizeImage=0;
		alpha.bmiHeader.biSize=sizeof(BITMAPINFOHEADER);
		alpha.bmiHeader.biClrUsed= 0;
		alpha.bmiHeader.biClrImportant= 0;
		VOID *pvBits;
		HBITMAP bufferBitmap = CreateDIBSection( bufferDC,
												&alpha,
												DIB_RGB_COLORS,
												&pvBits,
												NULL,
												0 );

		HBITMAP oldbufferBMP = (HBITMAP) SelectObject(bufferDC, bufferBitmap);

		// Use Complete Backup for WindowsInternal Repaint (Covering Windows, ...)
		if ( alphaFade && isFading && fadeBackupBitmap != NULL && fadeBackupDC != NULL )
		{
			BitBlt(bufferDC, 0, 0, width, height, fadeBackupDC, 0, 0, SRCCOPY);
		}
		else
		//1----------------------------------------------------------------
		{
		//1----------------------------------------------------------------
				
			// make a rendition of the background
			HDC backgroundDC = CreateCompatibleDC(NULL);
			BITMAPINFO alpha;
			ZeroMemory( &alpha.bmiHeader, sizeof(BITMAPINFOHEADER) );
			alpha.bmiHeader.biWidth=width;      // Set size you need
			alpha.bmiHeader.biHeight=height;    // Set size you need
			alpha.bmiHeader.biPlanes=1;
			alpha.bmiHeader.biBitCount=32;            // Can be 8, 16, 32 bpp or other number
			alpha.bmiHeader.biSizeImage=0;
			alpha.bmiHeader.biSize=sizeof(BITMAPINFOHEADER);
			alpha.bmiHeader.biClrUsed= 0;
			alpha.bmiHeader.biClrImportant= 0;
			VOID *pvBits;

			HBITMAP backgroundBitmap = CreateDIBSection( backgroundDC,
											&alpha,
											DIB_RGB_COLORS,
											&pvBits,
											NULL,
											0 );

			HBITMAP oldbackgroundBMP = (HBITMAP) SelectObject(backgroundDC, backgroundBitmap);

			BITMAP bm;
			GetObject(backgroundBitmap, sizeof(BITMAP), &bm);

			BYTE *pPixel = (LPBYTE)bm.bmBits;

			int iSize = bm.bmWidth * bm.bmHeight;

			for (int j = 0; j < iSize; ++j)
			{
				if ( fakealphaMap && !alphaMap )
				{
					pPixel[0] = 255;
					pPixel[1] = 0;
					pPixel[2] = 255;
					pPixel[3] = 255;
				}
				else
				{
					pPixel[0] = 0;
					pPixel[1] = 0;
					pPixel[2] = 0;
					if ( bTrackingIgnoreTrueTransparency )
						pPixel[3] = 1;
					else
						pPixel[3] = 0;
				}

				pPixel+= 4;
			}
			//Normal Background Painting
			pPaintClasses->pPopupTexture->apply(backgroundDC, 0, 0, width, height);

			OverlayPopupPainter(backgroundDC);
				
			// create & keep a cached rendition of the default popupmenu
			if (defaultBitmap == NULL && defaultDC == NULL)
			{
				defaultDC = CreateCompatibleDC(NULL);

				BITMAPINFO alpha;
				ZeroMemory( &alpha.bmiHeader, sizeof(BITMAPINFOHEADER) );
				alpha.bmiHeader.biWidth=width;      // Set size you need
				alpha.bmiHeader.biHeight=height;    // Set size you need
				alpha.bmiHeader.biPlanes=1;
				alpha.bmiHeader.biBitCount=32;            // Can be 8, 16, 32 bpp or other number
				alpha.bmiHeader.biSizeImage=0;
				alpha.bmiHeader.biSize=sizeof(BITMAPINFOHEADER);
				alpha.bmiHeader.biClrUsed= 0;
				alpha.bmiHeader.biClrImportant= 0;
				VOID *pvBits;
				defaultBitmap = CreateDIBSection( defaultDC,
												&alpha,
												DIB_RGB_COLORS,
												&pvBits,
												NULL,
												0 );

				olddefaultBMP = (HBITMAP) SelectObject(defaultDC, defaultBitmap);
					
				// blt background into double buffer
				BitBlt(defaultDC, 0, 0, width, height, backgroundDC, 0, 0, SRCCOPY);

				//Only Painted if NOT Shaded
				if ( shaded == -1 )
				{
					int state = 0;
					int arrowstate = 0;
					for (EntryListIterator eit = entrylist.begin(); eit != entrylist.end(); ++eit)
					{
						if ( iCurrentPart == (*eit)->hiddenPart )
						{
							//initialize
							state = 0;
							arrowstate = 0;

							if ( (*eit)->entryType == 0 )
							{
								state = 1;
								pPaintClasses->pEntryTexture->apply( defaultDC, (*eit)->x, (*eit)->y, (*eit)->width, (*eit)->height);
								OverlayButtonPainter( defaultDC, (*eit)->x, (*eit)->y, (*eit)->width, (*eit)->height, state );
					
								if (bShowIcon)
								{
									HICON hIcon = RetrieveIcon((*eit)->iconPath, (*eit)->hWnd, (*eit)->m_pidl, DefaultIcon, bBigEntryIcon);
									pPaintClasses->pEntryIcon->apply( defaultDC, (*eit)->x, (*eit)->y, (*eit)->width, (*eit)->height, hIcon);
									DestroyIcon(hIcon);
								}
								if (iQuicklaunchMode <= 0 && bShowText)
									pPaintClasses->pEntryText->apply( defaultDC, (*eit)->x, (*eit)->y, (*eit)->width, (*eit)->height, (*eit)->caption.c_str());
							}
							else if ( (*eit)->entryType < 0 )
							{
								state = 3;
								arrowstate = 31;
								pPaintClasses->pFolderTexture->apply( defaultDC, (*eit)->x, (*eit)->y, (*eit)->width, (*eit)->height);
								OverlayButtonPainter( defaultDC, (*eit)->x, (*eit)->y, (*eit)->width, (*eit)->height, state );
					
								if (bShowIcon)
								{
									HICON hIcon = RetrieveIcon((*eit)->iconPath, (*eit)->hWnd, (*eit)->m_pidl, DefaultFolderIcon, bBigFolderIcon);
									pPaintClasses->pFolderIcon->apply( defaultDC, (*eit)->x, (*eit)->y, (*eit)->width, (*eit)->height, hIcon);
									DestroyIcon(hIcon);
								}
								if (iQuicklaunchMode <= 0 && bShowText)
									pPaintClasses->pFolderText->apply( defaultDC, (*eit)->x, (*eit)->y, (*eit)->width, (*eit)->height, (*eit)->caption.c_str());
							}
							else if ( (*eit)->entryType == 1 )
							{
								state = 10;
								pPaintClasses->pSeparatorTexture->apply( defaultDC, (*eit)->x, (*eit)->y, (*eit)->width, (*eit)->height);
								OverlayButtonPainter( defaultDC, (*eit)->x, (*eit)->y, (*eit)->width, (*eit)->height, state );
							}
							else
							{
								continue;
							}
						
							//Paint Arrow
							if (arrowstate != 0)
							{
								POINT arrowSize = pPaintClasses->pArrowTexture->getImageSize();
								pPaintClasses->pArrowTexture->apply( defaultDC, (*eit)->x+(bArrowOnLeft ? iArrowSpacing : (*eit)->width-arrowSize.x-iArrowSpacing), (*eit)->y+((*eit)->height-arrowSize.y)/2, arrowSize.x, arrowSize.y);	
								
								if ( pPaintClasses->pArrowText )
									pPaintClasses->pArrowText->apply( defaultDC, (*eit)->x, (*eit)->y, (*eit)->width, (*eit)->height, strArrowText.c_str());
							}
						}
					}
				}
			}

			// blt default background into double buffer
			BitBlt(bufferDC, 0, 0, width, height, defaultDC, 0, 0, SRCCOPY);

			CString strHelpCaption = captionname.c_str();
			strHelpCaption.Replace("&&", "..--..");
			strHelpCaption.Replace("&", "");
			strHelpCaption.Replace("..--..", "&&");

			if ( iQuicklaunchMode == 0 )
			{
				if ( topBorder > 0 )
				{
					pPaintClasses->pTitleTexture->apply( bufferDC, 0, 0, width, topBorder );

					OverlayButtonPainter( bufferDC, 0, 0, width, topBorder, -1 );
				}
				if ( bottomBorder > 0 )
				{
					pPaintClasses->pBottomTexture->apply( bufferDC, 0, height-bottomBorder, width, bottomBorder );
					
					OverlayButtonPainter( bufferDC, 0, height-bottomBorder, width, bottomBorder, -2 );
				}
				if (showCaption == 1)
				{
					if ( pParentPopup != NULL && pParentEntry != NULL && pPaintClasses->pTitleIcon )
					{
						HICON hIcon = RetrieveIcon(pParentEntry->iconPath, pParentEntry->hWnd, pParentEntry->m_pidl, "", bBigTitleIcon);
						pPaintClasses->pTitleIcon->apply( bufferDC, 0, 0, width, topBorder, hIcon );
						DestroyIcon(hIcon);
					}
					if ( bPinnedPopupMode && pPaintClasses->pPinnedIcon )
					{		
						HICON hIcon = RetrieveIcon("", NULL, NULL, strPinnedIconPath, bBigPinnedIcon);
						pPaintClasses->pPinnedIcon->apply( bufferDC, 0, 0, width, topBorder, hIcon );
						DestroyIcon(hIcon);
					}

					pPaintClasses->pTitleText->apply( bufferDC, 0, 0, width, topBorder, strHelpCaption);
				}
				else if (showCaption == 2)
				{
					if ( pParentPopup != NULL && pParentEntry != NULL && pPaintClasses->pTitleIcon )
					{
						HICON hIcon = RetrieveIcon(pParentEntry->iconPath, pParentEntry->hWnd, pParentEntry->m_pidl, "", bBigTitleIcon);
						pPaintClasses->pTitleIcon->apply( bufferDC, 0, height-bottomBorder, width, bottomBorder, hIcon );
						DestroyIcon(hIcon);
					}
					if ( bPinnedPopupMode && pPaintClasses->pPinnedIcon )
					{
						HICON hIcon = RetrieveIcon("", NULL, NULL, strPinnedIconPath, bBigPinnedIcon);
						pPaintClasses->pPinnedIcon->apply( bufferDC, 0, height-bottomBorder, width, bottomBorder, hIcon );
						DestroyIcon(hIcon);
					}
					pPaintClasses->pTitleText->apply( bufferDC, 0, height-bottomBorder, width, bottomBorder, strHelpCaption);
				}
				else if (showCaption == 3)
				{
					if ( pParentPopup != NULL && pParentEntry != NULL && pPaintClasses->pTitleIcon )
					{
						HICON hIcon = RetrieveIcon(pParentEntry->iconPath, pParentEntry->hWnd, pParentEntry->m_pidl, "", bBigTitleIcon);
						pPaintClasses->pTitleIcon->apply( bufferDC, 0, 0, width, height, hIcon );
						DestroyIcon(hIcon);
					}
					if ( bPinnedPopupMode && pPaintClasses->pPinnedIcon )
					{
						HICON hIcon = RetrieveIcon("", NULL, NULL, strPinnedIconPath, bBigPinnedIcon);
						pPaintClasses->pPinnedIcon->apply( bufferDC, 0, 0, width, height, hIcon );
						DestroyIcon(hIcon);
					}
					pPaintClasses->pTitleText->apply( bufferDC, 0, 0, width, height, strHelpCaption);
				}
			}
			//Only Painted if NOT Shaded
			if ( shaded == -1 )
			{
				int state = 0;
				int arrowstate = 0;
				if ( iQuicklaunchMode == 0 && hasInfoEntry )
				{
					for (EntryListIterator eit = entrylist.begin(); eit != entrylist.end(); ++eit)
					{
						if ( iCurrentPart == (*eit)->hiddenPart && (*eit) != hoverentry && (*eit)->entryType == 2 )
						{
							state = 5;

							//Normal Background Painting behind ActiveEntry
							BitBlt(bufferDC, (*eit)->x, (*eit)->y, (*eit)->width, (*eit)->height, backgroundDC, (*eit)->x, (*eit)->y, SRCCOPY);

							//Paint Entry Texture
							pPaintClasses->pInfoTexture->apply( bufferDC, (*eit)->x, (*eit)->y, (*eit)->width, (*eit)->height);
							
							OverlayButtonPainter( bufferDC, (*eit)->x, (*eit)->y, (*eit)->width, (*eit)->height, state );
							
							HICON hIcon = RetrieveIcon((*eit)->iconPath, NULL, NULL, "", bBigInfoIcon);
							pPaintClasses->pInfoIcon->apply( bufferDC, (*eit)->x, (*eit)->y, (*eit)->width, (*eit)->height, hIcon);
							DestroyIcon(hIcon);
							
							pPaintClasses->pInfoText->apply( bufferDC, (*eit)->x, (*eit)->y, (*eit)->width, (*eit)->height, (*eit)->caption.c_str());
						}
						else if ( iCurrentPart == (*eit)->hiddenPart && (*eit) == hoverentry && (*eit)->entryType == 2 )
						{
							state = 6;

							//Normal Background Painting behind ActiveEntry
							BitBlt(bufferDC, (*eit)->x, (*eit)->y, (*eit)->width, (*eit)->height, backgroundDC, (*eit)->x, (*eit)->y, SRCCOPY);

							//Paint Entry Texture
							pPaintClasses->pActiveInfoTexture->apply( bufferDC, (*eit)->x, (*eit)->y, (*eit)->width, (*eit)->height);
						
							OverlayButtonPainter( bufferDC, (*eit)->x, (*eit)->y, (*eit)->width, (*eit)->height, state );
							
							HICON hIcon = RetrieveIcon((*eit)->iconPath, NULL, NULL, "", bBigActiveInfoIcon);
							pPaintClasses->pActiveInfoIcon->apply( bufferDC, (*eit)->x, (*eit)->y, (*eit)->width, (*eit)->height, hIcon);
							DestroyIcon(hIcon);
												
							pPaintClasses->pActiveInfoText->apply( bufferDC, (*eit)->x, (*eit)->y, (*eit)->width, (*eit)->height, (*eit)->caption.c_str());
						}
					}
				}
				for (EntryListIterator eit = entrylist.begin(); eit != entrylist.end(); ++eit)
				{
					//initialize
					state = 0;
					arrowstate = 0;
					
					if ( iCurrentPart == (*eit)->hiddenPart && (*eit) == hoverentry && (*eit) != NULL)
					{
						if ( (*eit)->entryType == 0 )
						{
							//Normal Background Painting behind ActiveEntry
							BitBlt(bufferDC, (*eit)->x, (*eit)->y, (*eit)->width, (*eit)->height, backgroundDC, (*eit)->x, (*eit)->y, SRCCOPY);

							state = 2;
							pPaintClasses->pActiveEntryTexture->apply( bufferDC, (*eit)->x, (*eit)->y, (*eit)->width, (*eit)->height);
							OverlayButtonPainter( bufferDC, (*eit)->x, (*eit)->y, (*eit)->width, (*eit)->height, state );
							//If No Info, Icon is valid
							if (bShowIcon)
							{
								HICON hIcon = RetrieveIcon((*eit)->iconPath, (*eit)->hWnd, (*eit)->m_pidl, DefaultIcon, bBigActiveEntryIcon);
								pPaintClasses->pActiveEntryIcon->apply( bufferDC, (*eit)->x, (*eit)->y, (*eit)->width, (*eit)->height, hIcon);
								DestroyIcon(hIcon);
							}
							if (iQuicklaunchMode <= 0 && bShowText)
								pPaintClasses->pActiveEntryText->apply( bufferDC, (*eit)->x, (*eit)->y, (*eit)->width, (*eit)->height, (*eit)->caption.c_str());
						}
						else if ( (*eit)->entryType < 0 )
						{
							//Normal Background Painting behind ActiveEntry
							BitBlt(bufferDC, (*eit)->x, (*eit)->y, (*eit)->width, (*eit)->height, backgroundDC, (*eit)->x, (*eit)->y, SRCCOPY);

							state = 4;
							arrowstate = 41;
							pPaintClasses->pActiveFolderTexture->apply( bufferDC, (*eit)->x, (*eit)->y, (*eit)->width, (*eit)->height);
							OverlayButtonPainter( bufferDC, (*eit)->x, (*eit)->y, (*eit)->width, (*eit)->height, state );
							//If No Info, Icon is valid
							if (bShowIcon)
							{
								HICON hIcon = RetrieveIcon((*eit)->iconPath, (*eit)->hWnd, (*eit)->m_pidl, DefaultFolderIcon,bBigActiveFolderIcon);
								pPaintClasses->pActiveFolderIcon->apply( bufferDC, (*eit)->x, (*eit)->y, (*eit)->width, (*eit)->height, hIcon);
								DestroyIcon(hIcon);
							}
							if (iQuicklaunchMode <= 0 && bShowText)	
								pPaintClasses->pActiveFolderText->apply( bufferDC, (*eit)->x, (*eit)->y, (*eit)->width, (*eit)->height, (*eit)->caption.c_str());
						}
						else
						{
							continue;
						}
					
						//Paint Arrow
						if (arrowstate != 0)
						{
							POINT arrowSize = pPaintClasses->pActiveArrowTexture->getImageSize();
							pPaintClasses->pActiveArrowTexture->apply( bufferDC, (*eit)->x+(bSelArrowOnLeft ? iSelArrowSpacing : (*eit)->width-arrowSize.x-iSelArrowSpacing), (*eit)->y+((*eit)->height-arrowSize.y)/2, arrowSize.x, arrowSize.y);
							
							if ( pPaintClasses->pActiveArrowText )
								pPaintClasses->pActiveArrowText->apply( bufferDC, (*eit)->x, (*eit)->y, (*eit)->width, (*eit)->height, strSelArrowText.c_str());
						}			
					}
				}
			}

			//Prepare Combined Image
			BITMAP bm0;
			GetObject(bufferBitmap, sizeof(BITMAP), &bm0);

			if ((bm0.bmBits != 0) && (bm0.bmBitsPixel == 32))
			{
				BYTE *pPixel = (LPBYTE)bm0.bmBits;

				int iSize = bm0.bmHeight * bm0.bmWidth;

				for (int j = 0; j < iSize; ++j)
				{
					//Remove Magic Pink
					if (pPixel[0] == 255 && pPixel[1] == 0 && pPixel[2] == 255)
					{
						pPixel[0] = 0;
						pPixel[1] = 0;
						pPixel[2] = 0;
						if ( bTrackingIgnoreTrueTransparency )
							pPixel[3] = 1;
						else
							pPixel[3] = 0;
					}
					//Make the Rest Opaque, if FakeAlphaMap
					else if ( fakealphaMap )
					{
						pPixel[3] = 255;
					}

					pPixel+= 4;
				}
			}
			//------------------------------------------------------------------

			//Create fadeBackupBitmap, if it doesn't exist!
			//----------------------------------------------------------------------
			if ( alphaFade && isFading && fadeBackupBitmap == NULL && fadeBackupDC == NULL )
			{
				fadeBackupDC = CreateCompatibleDC(NULL);
		
				BITMAPINFO alpha;
				ZeroMemory( &alpha.bmiHeader, sizeof(BITMAPINFOHEADER) );
				alpha.bmiHeader.biWidth=width;      // Set size you need
				alpha.bmiHeader.biHeight=height;    // Set size you need
				alpha.bmiHeader.biPlanes=1;
				alpha.bmiHeader.biBitCount=32;            // Can be 8, 16, 32 bpp or other number
				alpha.bmiHeader.biSizeImage=0;
				alpha.bmiHeader.biSize=sizeof(BITMAPINFOHEADER);
				alpha.bmiHeader.biClrUsed= 0;
				alpha.bmiHeader.biClrImportant= 0;
				VOID *pvBits;

				fadeBackupBitmap = CreateDIBSection( fadeBackupDC,
														&alpha,
														DIB_RGB_COLORS,
														&pvBits,
														NULL,
														0 );

				oldfadeBackupBMP = (HBITMAP) SelectObject(fadeBackupDC, fadeBackupBitmap);

				BitBlt(fadeBackupDC, 0, 0, width, height, bufferDC, 0, 0, SRCCOPY);
			}
			//----------------------------------------------------------------------

			backgroundBitmap = (HBITMAP) SelectObject(backgroundDC, oldbackgroundBMP);
			DeleteObject(backgroundBitmap);
			DeleteDC(backgroundDC);	
			DeleteObject(oldbackgroundBMP);
		//1
		}
		//1}
	
		POINT ptWindowScreenPosition = {rectPosition.left, rectPosition.top};
		SIZE szWindow = {width, height};

		POINT ptSrc = {0,0};

		BLENDFUNCTION blendPixelFunction = { AC_SRC_OVER, 0, factor, AC_SRC_ALPHA };

		pUpdateLayeredWindow( hWnd, NULL, &ptWindowScreenPosition, &szWindow, bufferDC, &ptSrc, 0, &blendPixelFunction, ULW_ALPHA );
		
		bufferBitmap = (HBITMAP) SelectObject(bufferDC, oldbufferBMP);
		DeleteObject(bufferBitmap);
		DeleteDC(bufferDC);	
		DeleteObject(oldbufferBMP);
	}
	else
	{
		pSetLayeredWindowAttributes( hWnd, 0, factor, LWA_ALPHA );
	}
}

//////////////////////////////////////////////////////////////////////
// Popup Creation -> StepParsing, Folder Scanning, Shell Scanning, ...
//////////////////////////////////////////////////////////////////////

void Popup::ScanKeyCodes(ENTRY* entryProcessedEntry, string strProcessedString, bool bIgnoreAmpersand)
{
	if ( !entryProcessedEntry || strProcessedString.empty() )
		return;

	//Ampersand Keyboard navigation
	//------------------------------------------------------------
	char* m_pszTitle = strdup(strProcessedString.c_str());

	if ( !bIgnoreAmpersand )
	{
		// search the string for '&', that is keyboard-shortcuts
		int len = lstrlen(m_pszTitle);
		for (int i = 0; i < len; i++)
		{
			if (m_pszTitle[i] == '&' && m_pszTitle[i + 1] != '\0' && m_pszTitle[i + 1] != ' ' && m_pszTitle[i + 1] != '&')
			{
				entryProcessedEntry->keyCode = islower(m_pszTitle[i + 1]) ? toupper(m_pszTitle[i + 1]) : m_pszTitle[i + 1];
				break;
			}
			else if (m_pszTitle[i + 1] != '&')
			{
				i++;
			}
		}
	}
	
	if (entryProcessedEntry->keyCode == -1)
	{
		entryProcessedEntry->keyCode = islower(m_pszTitle[0]) ? toupper(m_pszTitle[0]) : m_pszTitle[0];
		entryProcessedEntry->keyCode1 = islower(m_pszTitle[1]) ? toupper(m_pszTitle[1]) : m_pszTitle[1];
	}
	//------------------------------------------------------------
}

bool Popup::ParseDynamicEvars(string strInput, string &strOutput)
{
	char szExpandedEvar[MAX_LINE_LENGTH] = { 0 };

	CString worker0 = strInput.c_str();
	CString worker1 = "";

	int	pos = -1;

	worker0.Replace("%#", "$");

	pos = worker0.Find("$");
	pos = worker0.Find("$", pos+1);

	if ( pos != -1 )
	{
		while ( pos != -1 )
		{
			worker1 = worker0;
			worker1 = worker1.Right(worker1.GetLength()-worker1.Find("$"));
			worker1 = worker1.Left(worker1.Find("$", 1)+1);

			VarExpansion( szExpandedEvar, worker1 );

			worker0.Replace(worker1, szExpandedEvar);

			pos = worker0.Find("$");
			pos = worker0.Find("$", pos+1);
		}

		strOutput = (string) worker0;

		return true;
	}
	else
	{ 
		strOutput = strInput;

		return false;
	}
}

int Popup::ComputeEntryWidth(const string &measureText, bool bCheckTitle, bool bCheckInfo)
{
	int iMaxWidth = 0;

	HDC tmpDC = CreateCompatibleDC(NULL);

	if (!bCheckTitle && !bCheckInfo)
	{
		iMaxWidth = (pPaintClasses->pEntryText->measure(tmpDC, 0, 0, measureText.c_str(), true)).x;	
		iMaxWidth = max((pPaintClasses->pActiveEntryText->measure(tmpDC, 0, 0, measureText.c_str(), true)).x, iMaxWidth);
		iMaxWidth = max((pPaintClasses->pFolderText->measure(tmpDC, 0, 0, measureText.c_str(), true)).x, iMaxWidth);
		iMaxWidth = max((pPaintClasses->pActiveFolderText->measure(tmpDC, 0, 0, measureText.c_str(), true)).x, iMaxWidth);
	}
	else if (bCheckTitle)
	{
		iMaxWidth = (pPaintClasses->pTitleText->measure(tmpDC, 0, 0, measureText.c_str(), true)).x;	
	}
	else if (bCheckInfo)
	{
		iMaxWidth = (pPaintClasses->pInfoText->measure(tmpDC, 0, 0, measureText.c_str(), true)).x;	
		iMaxWidth = max((pPaintClasses->pActiveInfoText->measure(tmpDC, 0, 0, measureText.c_str(), true)).x, iMaxWidth);
	}

	DeleteDC(tmpDC);

	return iMaxWidth;
}

void Popup::ParseStepRC(bool externCall)
{
	if ( hasDynamicContent )
	{	
		string str_singleCheckString = "";
		string str_completeCheckString = "";

		for (StringListIterator sit = popupcontentList.begin(); sit != popupcontentList.end(); ++sit)
		{
			ParseDynamicEvars((*sit), str_singleCheckString);
		
			str_completeCheckString = (str_completeCheckString + str_singleCheckString);
		}
		
		//Check if the entryCheckString has changed (only reason for real updateing)
		if ( strcmp(str_completeCheckStringBackup.c_str(), str_completeCheckString.c_str()) != NULL )
		{		
			for (EntryListIterator eit = entrylist.begin(); eit != entrylist.end(); ++eit)
			{
				delete *eit;
				*eit = NULL;
			}

			entrylist.clear();

			for(PopupListIterator plit = subpopupList.begin(); plit != subpopupList.end(); plit++)
			{
				(*plit)->hide(true/*noFade*/);
				delete *plit;
				*plit = NULL;
			}

			subpopupList.clear();

			//Reset old Pointers
			//-------------------------
			iCurrentPart = 0;

			startentry = NULL;
			lastactiveentry = NULL;
			hoverentry = NULL;
			//-------------------------
			
			str_completeCheckStringBackup = str_completeCheckString;
		}
		else
		{
			if ((isTaskPopup || hasDynamicContent) && !hasInfoEntry)
				SetTimer(hWnd, TIMER_UPDATE, tasksUpdate, NULL);
			else if (hasInfoEntry)
				SetTimer(hWnd, TIMER_UPDATE, 1000, NULL);

			return;
		}
	}
	
	StringList hookcontentList;
	StringList settingsList = popupcontentList;

	if ( !settingsList.empty() && settingsList.size() == 1 )
	{
		char szFirst[MAX_LINE_LENGTH] = { 0 };
		char szSecond[MAX_LINE_LENGTH] = { 0 };
		char szThird[MAX_LINE_LENGTH] = { 0 };

		char *buffer[] = { szFirst, szSecond, szThird };

		int nTokens = LCTokenize((*settingsList.begin()).c_str(), buffer, 2, szThird);

		CString str_third = TrimAndStripQuote((string)szThird).c_str();

		if ( _stricmp(str_third.Left(19), "!popupdynamicfolder") == NULL || _stricmp(str_third.Left(14), "!dynamicfolder") == NULL || _stricmp(str_third.Left(12), "!popupfolder") == NULL )
		{
			CString easyHelper = TrimAndStripQuote( (string)szThird ).c_str();

			// Get FolderPath
			easyHelper = easyHelper.Right(easyHelper.GetLength()-easyHelper.Find(":")-1);

			// Strip surrounding '"'
			easyHelper = TrimAndStripQuote( (string)easyHelper ).c_str();

			rootScanPath = (string)easyHelper;

			if ( _stricmp(str_third.Left(19), "!popupdynamicfolder") == NULL || _stricmp(str_third.Left(14), "!dynamicfolder") == NULL )
			{
				ScanSubMenu( rootScanPath, -2 );

				bool newIteration = true;
				CString localworker1 = easyHelper;
				CString localworker2 = "";

				while (newIteration == true)
				{
					if (localworker1.Find("|") != -1)
					{
						CNotifyDirCheck *checkDirectory = new CNotifyDirCheck();
				
						checkdirectorylist.insert(checkdirectorylist.end(), checkDirectory);
					
						localworker2 = localworker1.Left(localworker1.Find("|"));

						if ( localworker2.Find("*") == -1 )
							checkDirectory->SetDirectory( localworker2 );
						else
							checkDirectory->SetDirectory( localworker2.Left(localworker2.Find("*")) );

						//set your callback to work with each new event
						checkDirectory->SetPopup( this );

						checkDirectory->Run();

						localworker1 = localworker1.Right(localworker1.GetLength()-localworker1.Find("|")-1);
					}
					else
					{
						CNotifyDirCheck *checkDirectory = new CNotifyDirCheck();
				
						checkdirectorylist.insert(checkdirectorylist.end(), checkDirectory);
					
						if ( localworker1.Find("*") == -1 )
							checkDirectory->SetDirectory( localworker1 );
						else
							checkDirectory->SetDirectory( localworker1.Left(localworker1.Find("*")) );

						//set your callback to work with each new event
						checkDirectory->SetPopup( this );

						checkDirectory->Run();

						newIteration = false;
					}
				}
			}
			else
			{
				ScanSubMenu( rootScanPath, -1 );
			}

			return;
		}
		else if ( _stricmp(str_third.Left(25), "!popupdynamicactionfolder") == NULL || _stricmp(str_third.Left(18), "!popupactionfolder") == NULL )
		{
			CString easyHelper = szThird;

			// Strip !action between ":" ... ":"
			int pos = easyHelper.Find(":");
			if ( easyHelper.Find(":", pos+3) == -1 )
			{
				return;
			}

			//Find 2nd ":" from backwards
			int sep2 = easyHelper.ReverseFind(':');
			sep2 = easyHelper.Find(":", sep2-3); // ...:c:\path

			// Get universalAction
			CString universalAction = easyHelper.Mid(pos+1, sep2-pos-1);

			easyHelper = easyHelper.Right(easyHelper.GetLength()-sep2-1);

			// Strip surrounding '"'
			easyHelper = TrimAndStripQuote( (string)easyHelper ).c_str();

			rootScanPath = (string)easyHelper;
			rootUniversalAction = (string)universalAction;

			if ( _stricmp(str_third.Left(25), "!popupdynamicactionfolder") == NULL )
			{
				ScanSubMenu( rootScanPath, -2, rootUniversalAction );

				bool newIteration = true;
				CString localworker1 = easyHelper;
				CString localworker2 = "";

				while (newIteration == true)
				{
					if (localworker1.Find("|") != -1)
					{
						CNotifyDirCheck *checkDirectory = new CNotifyDirCheck();
				
						checkdirectorylist.insert(checkdirectorylist.end(), checkDirectory);
					
						localworker2 = localworker1.Left(localworker1.Find("|"));

						if ( localworker2.Find("*") == -1 )
							checkDirectory->SetDirectory( localworker2 );
						else
							checkDirectory->SetDirectory( localworker2.Left(localworker2.Find("*")) );

						//set your callback to work with each new event
						checkDirectory->SetPopup( this );

						checkDirectory->Run();

						localworker1 = localworker1.Right(localworker1.GetLength()-localworker1.Find("|")-1);
					}
					else
					{
						CNotifyDirCheck *checkDirectory = new CNotifyDirCheck();
				
						checkdirectorylist.insert(checkdirectorylist.end(), checkDirectory);
					
						if ( localworker1.Find("*") == -1 )
							checkDirectory->SetDirectory( localworker1 );
						else
							checkDirectory->SetDirectory( localworker1.Left(localworker1.Find("*")) );

						//set your callback to work with each new event
						checkDirectory->SetPopup( this );

						checkDirectory->Run();

						newIteration = false;
					}
				}
			}
			else
			{
				ScanSubMenu( rootScanPath, -1, rootUniversalAction );
			}

			return;
		}
		else if ( _stricmp(str_third, "!popuptasks") == NULL )
		{
			iQuicklaunchMode = 0;

			isTaskPopup = true;

			ScanTasks();

			return;
		}
		else if ( _stricmp(str_third, "!popupcontrolpanel") == NULL || _stricmp(str_third, "!popupprinters") == NULL 
						|| _stricmp(str_third, "!popuprecyclebin") == NULL || _stricmp(str_third, "!popuprecentdocuments") == NULL
						|| _stricmp(str_third, "!popupnetwork") == NULL || _stricmp(str_third, "!popupmycomputer") == NULL )
		{
			iQuicklaunchMode = 0;

			if ( _stricmp(str_third, "!popupcontrolpanel") == NULL )
				ScanShellFolder( CSIDL_CONTROLS );
			else if (_stricmp(str_third, "!popupmycomputer") == NULL )
				ScanShellFolder( CSIDL_DRIVES );
			else if (_stricmp(str_third, "!popupnetwork") == NULL )
				ScanShellFolder( CSIDL_NETWORK );
			else if (_stricmp(str_third, "!popupprinters") == NULL )
				ScanShellFolder( CSIDL_PRINTERS );
			else if (_stricmp(str_third, "!popuprecentdocuments") == NULL )
				ScanShellFolder( CSIDL_RECENT );
			else if (_stricmp(str_third, "!popuprecyclebin") == NULL )
				ScanShellFolder( CSIDL_BITBUCKET );
		
			return;
		}
	}
	
	if (!settingsList.empty())
	{
		char szFirst[MAX_LINE_LENGTH] = { 0 };
		char szSecond[MAX_LINE_LENGTH] = { 0 };
		char szThird[MAX_LINE_LENGTH] = { 0 };

		char *buffer[] = { szFirst, szSecond, szThird };

		//Initializing of Values
		int itemCounter = 0;
		int nextXPosition = leftBorder;
		int nextYPosition = topBorder;
		int hiddenPart = 0;
		int menuBreaks = 0;
		int maxNeededWidth = 0;
		int adaptedmaxHeight = 0;
		bool customIcon = false;
		int nTokens = 0;

		//For Folder ~Folder
		//--------------------------------------
		Popup* addsettingPopup = NULL;
		string addsettingCustomIcon = "";
		bool input = false;
		int subfolderopen = 0;
		StringList sendToPopup;
		//--------------------------------------

		//Helper CStrings
		CString str_first = "";
		CString str_second = "";
		CString str_third = "";

		//SubPopupLayout
		CString str_fourth = "";
		//SubPopupLayout Group
		CString str_fifth = "";

		CString easyHelper = "";
		CString folderCaption = "";

		for (StringListIterator sit = settingsList.begin(); sit != settingsList.end(); ++sit)
		{
			//Check for Dynamic Settings %#
			string str_check = "";

			if ( !hasDynamicContent )
				hasDynamicContent = ParseDynamicEvars((*sit), str_check);
			else
				ParseDynamicEvars((*sit), str_check);

			//Read Settings line ( *(x)Popup )
			//Parse the Line for %# Dynamic Evars
			nTokens = LCTokenize(str_check.c_str() , buffer, 2, szThird);
			//-----------------------------------

			//Icon
			str_first = szFirst;
			str_first.MakeLower();
			str_first.Replace(".icon=", "");
			str_first = TrimAndStripQuote( (string) str_first ).c_str();

			if ( _stricmp(str_first, ".noexist") != NULL )
			{
				str_first.Replace("/", "\\");
				str_first.Replace("\\\\", "\\");
				customIcon = true;
			}
			else
			{
				customIcon = false;
			}
		
			//Caption
			str_second = szSecond;
			//Action
			str_third = TrimAndStripQuote( (string) szThird ).c_str();

			int foundExtraConfig = str_third.Find("%%");
			if ( foundExtraConfig != -1 && _stricmp(str_second.Left(5), "!hook") != NULL )
			{
				str_fourth = str_third.Right(str_third.GetLength()-foundExtraConfig-2);
				str_fourth.Replace("%%", "");

				int foundExtraGroupConfig = str_fourth.Find("%");
				if ( foundExtraGroupConfig != -1 )
				{
					str_fifth = str_fourth.Right(str_fourth.GetLength()-foundExtraGroupConfig-1);
					str_fourth = str_fourth.Left(foundExtraGroupConfig);
				}
				
				str_third = str_third.Left(foundExtraConfig);
				str_third = TrimAndStripQuote( (string) str_third ).c_str();
			}
			else
			{
				str_fourth = "";
				str_fifth = "";
			}
		
			//-----------------------------------------------------
			// MenuBreak
			//-----------------------------------------------------
			if ( nTokens >= 2 && (_stricmp(str_second, "Menubreak") == NULL || _stricmp(str_third, "Menubreak") == NULL) )
			{
				if ( iQuicklaunchMode != 0 )
					continue;

				nextXPosition = nextXPosition+maxNeededWidth+xSpacing;

				if (adaptedmaxHeight < nextYPosition-ySpacing+bottomBorder)
					adaptedmaxHeight = nextYPosition-ySpacing+bottomBorder;

				nextYPosition = topBorder;

				menuBreaks++;
			}
			//-----------------------------------------------------
			// Manually Created SubFolder START
			//-----------------------------------------------------
			else if ( nTokens >= 2 && (_stricmp(str_second, "Folder") == NULL || _stricmp(str_third, "Folder") == NULL) )
			{
				if (input)
				{
					//Add to specific Popup Settings list
					sendToPopup.insert(sendToPopup.end(), *sit);
					subfolderopen++;
				}
				else
				{
					//Search for Popup Start with !New and Create New Popup
					Popup *popup = NULL;
					if (_stricmp(str_second, "Folder") == NULL)
					{
						if ( customSubFolderList.empty() && str_fourth.IsEmpty() )
							popup = new Popup(strName, "sub_Unknown", pPaintClasses, plistGhostPopup, plistGhostButton, subFolderLevel);
						else if ( !str_fourth.IsEmpty() )
							popup = new Popup((string)str_fourth, "sub_Unknown", NULL, NULL, NULL, subFolderLevel, (str_fifth.IsEmpty() ? strName : (string)str_fifth));
						else if ( subFolderLevel < customSubFolderList.size() )
							popup = new Popup(customSubFolderList[subFolderLevel], "sub_Unknown", NULL, NULL, NULL, subFolderLevel, strName);
						else
							popup = new Popup(*(customSubFolderList.end()-1), "sub_Unknown", NULL, NULL, NULL, subFolderLevel, strName);
					}
					if (_stricmp(str_third, "Folder") == NULL)
					{
						folderCaption = str_second;
						if ( iTooltipMode != 0 )
						{
							if ( strstr(str_second, "||") != NULL )
							{
								easyHelper = str_second.Left(str_second.Find("||"));
								easyHelper = ("sub_" + easyHelper);

								if ( customSubFolderList.empty() && str_fourth.IsEmpty() )
									popup = new Popup(strName, (string)easyHelper, pPaintClasses, plistGhostPopup, plistGhostButton, subFolderLevel);
								else if ( !str_fourth.IsEmpty() )
									popup = new Popup((string)str_fourth, (string)easyHelper, NULL, NULL, NULL, subFolderLevel, (str_fifth.IsEmpty() ? strName : (string)str_fifth));
								else if ( subFolderLevel < customSubFolderList.size() )
									popup = new Popup(customSubFolderList[subFolderLevel], (string)easyHelper, NULL, NULL, NULL, subFolderLevel, strName);
								else
									popup = new Popup(*(customSubFolderList.end()-1), (string)easyHelper, NULL, NULL, NULL, subFolderLevel, strName);
							}
							else
							{
								easyHelper = ("sub_" + str_second);

								if ( customSubFolderList.empty() && str_fourth.IsEmpty() )
									popup = new Popup(strName, (string)easyHelper, pPaintClasses, plistGhostPopup, plistGhostButton, subFolderLevel);
								else if ( !str_fourth.IsEmpty() )
									popup = new Popup((string)str_fourth, (string)easyHelper, NULL, NULL, NULL, subFolderLevel, (str_fifth.IsEmpty() ? strName : (string)str_fifth));
								else if ( subFolderLevel < customSubFolderList.size() )
									popup = new Popup(customSubFolderList[subFolderLevel], (string)easyHelper, NULL, NULL, NULL, subFolderLevel, strName);
								else
									popup = new Popup(*(customSubFolderList.end()-1), (string)easyHelper, NULL, NULL, NULL, subFolderLevel, strName);
							}
						}
						else
						{
							if ( strstr(str_second, "||") != NULL )
							{
								easyHelper = str_second.Left(str_second.Find("||"));
								easyHelper = ("sub_" + easyHelper);
							}
							else
							{
								easyHelper = ("sub_" + str_second);
							}

							if ( customSubFolderList.empty() && str_fourth.IsEmpty() )
								popup = new Popup(strName, (string)easyHelper, pPaintClasses, plistGhostPopup, plistGhostButton, subFolderLevel);
							else if ( !str_fourth.IsEmpty() )
								popup = new Popup((string)str_fourth, (string)easyHelper, NULL, NULL, NULL, subFolderLevel, (str_fifth.IsEmpty() ? strName : (string)str_fifth));
							else if ( subFolderLevel < customSubFolderList.size() )
								popup = new Popup(customSubFolderList[subFolderLevel], (string)easyHelper, NULL, NULL, NULL, subFolderLevel, strName);
							else
								popup = new Popup(*(customSubFolderList.end()-1), (string)easyHelper, NULL, NULL, NULL, subFolderLevel, strName);
						}
					}
			
					if (popup)
					{
						subpopupList.insert(subpopupList.end(), popup);
								
						addsettingPopup = popup;
						addsettingCustomIcon = str_first;

						//Allow following settings to be tracked
						input = true;

						continue;
					}
				}
			}
			//-----------------------------------------------------
			// Manually Created SubFolder END and CREATION
			//-----------------------------------------------------
			else if (nTokens >= 1 && (_stricmp(str_first, "~Folder") == NULL || _stricmp(str_second, "~Folder") == NULL))
			{
				bool closeFolder = false;
				if (input)
				{
					//Add to specific Popup Settings list
					if (subfolderopen > 0)
					{
						sendToPopup.insert(sendToPopup.end(), *sit);
						subfolderopen--;
					}
					else
					{
						closeFolder = true;
					}
				}
				
				if (!input || closeFolder)
				{
					//Close Popup Definition, give Settings and Load
					if (addsettingPopup)
					{
						ENTRY* newitem = new ENTRY;
						
						//This is the parent for the Created PopupSubMenu
						addsettingPopup->pParentPopup = this;
						addsettingPopup->pParentEntry = newitem;

						addsettingPopup->load(hInstance);
						addsettingPopup->popupcontentList = sendToPopup;
						addsettingPopup->ParseStepRC();

						newitem->entryType = -1;

						if (iQuicklaunchMode != 0)
						{
							itemCounter++;

							newitem->tooltip = addsettingPopup->captionname;

							newitem->width = quicklaunchItemSize;
							newitem->height = quicklaunchItemSize;
							newitem->x = nextXPosition;
							newitem->y = nextYPosition;

							if (iQuicklaunchMode == 1 || iQuicklaunchMode == -1)
							{
								if (itemCounter%quicklaunchWrapCount != 0)
								{
									nextXPosition = nextXPosition+newitem->width+xSpacing;
									nextYPosition = nextYPosition;
								}
								else
								{
									nextXPosition = leftBorder;
									nextYPosition = nextYPosition+newitem->height+ySpacing;
								}
							}
							else if (iQuicklaunchMode == 2 || iQuicklaunchMode == -2)
							{
								if (itemCounter%quicklaunchWrapCount != 0)
								{
									nextXPosition = nextXPosition;
									nextYPosition = nextYPosition+newitem->height+ySpacing;
								}
								else
								{
									nextXPosition = nextXPosition+newitem->width+xSpacing;
									nextYPosition = topBorder;
								}
							}
						}
						else
						{
							newitem->x = nextXPosition;
							newitem->y = nextYPosition;
							newitem->height = folderHeight;

							if (nextYPosition+newitem->height+bottomBorder > maxHeight)
							{
								if (autoMenuBreak)
								{
									if (adaptedmaxHeight < nextYPosition-ySpacing+bottomBorder)
										adaptedmaxHeight = nextYPosition-ySpacing+bottomBorder;

									nextXPosition = nextXPosition+maxNeededWidth+xSpacing;
									nextYPosition = topBorder;

									newitem->x = nextXPosition;
									newitem->y = nextYPosition;

									menuBreaks++;

									nextYPosition = nextYPosition+newitem->height+ySpacing;
								}
								else
								{
									if (adaptedmaxHeight < nextYPosition-ySpacing+bottomBorder)
										adaptedmaxHeight = nextYPosition-ySpacing+bottomBorder;
									
									nextYPosition = topBorder;

									newitem->y = nextYPosition;

									hiddenPart++;
									
									nextYPosition = nextYPosition+newitem->height+ySpacing;
								}
							}
							else
							{
								nextYPosition = nextYPosition+newitem->height+ySpacing;
							}
						}
						
						newitem->hiddenPart = hiddenPart;
						newitem->multiPart = menuBreaks;
						
						easyHelper = addsettingPopup->subname.c_str();
						easyHelper.Replace("sub_", "");

						if ( iTooltipMode != 0 )
						{
							if ( strstr(folderCaption, "||") != NULL )
							{
								newitem->caption = folderCaption.Left(folderCaption.Find("||"));
								if ( strstr(folderCaption, "\\n") != NULL )
									folderCaption.Replace("\\n", "\n");
								newitem->tooltip = folderCaption.Right(folderCaption.GetLength()-folderCaption.Find("||")-2);
							}
							else
							{
								newitem->caption = easyHelper;
								if ( iTooltipMode == 1 )
									newitem->tooltip = easyHelper;
							}	
						}
						else
						{
							if ( strstr(folderCaption, "||") != NULL )
							{
								easyHelper = folderCaption.Left(folderCaption.Find("||"));
							}
						
							newitem->caption = easyHelper;
						}

						ScanKeyCodes(newitem, (string) easyHelper);

						if ( DefaultFolderIcon.empty() )
						{
							if ( _stricmp(addsettingCustomIcon.c_str(), ".noexist") == NULL )
								newitem->iconPath = "*";
							else
								newitem->iconPath = addsettingCustomIcon;
						}
						else
						{
							if ( _stricmp(addsettingCustomIcon.c_str(), ".noexist") == NULL )
								newitem->iconPath = DefaultFolderIcon;
							else
								newitem->iconPath = addsettingCustomIcon;
						}

						int tmp = ComputeEntryWidth(newitem->caption);
						
						if (maxNeededWidth < tmp)
							maxNeededWidth = tmp;

						entrylist.insert(entrylist.end(), newitem);
					}
					addsettingPopup = NULL;
					addsettingCustomIcon = "";
					input = false;
					sendToPopup.clear();
				}
			}
			//-----------------------------------------------------
			// SEPARATOR
			//-----------------------------------------------------
			else if ( _stricmp(str_second, "!separator") == NULL || _stricmp(str_second, "!seperator") == NULL )
			{
				if ( iQuicklaunchMode != 0 )
					continue;

				if (input)
				{
					//Add to specific Popup Settings list
					sendToPopup.insert(sendToPopup.end(), *sit);	
				}
				else
				{
					ENTRY* newitem = new ENTRY;	
					
					newitem->entryType = 1;

					newitem->x = nextXPosition;
					if ( nextYPosition != topBorder )
					{
						newitem->y = nextYPosition+separatorySpacing;
						nextYPosition = newitem->y;
					}
					else
					{
						newitem->y = nextYPosition;
					}
					
					newitem->height = separatorHeight;
					
					if (nextYPosition+newitem->height+bottomBorder > maxHeight)
					{
						if (autoMenuBreak)
						{
							if (adaptedmaxHeight < nextYPosition-ySpacing+bottomBorder)
								adaptedmaxHeight = nextYPosition-ySpacing+bottomBorder;

							nextXPosition = nextXPosition+maxNeededWidth+xSpacing;
							nextYPosition = topBorder;

							newitem->x = nextXPosition;
							newitem->y = nextYPosition;

							menuBreaks++;

							nextYPosition = nextYPosition+newitem->height+ySpacing+separatorySpacing;
						}
						else
						{
							if (adaptedmaxHeight < nextYPosition-ySpacing+bottomBorder)
								adaptedmaxHeight = nextYPosition-ySpacing+bottomBorder;
							
							nextYPosition = topBorder;

							newitem->y = nextYPosition;

							hiddenPart++;
							
							nextYPosition = nextYPosition+newitem->height+ySpacing+separatorySpacing;
						}
					}
					else
					{
						nextYPosition = nextYPosition+newitem->height+ySpacing+separatorySpacing;
					}
						
					newitem->hiddenPart = hiddenPart;
					newitem->multiPart = menuBreaks;

					entrylist.insert(entrylist.end(), newitem);
				}
			}
			//-----------------------------------------------------
			// MODULE HOOK
			//-----------------------------------------------------
			else if ( _stricmp(str_second.Left(5), "!hook") == NULL )
			{
				if ( iQuicklaunchMode != 0 )
					continue;

				if (input)
				{
					//Add to specific Popup Settings list
					sendToPopup.insert(sendToPopup.end(), *sit);	
				}
				else
				{
					//Disable FakeAlphaMap
					fakealphaMap = false;

					str_second = str_second.Right(str_second.GetLength()-6);
					int hookleft = atoi( str_second.Left(str_second.Find(":")) );
					str_second = str_second.Right(str_second.GetLength()-str_second.Find(":")-1);
					int hookright = atoi(str_second.Left(str_second.Find(":")));
					str_second = str_second.Right(str_second.GetLength()-str_second.Find(":")-1);
					int hookwidth = atoi(str_second.Left(str_second.Find(":")));
					str_second = str_second.Right(str_second.GetLength()-str_second.Find(":")-1);
					int hookheight = atoi(str_second);

					string hWndbuffer = IntToString( (int) hWnd );
					string config = str_third;

					config.append( " \"$themedir$theme.rc\" " );
					config.append( hWndbuffer );

					LSExecute(NULL, config.c_str(), SW_SHOWNORMAL);

					if (hookleft == 0 && hookright == 0 && hookwidth == 0 && hookheight == 0)
						continue;

					if (nextYPosition-ySpacing+hookheight+bottomBorder > maxHeight)
					{
						if (autoMenuBreak)
						{
							if (adaptedmaxHeight < nextYPosition-ySpacing+bottomBorder)
								adaptedmaxHeight = nextYPosition-ySpacing+bottomBorder;

							nextXPosition = nextXPosition+maxNeededWidth+xSpacing;
							nextYPosition = topBorder;

							easyHelper = str_third;
							if (easyHelper.Find("%%") != -1)
							{
								easyHelper = easyHelper.Right(easyHelper.GetLength()-easyHelper.Find("%%")-2);
								easyHelper = easyHelper.Left(easyHelper.Find("%%"));
								easyHelper += " ";
								easyHelper += IntToString(hookleft).c_str();
								easyHelper += " ";
								easyHelper += IntToString(nextXPosition).c_str();
								easyHelper += " ";
								easyHelper += IntToString(nextYPosition).c_str();
							
								hookcontentList.insert(hookcontentList.end(), (string)easyHelper);
							}

							menuBreaks++;

							nextYPosition = nextYPosition+hookheight+ySpacing;
						}
						else
						{
							if (adaptedmaxHeight < nextYPosition-ySpacing+bottomBorder)
								adaptedmaxHeight = nextYPosition-ySpacing+bottomBorder;
							
							nextYPosition = topBorder;

							easyHelper = str_third;
							if (easyHelper.Find("%%") != -1)
							{
								easyHelper = easyHelper.Right(easyHelper.GetLength()-easyHelper.Find("%%")-2);
								easyHelper = easyHelper.Left(easyHelper.Find("%%"));
								easyHelper += " ";
								easyHelper += IntToString(hookleft).c_str();
								easyHelper += " ";
								easyHelper += IntToString(nextXPosition).c_str();
								easyHelper += " ";
								easyHelper += IntToString(nextYPosition).c_str();
							
								hookcontentList.insert(hookcontentList.end(), (string)easyHelper);
							}

							hiddenPart++;
							
							nextYPosition = nextYPosition+hookheight+ySpacing;
						}
					}
					else
					{
						easyHelper = str_third;
						if (easyHelper.Find("%%") != -1)
						{
							easyHelper = easyHelper.Right(easyHelper.GetLength()-easyHelper.Find("%%")-2);
							easyHelper = easyHelper.Left(easyHelper.Find("%%"));
							easyHelper += " ";
							easyHelper += IntToString(hookleft).c_str();
							easyHelper += " ";
							easyHelper += IntToString(nextXPosition).c_str();
							easyHelper += " ";
							easyHelper += IntToString(nextYPosition).c_str();
							
							hookcontentList.insert(hookcontentList.end(), (string)easyHelper);
						}

						nextYPosition = nextYPosition+hookheight+ySpacing;
					}

					if (maxNeededWidth < hookleft+hookwidth+hookright)
						maxNeededWidth = hookleft+hookwidth+hookright;

					hasHookedWindows = true;
				}
			}
			//-----------------------------------------------------
			// INFO ENTRY (xLabel(Light)) REQUIRED
			//-----------------------------------------------------
			else if ( _stricmp(str_second.Left(5), "!info") == NULL )
			{
				if ( iQuicklaunchMode != 0 )
					continue;

				if (input)
				{
					//Add to specific Popup Settings list
					sendToPopup.insert(sendToPopup.end(), *sit);	
				}
				else
				{
					ENTRY* newitem = new ENTRY;

					//Needs the original "szThird", cause "str_third" hasn't Quotations anymore
					newitem->info = szThird;

					newitem->entryType = 2;

					newitem->x = nextXPosition;
					newitem->y = nextYPosition;
					newitem->height = infoHeight;
					
					if (nextYPosition+newitem->height+bottomBorder > maxHeight)
					{
						if (autoMenuBreak)
						{
							if (adaptedmaxHeight < nextYPosition-ySpacing+bottomBorder)
								adaptedmaxHeight = nextYPosition-ySpacing+bottomBorder;

							nextXPosition = nextXPosition+maxNeededWidth+xSpacing;
							nextYPosition = topBorder;

							newitem->x = nextXPosition;
							newitem->y = nextYPosition;

							menuBreaks++;

							nextYPosition = nextYPosition+newitem->height+ySpacing;
						}
						else
						{
							if (adaptedmaxHeight < nextYPosition-ySpacing+bottomBorder)
								adaptedmaxHeight = nextYPosition-ySpacing+bottomBorder;
							
							nextYPosition = topBorder;

							newitem->y = nextYPosition;

							hiddenPart++;
							
							nextYPosition = nextYPosition+newitem->height+ySpacing;
						}
					}
					else
					{
						nextYPosition = nextYPosition+newitem->height+ySpacing;
					}
						
					newitem->hiddenPart = hiddenPart;
					newitem->multiPart = menuBreaks;

					if ( customIcon )
						newitem->iconPath = str_first;

					newitem->caption = InfoUpdate( newitem->info );

					//Measure with Type 2 only check InfoFont Settings
					int tmp = ComputeEntryWidth(newitem->caption, false, true);
					
					if (maxNeededWidth < tmp)
						maxNeededWidth = tmp;

					entrylist.insert(entrylist.end(), newitem);

					hasInfoEntry = true;
				}
			}
			//-----------------------------------------------------
			// INSERT POPUPFOLDERS
			//-----------------------------------------------------
			else if ( _stricmp(str_second.Left(13), "!insertfolder") == NULL || _stricmp(str_second.Left(14), "!xinsertfolder") == NULL )
			{
				if (input)
				{
					//Add to specific Popup Settings list
					sendToPopup.insert(sendToPopup.end(), *sit);	
				}
				else
				{
					easyHelper = str_second;
					
					// Get FolderPath
					easyHelper = easyHelper.Right(easyHelper.GetLength()-easyHelper.Find(":")-1);
					easyHelper = TrimAndStripQuote( (string)easyHelper ).c_str();
					//---------------------------------------------------------------------------------------
					// START Scanning
					//---------------------------------------------------------------------------------------
					string searchPath = (string) easyHelper;
					string wildcardPath = "";
					string itemName = "";
					string tmpHelper = "";

					HANDLE hFindHandle;
					WIN32_FIND_DATA FindData;

					vector<string> folders;
					vector<string> files;

					bool newIteration = true;
					CString worker = searchPath.c_str();
					while (newIteration == true)
					{
						if (worker.Find("|") != -1)
						{
							searchPath = worker.Left(worker.Find("|"));
							worker = worker.Right(worker.GetLength()-worker.Find("|")-1);
						}
						else
						{
							searchPath = worker;
							newIteration = false;
						}
						
						//Check for ending BackSlash
						tmpHelper = searchPath.substr(searchPath.length()-1, searchPath.length());
						bool addBackslash = false;
						if (tmpHelper != "\\" && strstr(searchPath.c_str(), "*.") == NULL)
						{
							wildcardPath = (searchPath + "\\*.*");
							addBackslash = true;
						}
						else if (strstr(searchPath.c_str(), "*.") == NULL)
						{
							wildcardPath = (searchPath + "*.*");
						}
						else
						{
							wildcardPath = searchPath;
							searchPath = searchPath.substr(0, searchPath.find('*'));
						}

						hFindHandle = FindFirstFile(wildcardPath.c_str(), &FindData);
						
						for(;;)
						{
							if(INVALID_HANDLE_VALUE == hFindHandle)
								break;

							itemName = FindData.cFileName;

							//It is a folder
							if (_stricmp(itemName.c_str(), ".") != NULL && _stricmp(itemName.c_str(), "..") != NULL /*&& !(FindData.dwFileAttributes & FILE_ATTRIBUTE_HIDDEN)*/)
							{
								if (addBackslash)
									tmpHelper = (searchPath + "\\" + itemName);
								else
									tmpHelper = (searchPath + itemName);
		
								if ( (FindData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) && (!(FindData.dwFileAttributes & FILE_ATTRIBUTE_HIDDEN) || ((FindData.dwFileAttributes & FILE_ATTRIBUTE_HIDDEN) && (iShowHiddenFiles == 1 || iShowHiddenFiles == 3)) ) )
								{
									//Check for double folders and unite
									CString helper = "";
									bool found = false;
									for (StringListIterator sit = folders.begin(); sit != folders.end(); ++sit)
									{
										easyHelper = (*sit).c_str();
										
										if (_stricmp( itemName.c_str(), easyHelper.Right( itemName.length() ) ) == NULL)
										{
											easyHelper += "|";
											easyHelper += tmpHelper.c_str();
											(*sit) = (string)easyHelper;

											found = true;
											break;
										}
									}
									
									if ( !found )
										folders.insert(folders.end(), tmpHelper);
								}
								else if ( !(FindData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) && (!(FindData.dwFileAttributes & FILE_ATTRIBUTE_HIDDEN) || ((FindData.dwFileAttributes & FILE_ATTRIBUTE_HIDDEN) && (iShowHiddenFiles == 2 || iShowHiddenFiles == 3) )) )
								{
									files.insert(files.end(), tmpHelper);
								}
							}

							if(!FindNextFile(hFindHandle, &FindData))
								break;
						}
					}

					FindClose(hFindHandle);
					//---------------------------------------------------------------------------------------
					// END Scanning
					//---------------------------------------------------------------------------------------

					//Sorting
					Sort(folders);
					Sort(files);

					StringListIterator sit;

					for (sit = folders.begin(); sit != folders.end(); ++sit)
					{
						ENTRY* newitem = new ENTRY;
						
						easyHelper = (*sit).c_str();
						easyHelper = easyHelper.Right(easyHelper.GetLength()-easyHelper.ReverseFind('\\')-1);

						//Remove Custom Sort HelperStrings
						if ( easyHelper.Find("[") == 0 && easyHelper.Find("]") == 3 )
							easyHelper = easyHelper.Right(easyHelper.GetLength()-4);
						//--------------------------------
						newitem->caption = easyHelper;

						ScanKeyCodes(newitem, (string) easyHelper, true);
									
						newitem->entryType = -2;

						if (iQuicklaunchMode != 0)
						{
							itemCounter++;

							newitem->width = quicklaunchItemSize;
							newitem->height = quicklaunchItemSize;
							newitem->x = nextXPosition;
							newitem->y = nextYPosition;

							if (iQuicklaunchMode == 1 || iQuicklaunchMode == -1)
							{
								if (itemCounter%quicklaunchWrapCount != 0)
								{
									nextXPosition = nextXPosition+newitem->width+xSpacing;
									nextYPosition = nextYPosition;
								}
								else
								{
									nextXPosition = leftBorder;
									nextYPosition = nextYPosition+newitem->height+ySpacing;
								}
							}
							else if (iQuicklaunchMode == 2 || iQuicklaunchMode == -2)
							{
								if (itemCounter%quicklaunchWrapCount != 0)
								{
									nextXPosition = nextXPosition;
									nextYPosition = nextYPosition+newitem->height+ySpacing;
								}
								else
								{
									nextXPosition = nextXPosition+newitem->width+xSpacing;
									nextYPosition = topBorder;
								}
							}
						}
						else
						{
							newitem->x = nextXPosition;
							newitem->y = nextYPosition;
							newitem->height = folderHeight;
										
							if (nextYPosition+newitem->height+bottomBorder > maxHeight)
							{
								if (autoMenuBreak)
								{
									if (adaptedmaxHeight < nextYPosition-ySpacing+bottomBorder)
										adaptedmaxHeight = nextYPosition-ySpacing+bottomBorder;

									nextXPosition = nextXPosition+maxNeededWidth+xSpacing;
									nextYPosition = topBorder;

									newitem->x = nextXPosition;
									newitem->y = nextYPosition;

									menuBreaks++;

									nextYPosition = nextYPosition+newitem->height+ySpacing;
								}
								else
								{
									if (adaptedmaxHeight < nextYPosition-ySpacing+bottomBorder)
										adaptedmaxHeight = nextYPosition-ySpacing+bottomBorder;
										
									nextYPosition = topBorder;

									newitem->y = nextYPosition;

									hiddenPart++;
										
									nextYPosition = nextYPosition+newitem->height+ySpacing;
								}
							}
							else
							{
								nextYPosition = nextYPosition+newitem->height+ySpacing;
							}
						}
								
						newitem->hiddenPart = hiddenPart;
						newitem->multiPart = menuBreaks;

						//Folders MUSTN'T have Quotes
						//Scan and process Piped Folders
						bool newIteration = true;
						CString worker = (*sit).c_str();
						while (newIteration == true)
						{
							if (worker.Find("|") != -1)
							{
								newitem->pipedList.insert(newitem->pipedList.end(), (string)worker.Left(worker.Find("|")));
								worker = worker.Right(worker.GetLength()-worker.Find("|")-1);
							}
							else
							{
								newitem->pipedList.insert(newitem->pipedList.end(), (string)worker);
								newIteration = false;
							}
						}

						newitem->action = (*sit);
					
						if ( DefaultFolderIcon.empty() )
							newitem->iconPath = "*";
						else
							newitem->iconPath = DefaultFolderIcon;

						int tmp = ComputeEntryWidth(newitem->caption);
						
						if (maxNeededWidth < tmp)
							maxNeededWidth = tmp;

						entrylist.insert(entrylist.end(), newitem);
					}

					for (sit = files.begin(); sit != files.end(); ++sit)
					{
						ENTRY* newitem = new ENTRY;
						
						easyHelper = (*sit).c_str();
						easyHelper = easyHelper.Right(easyHelper.GetLength()-easyHelper.ReverseFind('\\')-1);
						if (easyHelper.ReverseFind('.') != -1 && !showExtensions)
						{
							easyHelper = easyHelper.Left(easyHelper.ReverseFind('.'));
						}
						else if (easyHelper.ReverseFind('.') != -1 && noshowExtensions.size() != 0)
						{
							CString help1;
							CString help2;

							for (StringListIterator isit = noshowExtensions.begin(); isit != noshowExtensions.end(); ++isit)
							{
								help1 = easyHelper;
								help2 = (*isit).c_str();
								help1.MakeLower();
								help2.MakeLower();

								if ( help1.Find(help2) >= help1.GetLength()-5 )
								{
									easyHelper = easyHelper.Left(easyHelper.ReverseFind('.'));
									break;
								}
							}
						}

						//Remove Custom Sort HelperStrings
						if ( easyHelper.Find("[") == 0 && easyHelper.Find("]") == 3 )
							easyHelper = easyHelper.Right(easyHelper.GetLength()-4);
						//--------------------------------
						newitem->caption = easyHelper;

						ScanKeyCodes(newitem, (string) easyHelper, true);
					
						if (iQuicklaunchMode != 0)
						{
							itemCounter++;

							newitem->width = quicklaunchItemSize;
							newitem->height = quicklaunchItemSize;
							newitem->x = nextXPosition;
							newitem->y = nextYPosition;

							if (iQuicklaunchMode == 1 || iQuicklaunchMode == -1)
							{
								if (itemCounter%quicklaunchWrapCount != 0)
								{
									nextXPosition = nextXPosition+newitem->width+xSpacing;
									nextYPosition = nextYPosition;
								}
								else
								{
									nextXPosition = leftBorder;
									nextYPosition = nextYPosition+newitem->height+ySpacing;
								}
							}
							else if (iQuicklaunchMode == 2 || iQuicklaunchMode == -2)
							{
								if (itemCounter%quicklaunchWrapCount != 0)
								{
									nextXPosition = nextXPosition;
									nextYPosition = nextYPosition+newitem->height+ySpacing;
								}
								else
								{
									nextXPosition = nextXPosition+newitem->width+xSpacing;
									nextYPosition = topBorder;
								}
							}
						}
						else
						{
							newitem->x = nextXPosition;
							newitem->y = nextYPosition;
							newitem->height = entryHeight;
										
							if (nextYPosition+newitem->height+bottomBorder > maxHeight)
							{
								if (autoMenuBreak)
								{
									if (adaptedmaxHeight < nextYPosition-ySpacing+bottomBorder)
										adaptedmaxHeight = nextYPosition-ySpacing+bottomBorder;

									nextXPosition = nextXPosition+maxNeededWidth+xSpacing;
									nextYPosition = topBorder;

									newitem->x = nextXPosition;
									newitem->y = nextYPosition;

									menuBreaks++;

									nextYPosition = nextYPosition+newitem->height+ySpacing;
								}
								else
								{
									if (adaptedmaxHeight < nextYPosition-ySpacing+bottomBorder)
										adaptedmaxHeight = nextYPosition-ySpacing+bottomBorder;
											
									nextYPosition = topBorder;

									newitem->y = nextYPosition;

									hiddenPart++;
										
									nextYPosition = nextYPosition+newitem->height+ySpacing;
								}
							}
							else
							{
								nextYPosition = nextYPosition+newitem->height+ySpacing;
							}
						}
										
						newitem->hiddenPart = hiddenPart;
						newitem->multiPart = menuBreaks;

						//Files MUST have Quotes
						newitem->action = ("\"" + (*sit) + "\"");

						newitem->iconPath = (*sit);

						int tmp = ComputeEntryWidth(newitem->caption);
						
						if (maxNeededWidth < tmp)
							maxNeededWidth = tmp;

						entrylist.insert(entrylist.end(), newitem);
					}
				}
			}
			//-----------------------------------------------------
			// DYNAMIC POPUPFOLDERS
			//-----------------------------------------------------
			else if ( _stricmp(str_third.Left(19), "!popupdynamicfolder") == NULL || _stricmp(str_third.Left(14), "!dynamicfolder") == NULL || _stricmp(str_third.Left(12), "!popupfolder") == NULL )
			{
				if (input)
				{
					//Add to specific Popup Settings list
					sendToPopup.insert(sendToPopup.end(), *sit);	
				}
				else
				{
					ENTRY* newitem = new ENTRY;

					newitem->entryType = -1;
					if ( _stricmp(str_third.Left(19), "!popupdynamicfolder") == NULL || _stricmp(str_third.Left(14), "!dynamicfolder") == NULL )
						newitem->entryType = -2;

					if (iQuicklaunchMode != 0)
					{
						itemCounter++;

						newitem->width = quicklaunchItemSize;
						newitem->height = quicklaunchItemSize;
						newitem->x = nextXPosition;
						newitem->y = nextYPosition;

						if (iQuicklaunchMode == 1 || iQuicklaunchMode == -1)
						{
							if (itemCounter%quicklaunchWrapCount != 0)
							{
								nextXPosition = nextXPosition+newitem->width+xSpacing;
								nextYPosition = nextYPosition;
							}
							else
							{
								nextXPosition = leftBorder;
								nextYPosition = nextYPosition+newitem->height+ySpacing;
							}
						}
						else if (iQuicklaunchMode == 2 || iQuicklaunchMode == -2)
						{
							if (itemCounter%quicklaunchWrapCount != 0)
							{
								nextXPosition = nextXPosition;
								nextYPosition = nextYPosition+newitem->height+ySpacing;
							}
							else
							{
								nextXPosition = nextXPosition+newitem->width+xSpacing;
								nextYPosition = topBorder;
							}
						}
					}
					else
					{
						newitem->x = nextXPosition;
						newitem->y = nextYPosition;
						newitem->height = folderHeight;
					
						if (nextYPosition+newitem->height+bottomBorder > maxHeight)
						{
							if (autoMenuBreak)
							{
								if (adaptedmaxHeight < nextYPosition-ySpacing+bottomBorder)
									adaptedmaxHeight = nextYPosition-ySpacing+bottomBorder;

								nextXPosition = nextXPosition+maxNeededWidth+xSpacing;
								nextYPosition = topBorder;

								newitem->x = nextXPosition;
								newitem->y = nextYPosition;

								menuBreaks++;

								nextYPosition = nextYPosition+newitem->height+ySpacing;
							}
							else
							{
								if (adaptedmaxHeight < nextYPosition-ySpacing+bottomBorder)
									adaptedmaxHeight = nextYPosition-ySpacing+bottomBorder;
								
								nextYPosition = topBorder;

								newitem->y = nextYPosition;

								hiddenPart++;
								
								nextYPosition = nextYPosition+newitem->height+ySpacing;
							}
						}
						else
						{
							nextYPosition = nextYPosition+newitem->height+ySpacing;
						}
					}

					newitem->hiddenPart = hiddenPart;
					newitem->multiPart = menuBreaks;

					if ( iTooltipMode != 0 )
					{
						if ( strstr(str_second, "||") != NULL )
						{
							newitem->caption = str_second.Left(str_second.Find("||"));
							if ( strstr(str_second, "\\n") != NULL )
								str_second.Replace("\\n", "\n");
							newitem->tooltip = str_second.Right(str_second.GetLength()-str_second.Find("||")-2);
						}
						else
						{
							newitem->caption = str_second;
							if ( iTooltipMode == 1 )
								newitem->tooltip = str_second;
						}
					}
					else
					{
						if ( strstr(str_second, "||") != NULL )
						{
							easyHelper = str_second.Left(str_second.Find("||"));
							newitem->caption = easyHelper;
						}
						else
						{
							newitem->caption = str_second;
						}
					}

					ScanKeyCodes(newitem, (string) str_second);
									
					easyHelper = str_third;
					
					// Get FolderPath
					easyHelper = easyHelper.Right(easyHelper.GetLength()-easyHelper.Find(":")-1);
					easyHelper = TrimAndStripQuote( (string)easyHelper ).c_str();

					//Scan and process Piped Folders
					bool newIteration = true;
					CString worker = easyHelper;
					while (newIteration == true)
					{
						if (worker.Find("|") != -1)
						{
							newitem->pipedList.insert(newitem->pipedList.end(), (string)worker.Left(worker.Find("|")));
							worker = worker.Right(worker.GetLength()-worker.Find("|")-1);
						}
						else
						{
							newitem->pipedList.insert(newitem->pipedList.end(), (string)worker);
							newIteration = false;
						}
					}

					newitem->action = easyHelper;

					if ( DefaultFolderIcon.empty() )
					{
						if (customIcon)
							newitem->iconPath = str_first;
						else
							newitem->iconPath = "*";
					}
					else
					{
						if (customIcon)
							newitem->iconPath = str_first;
						else
							newitem->iconPath = DefaultFolderIcon;
					}

					newitem->layout = (string) str_fourth;
					newitem->defaultlayout = (string) str_fifth;

					int tmp = ComputeEntryWidth(newitem->caption);
					
					if (maxNeededWidth < tmp)
						maxNeededWidth = tmp;

					entrylist.insert(entrylist.end(), newitem);
				}
			}
			//-----------------------------------------------------
			// DYNAMIC ACTION POPUPFOLDERS
			//-----------------------------------------------------
			else if ( _stricmp(str_third.Left(25), "!popupdynamicactionfolder") == NULL || _stricmp(str_third.Left(18), "!popupactionfolder") == NULL )
			{
				if (input)
				{
					//Add to specific Popup Settings list
					sendToPopup.insert(sendToPopup.end(), *sit);	
				}
				else
				{
					ENTRY* newitem = new ENTRY;

					newitem->entryType = -1;
					if ( _stricmp(str_third.Left(25), "!popupdynamicactionfolder") == NULL )
						newitem->entryType = -2;

					if (iQuicklaunchMode != 0)
					{
						itemCounter++;

						newitem->width = quicklaunchItemSize;
						newitem->height = quicklaunchItemSize;
						newitem->x = nextXPosition;
						newitem->y = nextYPosition;

						if (iQuicklaunchMode == 1 || iQuicklaunchMode == -1)
						{
							if (itemCounter%quicklaunchWrapCount != 0)
							{
								nextXPosition = nextXPosition+newitem->width+xSpacing;
								nextYPosition = nextYPosition;
							}
							else
							{
								nextXPosition = leftBorder;
								nextYPosition = nextYPosition+newitem->height+ySpacing;
							}
						}
						else if (iQuicklaunchMode == 2 || iQuicklaunchMode == -2)
						{
							if (itemCounter%quicklaunchWrapCount != 0)
							{
								nextXPosition = nextXPosition;
								nextYPosition = nextYPosition+newitem->height+ySpacing;
							}
							else
							{
								nextXPosition = nextXPosition+newitem->width+xSpacing;
								nextYPosition = topBorder;
							}
						}
					}
					else
					{
						newitem->x = nextXPosition;
						newitem->y = nextYPosition;
						newitem->height = folderHeight;
					
						if (nextYPosition+newitem->height+bottomBorder > maxHeight)
						{
							if (autoMenuBreak)
							{
								if (adaptedmaxHeight < nextYPosition-ySpacing+bottomBorder)
									adaptedmaxHeight = nextYPosition-ySpacing+bottomBorder;

								nextXPosition = nextXPosition+maxNeededWidth+xSpacing;
								nextYPosition = topBorder;

								newitem->x = nextXPosition;
								newitem->y = nextYPosition;

								menuBreaks++;

								nextYPosition = nextYPosition+newitem->height+ySpacing;
							}
							else
							{
								if (adaptedmaxHeight < nextYPosition-ySpacing+bottomBorder)
									adaptedmaxHeight = nextYPosition-ySpacing+bottomBorder;
								
								nextYPosition = topBorder;

								newitem->y = nextYPosition;

								hiddenPart++;
								
								nextYPosition = nextYPosition+newitem->height+ySpacing;
							}
						}
						else
						{
							nextYPosition = nextYPosition+newitem->height+ySpacing;
						}
					}

					newitem->hiddenPart = hiddenPart;
					newitem->multiPart = menuBreaks;

					if ( iTooltipMode != 0 )
					{
						if ( strstr(str_second, "||") != NULL )
						{
							newitem->caption = str_second.Left(str_second.Find("||"));
							if ( strstr(str_second, "\\n") != NULL )
								str_second.Replace("\\n", "\n");
							newitem->tooltip = str_second.Right(str_second.GetLength()-str_second.Find("||")-2);
						}
						else
						{
							newitem->caption = str_second;
							if ( iTooltipMode == 1 )
								newitem->tooltip = str_second;
						}
					}
					else
					{
						if ( strstr(str_second, "||") != NULL )
						{
							easyHelper = str_second.Left(str_second.Find("||"));
							newitem->caption = easyHelper;
						}
						else
						{
							newitem->caption = str_second;
						}
					}

					ScanKeyCodes(newitem, (string) str_second);
									
					easyHelper = str_third;
					
					// Get FolderPath (Strip !action between ":" ... ":"
					int pos = easyHelper.Find(":");
					if ( easyHelper.Find(":", pos+3) == -1 )
					{
						delete newitem;
						newitem = NULL;
						continue;
					}

					//Find 2nd ":" from backwards
					int sep2 = easyHelper.ReverseFind(':');
					sep2 = easyHelper.Find(":", sep2-3); // ...:c:\path

					// Get universalAction
					newitem->universalaction = easyHelper.Mid(pos+1, sep2-pos-1);

					easyHelper = easyHelper.Right(easyHelper.GetLength()-sep2-1);
					
					easyHelper = TrimAndStripQuote( (string)easyHelper ).c_str();

					//Scan and process Piped Folders
					bool newIteration = true;
					CString worker = easyHelper;
					while (newIteration == true)
					{
						if (worker.Find("|") != -1)
						{
							newitem->pipedList.insert(newitem->pipedList.end(), (string)worker.Left(worker.Find("|")));
							worker = worker.Right(worker.GetLength()-worker.Find("|")-1);
						}
						else
						{
							newitem->pipedList.insert(newitem->pipedList.end(), (string)worker);
							newIteration = false;
						}
					}

					newitem->action = easyHelper;

					if ( DefaultFolderIcon.empty() )
					{
						if (customIcon)
							newitem->iconPath = str_first;
						else
							newitem->iconPath = "*";
					}
					else
					{
						if (customIcon)
							newitem->iconPath = str_first;
						else
							newitem->iconPath = DefaultFolderIcon;
					}

					newitem->layout = (string) str_fourth;
					newitem->defaultlayout = (string) str_fifth;

					int tmp = ComputeEntryWidth(newitem->caption);
					
					if (maxNeededWidth < tmp)
						maxNeededWidth = tmp;

					entrylist.insert(entrylist.end(), newitem);
				}
			}
			//-----------------------------------------------------
			// POPUPTASKS
			//-----------------------------------------------------
			else if ( _stricmp(str_third, "!popuptasks") == NULL )
			{
				if ( iQuicklaunchMode != 0 )
					continue;

				if (input)
				{
					//Add to specific Popup Settings list
					sendToPopup.insert(sendToPopup.end(), *sit);	
				}
				else
				{
					ENTRY* newitem = new ENTRY;

					newitem->hWnd = hWnd;

					newitem->entryType = -1;

					newitem->x = nextXPosition;
					newitem->y = nextYPosition;
					newitem->height = folderHeight;
					
					if (nextYPosition+newitem->height+bottomBorder > maxHeight)
					{
						if (autoMenuBreak)
						{
							if (adaptedmaxHeight < nextYPosition-ySpacing+bottomBorder)
								adaptedmaxHeight = nextYPosition-ySpacing+bottomBorder;

							nextXPosition = nextXPosition+maxNeededWidth+xSpacing;
							nextYPosition = topBorder;

							newitem->x = nextXPosition;
							newitem->y = nextYPosition;

							menuBreaks++;

							nextYPosition = nextYPosition+newitem->height+ySpacing;
						}
						else
						{
							if (adaptedmaxHeight < nextYPosition-ySpacing+bottomBorder)
								adaptedmaxHeight = nextYPosition-ySpacing+bottomBorder;
							
							nextYPosition = topBorder;

							newitem->y = nextYPosition;

							hiddenPart++;
							
							nextYPosition = nextYPosition+newitem->height+ySpacing;
						}
					}
					else
					{
						nextYPosition = nextYPosition+newitem->height+ySpacing;
					}
						
					newitem->hiddenPart = hiddenPart;
					newitem->multiPart = menuBreaks;

					if ( iTooltipMode != 0 )
					{
						if ( strstr(str_second, "||") != NULL )
						{
							newitem->caption = str_second.Left(str_second.Find("||"));
							if ( strstr(str_second, "\\n") != NULL )
								str_second.Replace("\\n", "\n");
							newitem->tooltip = str_second.Right(str_second.GetLength()-str_second.Find("||")-2);
						}
						else
						{
							newitem->caption = str_second;
							if ( iTooltipMode == 1 )
								newitem->tooltip = str_second;
						}
						
					}
					else
					{
						if ( strstr(str_second, "||") != NULL )
						{
							easyHelper = str_second.Left(str_second.Find("||"));
							newitem->caption = easyHelper;
						}
						else
						{
							newitem->caption = str_second;
						}
					}

					ScanKeyCodes(newitem, (string) str_second, true);
			
					if (customIcon)
						easyHelper = str_first;
					else
						easyHelper = DefaultFolderIcon.c_str();

					newitem->iconPath = easyHelper;

					newitem->layout = (string) str_fourth;
					newitem->defaultlayout = (string) str_fifth;

					int tmp = ComputeEntryWidth(newitem->caption);
					
					if (maxNeededWidth < tmp)
						maxNeededWidth = tmp;

					entrylist.insert(entrylist.end(), newitem);
				}
			}
			//-----------------------------------------------------
			// POPUP SHELLFOLDERS
			//-----------------------------------------------------
			else if ( _stricmp(str_third, "!popupcontrolpanel") == NULL || _stricmp(str_third, "!popupprinters") == NULL 
						|| _stricmp(str_third, "!popuprecyclebin") == NULL || _stricmp(str_third, "!popuprecentdocuments") == NULL
						|| _stricmp(str_third, "!popupnetwork") == NULL || _stricmp(str_third, "!popupmycomputer") == NULL )
			{
				if ( iQuicklaunchMode != 0 )
					continue;

				if (input)
				{
					//Add to specific Popup Settings list
					sendToPopup.insert(sendToPopup.end(), *sit);	
				}
				else
				{	
					ENTRY* newitem = new ENTRY;

					if ( _stricmp(str_third, "!popupcontrolpanel") == NULL )
						newitem->clsid = CSIDL_CONTROLS;
					else if (_stricmp(str_third, "!popupmycomputer") == NULL )
						newitem->clsid = CSIDL_DRIVES;
					else if (_stricmp(str_third, "!popupnetwork") == NULL )
						newitem->clsid = CSIDL_NETWORK;
					else if (_stricmp(str_third, "!popupprinters") == NULL )
						newitem->clsid = CSIDL_PRINTERS;
					else if (_stricmp(str_third, "!popuprecentdocuments") == NULL )
						newitem->clsid = CSIDL_RECENT;
					else if (_stricmp(str_third, "!popuprecyclebin") == NULL )
						newitem->clsid = CSIDL_BITBUCKET;
					else
					{
						delete newitem;
						continue;
					}

					newitem->entryType = -1;

					newitem->x = nextXPosition;
					newitem->y = nextYPosition;
					newitem->height = entryHeight;

					//Shell Folder PIDL for Icon Extraction
					SHGetSpecialFolderLocation(NULL, newitem->clsid, &newitem->m_pidl);
					
					if (nextYPosition+newitem->height+bottomBorder > maxHeight)
					{
						if (autoMenuBreak)
						{
							if (adaptedmaxHeight < nextYPosition-ySpacing+bottomBorder)
								adaptedmaxHeight = nextYPosition-ySpacing+bottomBorder;

							nextXPosition = nextXPosition+maxNeededWidth+xSpacing;
							nextYPosition = topBorder;

							newitem->x = nextXPosition;
							newitem->y = nextYPosition;

							menuBreaks++;

							nextYPosition = nextYPosition+newitem->height+ySpacing;
						}
						else
						{
							if (adaptedmaxHeight < nextYPosition-ySpacing+bottomBorder)
								adaptedmaxHeight = nextYPosition-ySpacing+bottomBorder;
							
							nextYPosition = topBorder;

							newitem->y = nextYPosition;

							hiddenPart++;
							
							nextYPosition = nextYPosition+newitem->height+ySpacing;
						}
					}
					else
					{
						nextYPosition = nextYPosition+newitem->height+ySpacing;
					}
						
					newitem->hiddenPart = hiddenPart;
					newitem->multiPart = menuBreaks;

					if ( iTooltipMode != 0 )
					{
						if ( strstr(str_second, "||") != NULL )
						{
							newitem->caption = str_second.Left(str_second.Find("||"));
							if ( strstr(str_second, "\\n") != NULL )
								str_second.Replace("\\n", "\n");
							newitem->tooltip = str_second.Right(str_second.GetLength()-str_second.Find("||")-2);
						}
						else
						{
							newitem->caption = str_second;
							if ( iTooltipMode == 1 )
								newitem->tooltip = str_second;
						}
						
					}
					else
					{
						if ( strstr(str_second, "||") != NULL )
						{
							easyHelper = str_second.Left(str_second.Find("||"));
							newitem->caption = easyHelper;
						}
						else
						{
							newitem->caption = str_second;
						}
					}

					ScanKeyCodes(newitem, (string) str_second, true);
			
					if (customIcon)
						newitem->iconPath = str_first;

					newitem->layout = (string) str_fourth;
					newitem->defaultlayout = (string) str_fifth;

					int tmp = ComputeEntryWidth(newitem->caption);
					
					if (maxNeededWidth < tmp)
						maxNeededWidth = tmp;

					entrylist.insert(entrylist.end(), newitem);
				}
			}
			//-----------------------------------------------------
			// NORMAL ENTRYS
			//-----------------------------------------------------
			else if ( _stricmp(str_second, "!popupimage") != NULL && _stricmp(str_third, "!popuprun") != NULL
					 && _stricmp(str_third, "!datetime") != NULL && _stricmp(str_second, "!remoteamp") != NULL 
					 && _stricmp(str_third, "!popupdesktops") != NULL )
			{
				if (input)
				{
					//Add to specific Popup Settings list
					sendToPopup.insert(sendToPopup.end(), *sit);	
				}
				else
				{
					ENTRY* newitem = new ENTRY;

					if (iQuicklaunchMode != 0)
					{
						itemCounter++;

						newitem->tooltip = str_second;

						newitem->width = quicklaunchItemSize;
						newitem->height = quicklaunchItemSize;
						newitem->x = nextXPosition;
						newitem->y = nextYPosition;

						if (iQuicklaunchMode == 1 || iQuicklaunchMode == -1)
						{
							if (itemCounter%quicklaunchWrapCount != 0)
							{
								nextXPosition = nextXPosition+newitem->width+xSpacing;
								nextYPosition = nextYPosition;
							}
							else
							{
								nextXPosition = leftBorder;
								nextYPosition = nextYPosition+newitem->height+ySpacing;
							}
						}
						else if (iQuicklaunchMode == 2 || iQuicklaunchMode == -2)
						{
							if (itemCounter%quicklaunchWrapCount != 0)
							{
								nextXPosition = nextXPosition;
								nextYPosition = nextYPosition+newitem->height+ySpacing;
							}
							else
							{
								nextXPosition = nextXPosition+newitem->width+xSpacing;
								nextYPosition = topBorder;
							}
						}
					}
					else
					{
						newitem->x = nextXPosition;
						newitem->y = nextYPosition;
						newitem->height = entryHeight;
						
						if (nextYPosition+newitem->height+bottomBorder > maxHeight)
						{
							if (autoMenuBreak)
							{
								if (adaptedmaxHeight < nextYPosition-ySpacing+bottomBorder)
									adaptedmaxHeight = nextYPosition-ySpacing+bottomBorder;

								nextXPosition = nextXPosition+maxNeededWidth+xSpacing;
								nextYPosition = topBorder;

								newitem->x = nextXPosition;
								newitem->y = nextYPosition;

								menuBreaks++;

								nextYPosition = nextYPosition+newitem->height+ySpacing;
							}
							else
							{
								if (adaptedmaxHeight < nextYPosition-ySpacing+bottomBorder)
									adaptedmaxHeight = nextYPosition-ySpacing+bottomBorder;
								
								nextYPosition = topBorder;

								newitem->y = nextYPosition;

								hiddenPart++;
								
								nextYPosition = nextYPosition+newitem->height+ySpacing;
							}
						}
						else
						{
							nextYPosition = nextYPosition+newitem->height+ySpacing;
						}
					}
						
					newitem->hiddenPart = hiddenPart;
					newitem->multiPart = menuBreaks;

					if ( iTooltipMode != 0 )
					{
						if ( strstr(str_second, "||") != NULL )
						{
							newitem->caption = str_second.Left(str_second.Find("||"));
							if ( strstr(str_second, "\\n") != NULL )
								str_second.Replace("\\n", "\n");
							newitem->tooltip = str_second.Right(str_second.GetLength()-str_second.Find("||")-2);
						}
						else
						{
							newitem->caption = str_second;
							if ( iTooltipMode == 1 )
								newitem->tooltip = str_second;
						}
						
					}
					else
					{
						if ( strstr(str_second, "||") != NULL )
						{
							easyHelper = str_second.Left(str_second.Find("||"));
							newitem->caption = easyHelper;
						}
						else
						{
							newitem->caption = str_second;
						}
					}

					ScanKeyCodes(newitem, (string) str_second);
		
					if ( PathFileExists(str_third) )
						newitem->action = ("\"" + str_third + "\"");
					else
						newitem->action = str_third;

					if (customIcon)
						easyHelper = str_first;
					else
						easyHelper = TrimAndStripQuote( newitem->action ).c_str();

					if ( customIcon )
						newitem->iconPath = easyHelper;
					else if ( easyHelper.Left(1) != "!" )
						newitem->iconPath = easyHelper;
					else 
						newitem->iconPath = DefaultIcon;

					int tmp = ComputeEntryWidth(newitem->caption);
					
					if (maxNeededWidth < tmp)
						maxNeededWidth = tmp;

					entrylist.insert(entrylist.end(), newitem);
				}
			}
		}

		if (iQuicklaunchMode != 0)
		{
			int tmpCount = 0;
			if (itemCounter < quicklaunchWrapCount)
				tmpCount = itemCounter;
			else
				tmpCount = quicklaunchWrapCount;

			if (iQuicklaunchMode == 1 || iQuicklaunchMode == -1)
			{
				actualwidth = leftBorder+tmpCount*quicklaunchItemSize+(tmpCount-1)*xSpacing+rightBorder;
				actualheight = nextYPosition+((itemCounter%quicklaunchWrapCount != 0) ? quicklaunchItemSize : 0)+bottomBorder;
			}
			else if (iQuicklaunchMode == 2 || iQuicklaunchMode == -2)
			{
				actualwidth = nextXPosition+((itemCounter%quicklaunchWrapCount != 0) ? quicklaunchItemSize : 0)+rightBorder;
				actualheight = topBorder+tmpCount*quicklaunchItemSize+(tmpCount-1)*ySpacing+bottomBorder;
			}
		}
		else
		{
			//Check Max/Min EntryWidth
			maxNeededWidth = max(min(maxNeededWidth, maxEntryWidth), minEntryWidth);

			// Check Last MenuColumn for Height
			if (adaptedmaxHeight != 0 && adaptedmaxHeight < nextYPosition-ySpacing+bottomBorder)
				adaptedmaxHeight = nextYPosition-ySpacing+bottomBorder;

			//Measure with Type 1 only check TitleFont Settings for maxNeededWidth
			int tmp = ComputeEntryWidth(captionname, true);
			int neededTitleWidth = leftBorder+tmp+rightBorder;

			//Adapt EntryWidth based on Max/Min Width
			if ( leftBorder+(menuBreaks+1)*maxNeededWidth+menuBreaks*xSpacing+rightBorder >= maxWidth )
			{
				actualwidth = maxWidth;
				maxNeededWidth = (actualwidth-leftBorder-rightBorder-menuBreaks*xSpacing)/(menuBreaks+1);
			}
			else if (leftBorder+(menuBreaks+1)*maxNeededWidth+menuBreaks*xSpacing+rightBorder <= minWidth )
			{
				actualwidth = minWidth;
				maxNeededWidth = (actualwidth-leftBorder-rightBorder-menuBreaks*xSpacing)/(menuBreaks+1);
			}
			else
			{
				actualwidth = leftBorder+(menuBreaks+1)*maxNeededWidth+menuBreaks*xSpacing+rightBorder;
				actualwidth = min(max(neededTitleWidth, actualwidth), maxWidth);
			}

			if (hiddenPart != 0 || menuBreaks != 0)
				actualheight = adaptedmaxHeight;
			else
				actualheight = min(nextYPosition-ySpacing+bottomBorder, maxHeight);
			actualheight = max(actualheight, minHeight);

			//Hook Positioning
			//----------------------------------------------
			string x = "";
			string y = "";
			string space = "";
		
			for (sit = hookcontentList.begin(); sit != hookcontentList.end(); ++sit)
			{
				x = "0";
				y = "0";
				space = "0";
				easyHelper = (*sit).c_str();
				y = easyHelper.Right(easyHelper.GetLength()-easyHelper.ReverseFind(' '));
				easyHelper = easyHelper.Left(easyHelper.ReverseFind(' '));
				x = easyHelper.Right(easyHelper.GetLength()-easyHelper.ReverseFind(' '));
				easyHelper = easyHelper.Left(easyHelper.ReverseFind(' '));
				space = easyHelper.Right(easyHelper.GetLength()-easyHelper.ReverseFind(' '));
				easyHelper = easyHelper.Left(easyHelper.ReverseFind(' '));
				
				int iterate = leftBorder;
				while ( atoi(x.c_str()) > iterate )
					iterate = iterate+maxNeededWidth+xSpacing;

				x = IntToString(iterate+atoi(space.c_str()));

				easyHelper += " ";
				easyHelper += x.c_str();
				easyHelper += " ";
				easyHelper += y.c_str();

				LSExecute(NULL, easyHelper, SW_SHOWNORMAL);
			}
		}

		hookcontentList.clear();
		//----------------------------------------------

		startentry = NULL;

		for (EntryListIterator eit = entrylist.begin(); eit != entrylist.end(); ++eit)
		{
			if (startentry == NULL && (*eit)->entryType <= 0)
				startentry = *eit;

			if (iQuicklaunchMode == 0)
			{
				if (menuBreaks == 0)
				{
					(*eit)->width = max(maxNeededWidth, actualwidth-leftBorder-rightBorder);
				}
				else
				{
					(*eit)->width = maxNeededWidth;
					
					(*eit)->x = leftBorder + (*eit)->multiPart*(maxNeededWidth + xSpacing);
				}
			}

			if ( subFolderLevel <= iCacheLevel )
			{
				if ( (*eit)->entryType < 0 && (*eit)->clsid == 0 && (*eit)->hWnd == NULL )
					CreateSubMenu( ("sub_"+(*eit)->caption), (*eit)->action, (*eit), false, true );
				else if ( (*eit)->clsid != 0 && (*eit)->hWnd == NULL )
					CreateShellSubMenu( ("sub_"+(*eit)->caption), (*eit)->clsid, (*eit), false, true );
			}
		}

		sendToPopup.clear();

		iMaxParts = hiddenPart;
		iMaxMenuBreaks = menuBreaks;
	}

	reposition(2, 0, 0, actualwidth, actualheight);
	repaint(true);
	
	settingsList.clear();

	if ( !hasDynamicContent )
	{
		popupcontentList.clear();
		str_completeCheckStringBackup = "";
	}

	KillTimer(hWnd, TIMER_UPDATE);

	if ((isTaskPopup || hasDynamicContent) && !hasInfoEntry)
		SetTimer(hWnd, TIMER_UPDATE, tasksUpdate, NULL);
	else if (hasInfoEntry)
		SetTimer(hWnd, TIMER_UPDATE, 1000, NULL);
}

void Popup::ScanSubMenu(string scanPath, int dynamic, string universalAction)
{
	bool forcedDriveRescan = true;
	bool NotifyUpdate = false;

	if ( scanPath.empty() && dynamic == -100 && universalAction.empty() )
	{
		scanPath = rootScanPath;
		dynamic = -2;
		universalAction = rootUniversalAction;
		
		forcedDriveRescan = true;
		NotifyUpdate = true;
	}

	for(PopupListIterator plit = subpopupList.begin(); plit != subpopupList.end(); plit++)
	{
		(*plit)->hide(true);
		(*plit)->captionname = "123deleted-remove";
	}

	for (EntryListIterator deit = entrylist.begin(); deit != entrylist.end(); ++deit)
	{	
		delete *deit;
		*deit = NULL;
	}

	entrylist.clear();

	//Reset old Pointers
	//-------------------------
	iCurrentPart = 0;

	startentry = NULL;
	lastactiveentry = NULL;
	hoverentry = NULL;
	//-------------------------

	CString easyHelper = "";

	//---------------------------------------------------------------------------------------
	// START Scanning
	//---------------------------------------------------------------------------------------

		string searchPath = "";
		string wildcardPath = "";
		string itemName = "";
		string tmpHelper = "";
		
		//BOOL for only FolderMatching ActionFolder
		bool actionFolder_FolderList = false;
		
		StringList folders;
		StringList files;
		
		//---------------------------------------------------------------------
		if ( forcedDriveRescan )
		{
		//---------------------------------------------------------------------
			
		HANDLE hFindHandle;
		WIN32_FIND_DATA FindData;

		bool newIteration = true;
		CString worker = scanPath.c_str();

		while (newIteration == true)
		{
			if (worker.Find("|") != -1)
			{
				searchPath = worker.Left(worker.Find("|"));
				worker = worker.Right(worker.GetLength()-worker.Find("|")-1);
			}
			else
			{
				searchPath = worker;
				newIteration = false;
			}

			//Check for ending BackSlash
			tmpHelper = searchPath.substr(searchPath.length()-1, searchPath.length());
			bool addBackslash = false;
			
			if ( !universalAction.empty() && (strstr(searchPath.c_str(), "*\\") != NULL || strstr(searchPath.c_str(), "*/") != NULL) )
			{
				searchPath = searchPath.substr(0, searchPath.length()-2);
				wildcardPath = (searchPath + "*.*");
				actionFolder_FolderList = true;
			}
			else if ( tmpHelper != "\\" && tmpHelper != "/" && strstr(searchPath.c_str(), "*.") == NULL)
			{
				wildcardPath = (searchPath + "\\*.*");
				addBackslash = true;
			}
			else if (strstr(searchPath.c_str(), "*.") == NULL)
			{
				wildcardPath = (searchPath + "*.*");
			}
			else
			{
				wildcardPath = searchPath;
				searchPath = searchPath.substr(0, searchPath.find('*'));
			}

			hFindHandle = FindFirstFile(wildcardPath.c_str(), &FindData);
			
			for(;;)
			{
				if(INVALID_HANDLE_VALUE == hFindHandle)
					break;

				itemName = FindData.cFileName;

				//It is a folder
				if (_stricmp(itemName.c_str(), ".") != NULL && _stricmp(itemName.c_str(), "..") != NULL /*&& !(FindData.dwFileAttributes & FILE_ATTRIBUTE_HIDDEN)*/)
				{
					if (addBackslash)
						tmpHelper = (searchPath + "\\" + itemName);
					else
						tmpHelper = (searchPath + itemName);
					
					if ( (FindData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) && (!(FindData.dwFileAttributes & FILE_ATTRIBUTE_HIDDEN) || ((FindData.dwFileAttributes & FILE_ATTRIBUTE_HIDDEN) && (iShowHiddenFiles == 1 || iShowHiddenFiles == 3)) ) )
					{
						//Check for double folders and unite
						bool found = false;
						for (StringListIterator sit = folders.begin(); sit != folders.end(); ++sit)
						{
							easyHelper = (*sit).c_str();

							if (_stricmp(itemName.c_str(), easyHelper.Right(easyHelper.GetLength()-easyHelper.ReverseFind('\\')-1)) == NULL)
							{
								easyHelper += "|";
								easyHelper += tmpHelper.c_str();
								(*sit) = (string)easyHelper;

								found = true;
								break;
							}
						}
						
						if ( !found )
						{
							folders.insert(folders.end(), tmpHelper);
						}
					}
					else if ( !actionFolder_FolderList && !(FindData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) && (!(FindData.dwFileAttributes & FILE_ATTRIBUTE_HIDDEN) || ((FindData.dwFileAttributes & FILE_ATTRIBUTE_HIDDEN) && (iShowHiddenFiles == 2 || iShowHiddenFiles == 3) )) )
					{
						files.insert(files.end(), tmpHelper);
					}
				}

				if(!FindNextFile(hFindHandle, &FindData))
					break;
			}
		}

		FindClose(hFindHandle);

		//---------------------------------------------------------------------
		}
		//---------------------------------------------------------------------
	
	//---------------------------------------------------------------------------------------
	// END Scanning
	//---------------------------------------------------------------------------------------

	//No SubFolders in !popupactionfolder
	if ( !universalAction.empty() )
	{
		if ( !actionFolder_FolderList )
		{
			folders.clear();
		}
		else 
		{
			files = folders;
			folders.clear();
		}
	}
	
	//Sorting
	Sort(folders);
	Sort(files);

	//Initializing of Values
	int itemCounter = 0;
	int nextXPosition = leftBorder;
	int nextYPosition = topBorder;
	int hiddenPart = 0;
	int menuBreaks = 0;
	int maxNeededWidth = 0;
	int adaptedmaxHeight = 0;

	//Insert AutoSeparator if wished between Title and Entrys
	if (autoSeparator >= 2 && iQuicklaunchMode == 0)
	{
		ENTRY* newitem = new ENTRY;	
					
		newitem->entryType = 1;

		newitem->x = nextXPosition;
		if ( nextYPosition != topBorder )
		{
			newitem->y = nextYPosition+separatorySpacing;
			nextYPosition = newitem->y;
		}
		else
		{
			newitem->y = nextYPosition;
		}
			
		newitem->height = separatorHeight;
			
		if (nextYPosition+newitem->height+bottomBorder > maxHeight)
		{
			if (autoMenuBreak)
			{
				if (adaptedmaxHeight < nextYPosition-ySpacing+bottomBorder)
					adaptedmaxHeight = nextYPosition-ySpacing+bottomBorder;

				nextXPosition = nextXPosition+maxNeededWidth+xSpacing;
				nextYPosition = topBorder;

				newitem->x = nextXPosition;
				newitem->y = nextYPosition;

				menuBreaks++;

				nextYPosition = nextYPosition+newitem->height+ySpacing+separatorySpacing;
			}
			else
			{
				if (adaptedmaxHeight < nextYPosition-ySpacing+bottomBorder)
					adaptedmaxHeight = nextYPosition-ySpacing+bottomBorder;
							
				nextYPosition = topBorder;

				newitem->y = nextYPosition;

				hiddenPart++;
							
				nextYPosition = nextYPosition+newitem->height+ySpacing+separatorySpacing;
			}
		}
		else
		{
			nextYPosition = nextYPosition+newitem->height+ySpacing+separatorySpacing;
		}
						
		newitem->hiddenPart = hiddenPart;
		newitem->multiPart = menuBreaks;

		entrylist.insert(entrylist.end(), newitem);
	}

	StringListIterator sit;

	for (sit = folders.begin(); sit != folders.end(); ++sit)
	{
		itemCounter++;

		ENTRY* newitem = new ENTRY;
		
		easyHelper = (*sit).c_str();
		easyHelper = easyHelper.Right(easyHelper.GetLength()-easyHelper.ReverseFind('\\')-1);

		//Remove Custom Sort HelperStrings
		if ( easyHelper.Find("[") == 0 && easyHelper.Find("]") == 3 )
			easyHelper = easyHelper.Right(easyHelper.GetLength()-4);

		easyHelper.Replace("&", "&&");
		//--------------------------------
		newitem->caption = easyHelper;

		ScanKeyCodes(newitem, (string) easyHelper, true);
			
		newitem->entryType = -1;
		if (dynamic == -2)
			newitem->entryType = -2;

		if (iQuicklaunchMode != 0)
		{
			newitem->width = quicklaunchItemSize;
			newitem->height = quicklaunchItemSize;
			newitem->x = nextXPosition;
			newitem->y = nextYPosition;

			if (iQuicklaunchMode == 1 || iQuicklaunchMode == -1)
			{
				if (itemCounter%quicklaunchWrapCount != 0)
				{
					nextXPosition = nextXPosition+newitem->width+xSpacing;
					nextYPosition = nextYPosition;
				}
				else
				{
					nextXPosition = leftBorder;
					nextYPosition = nextYPosition+newitem->height+ySpacing;
				}
			}
			else if (iQuicklaunchMode == 2 || iQuicklaunchMode == -2)
			{
				if (itemCounter%quicklaunchWrapCount != 0)
				{
					nextXPosition = nextXPosition;
					nextYPosition = nextYPosition+newitem->height+ySpacing;
				}
				else
				{
					nextXPosition = nextXPosition+newitem->width+xSpacing;
					nextYPosition = topBorder;
				}
			}
		}
		else
		{
			newitem->x = nextXPosition;
			newitem->y = nextYPosition;
			newitem->height = folderHeight;
					
			if (nextYPosition+newitem->height+bottomBorder > maxHeight)
			{
				if (autoMenuBreak)
				{
					if (adaptedmaxHeight < nextYPosition-ySpacing+bottomBorder)
						adaptedmaxHeight = nextYPosition-ySpacing+bottomBorder;

					nextXPosition = nextXPosition+maxNeededWidth+xSpacing;
					nextYPosition = topBorder;

					newitem->x = nextXPosition;
					newitem->y = nextYPosition;

					menuBreaks++;

					nextYPosition = nextYPosition+newitem->height+ySpacing;
				}
				else
				{
					if (adaptedmaxHeight < nextYPosition-ySpacing+bottomBorder)
						adaptedmaxHeight = nextYPosition-ySpacing+bottomBorder;
					
					nextYPosition = topBorder;

					newitem->y = nextYPosition;

					hiddenPart++;
					
					nextYPosition = nextYPosition+newitem->height+ySpacing;
				}
			}
			else
			{
				nextYPosition = nextYPosition+newitem->height+ySpacing;
			}
		}
				
		newitem->hiddenPart = hiddenPart;
		newitem->multiPart = menuBreaks;

		//Folders MUSTN'T have Quotes
		//Scan and process Piped Folders
		bool newIteration = true;
		CString worker = (*sit).c_str();
		while (newIteration == true)
		{
			if (worker.Find("|") != -1)
			{
				newitem->pipedList.insert(newitem->pipedList.end(), (string)worker.Left(worker.Find("|")));
				worker = worker.Right(worker.GetLength()-worker.Find("|")-1);
			}
			else
			{
				newitem->pipedList.insert(newitem->pipedList.end(), (string)worker);
				newIteration = false;
			}
		}

		newitem->action = (*sit);
	
		if ( DefaultFolderIcon.empty() )
			newitem->iconPath = "*";
		else
			newitem->iconPath = DefaultFolderIcon;

		int tmp = ComputeEntryWidth(newitem->caption);

		if (maxNeededWidth < tmp)
			maxNeededWidth = tmp;

		entrylist.insert(entrylist.end(), newitem);
	}

	//Insert AutoSeparator if wished between Folder and Files
	if (autoSeparator >= 1 && folders.size() >= 1 && files.size() >= 1 && iQuicklaunchMode == 0)
	{
		ENTRY* newitem = new ENTRY;	
					
		newitem->entryType = 1;

		newitem->x = nextXPosition;
		if ( nextYPosition != topBorder )
		{
			newitem->y = nextYPosition+separatorySpacing;
			nextYPosition = newitem->y;
		}
		else
		{
			newitem->y = nextYPosition;
		}
					
		newitem->height = separatorHeight;
					
		if (nextYPosition+newitem->height+bottomBorder > maxHeight)
		{
			if (autoMenuBreak)
			{
				if (adaptedmaxHeight < nextYPosition-ySpacing+bottomBorder)
					adaptedmaxHeight = nextYPosition-ySpacing+bottomBorder;

				nextXPosition = nextXPosition+maxNeededWidth+xSpacing;
				nextYPosition = topBorder;

				newitem->x = nextXPosition;
				newitem->y = nextYPosition;

				menuBreaks++;

				nextYPosition = nextYPosition+newitem->height+ySpacing+separatorySpacing;
			}
			else
			{
				if (adaptedmaxHeight < nextYPosition-ySpacing+bottomBorder)
					adaptedmaxHeight = nextYPosition-ySpacing+bottomBorder;
							
				nextYPosition = topBorder;

				newitem->y = nextYPosition;

				hiddenPart++;
							
				nextYPosition = nextYPosition+newitem->height+ySpacing+separatorySpacing;
			}
		}
		else
		{
			nextYPosition = nextYPosition+newitem->height+ySpacing+separatorySpacing;
		}
						
		newitem->hiddenPart = hiddenPart;
		newitem->multiPart = menuBreaks;

		entrylist.insert(entrylist.end(), newitem);
	}

	for (sit = files.begin(); sit != files.end(); ++sit)
	{
		itemCounter++;

		ENTRY* newitem = new ENTRY;
		
		easyHelper = (*sit).c_str();
		easyHelper = easyHelper.Right(easyHelper.GetLength()-easyHelper.ReverseFind('\\')-1);
		if (!actionFolder_FolderList)
		{
			if (easyHelper.ReverseFind('.') != -1 && !showExtensions)
			{
				easyHelper = easyHelper.Left(easyHelper.ReverseFind('.'));
			}
			else if (easyHelper.ReverseFind('.') != -1 && noshowExtensions.size() != 0)
			{
				CString help1;
				CString help2;

				for (StringListIterator isit = noshowExtensions.begin(); isit != noshowExtensions.end(); ++isit)
				{
					help1 = easyHelper;
					help2 = (*isit).c_str();
					help1.MakeLower();
					help2.MakeLower();

					if ( help1.Find(help2) >= help1.GetLength()-5 )
					{
						easyHelper = easyHelper.Left(easyHelper.ReverseFind('.'));
						break;
					}
				}
			}
		}

		//Remove Custom Sort HelperStrings
		if ( easyHelper.Find("[") == 0 && easyHelper.Find("]") == 3 )
			easyHelper = easyHelper.Right(easyHelper.GetLength()-4);

		easyHelper.Replace("&", "&&");
		//--------------------------------
		newitem->caption = easyHelper;

		ScanKeyCodes(newitem, (string) easyHelper, true);

		if (iQuicklaunchMode != 0)
		{
			newitem->width = quicklaunchItemSize;
			newitem->height = quicklaunchItemSize;
			newitem->x = nextXPosition;
			newitem->y = nextYPosition;

			if (iQuicklaunchMode == 1 || iQuicklaunchMode == -1)
			{
				if (itemCounter%quicklaunchWrapCount != 0)
				{
					nextXPosition = nextXPosition+newitem->width+xSpacing;
					nextYPosition = nextYPosition;
				}
				else
				{
					nextXPosition = leftBorder;
					nextYPosition = nextYPosition+newitem->height+ySpacing;
				}
			}
			else if (iQuicklaunchMode == 2 || iQuicklaunchMode == -2)
			{
				if (itemCounter%quicklaunchWrapCount != 0)
				{
					nextXPosition = nextXPosition;
					nextYPosition = nextYPosition+newitem->height+ySpacing;
				}
				else
				{
					nextXPosition = nextXPosition+newitem->width+xSpacing;
					nextYPosition = topBorder;
				}
			}
		}
		else
		{
			newitem->x = nextXPosition;
			newitem->y = nextYPosition;
			newitem->height = entryHeight;
				
			if (nextYPosition+newitem->height+bottomBorder > maxHeight)
			{
				if (autoMenuBreak)
				{
					if (adaptedmaxHeight < nextYPosition-ySpacing+bottomBorder)
						adaptedmaxHeight = nextYPosition-ySpacing+bottomBorder;

					nextXPosition = nextXPosition+maxNeededWidth+xSpacing;
					nextYPosition = topBorder;

					newitem->x = nextXPosition;
					newitem->y = nextYPosition;

					menuBreaks++;

					nextYPosition = nextYPosition+newitem->height+ySpacing;
				}
				else
				{
					if (adaptedmaxHeight < nextYPosition-ySpacing+bottomBorder)
						adaptedmaxHeight = nextYPosition-ySpacing+bottomBorder;
						
					nextYPosition = topBorder;

					newitem->y = nextYPosition;

					hiddenPart++;
					
					nextYPosition = nextYPosition+newitem->height+ySpacing;
				}
			}
			else
			{
				nextYPosition = nextYPosition+newitem->height+ySpacing;
			}
		}
						
		newitem->hiddenPart = hiddenPart;
		newitem->multiPart = menuBreaks;

		//Files MUST have Quotes
		if ( !universalAction.empty() )
			newitem->universalaction = (universalAction + " \"" + (*sit) + "\"");
		
		newitem->action = ("\"" + (*sit) + "\"");

		newitem->iconPath = (*sit);

		int tmp = ComputeEntryWidth(newitem->caption);

		if (maxNeededWidth < tmp)
			maxNeededWidth = tmp;

		entrylist.insert(entrylist.end(), newitem);
	}

	//Insert AutoSeparator if wished at the Bottom
	if (autoSeparator == 3 && iQuicklaunchMode == 0)
	{
		ENTRY* newitem = new ENTRY;	
					
		newitem->entryType = 1;

		newitem->x = nextXPosition;
		if ( nextYPosition != topBorder )
		{
			newitem->y = nextYPosition+separatorySpacing;
			nextYPosition = newitem->y;
		}
		else
		{
			newitem->y = nextYPosition;
		}
					
		newitem->height = separatorHeight;
					
		if (nextYPosition+newitem->height+bottomBorder > maxHeight)
		{
			if (autoMenuBreak)
			{
				if (adaptedmaxHeight < nextYPosition-ySpacing+bottomBorder)
					adaptedmaxHeight = nextYPosition-ySpacing+bottomBorder;

				nextXPosition = nextXPosition+maxNeededWidth+xSpacing;
				nextYPosition = topBorder;

				newitem->x = nextXPosition;
				newitem->y = nextYPosition;

				menuBreaks++;

				nextYPosition = nextYPosition+newitem->height+ySpacing+separatorySpacing;
			}
			else
			{
				if (adaptedmaxHeight < nextYPosition-ySpacing+bottomBorder)
					adaptedmaxHeight = nextYPosition-ySpacing+bottomBorder;
							
				nextYPosition = topBorder;

				newitem->y = nextYPosition;

				hiddenPart++;
							
				nextYPosition = nextYPosition+newitem->height+ySpacing+separatorySpacing;
			}
		}
		else
		{
			nextYPosition = nextYPosition+newitem->height+ySpacing+separatorySpacing;
		}
						
		newitem->hiddenPart = hiddenPart;
		newitem->multiPart = menuBreaks;

		entrylist.insert(entrylist.end(), newitem);
	}

	if (iQuicklaunchMode != 0)
	{
		int tmpCount = 0;
		if (itemCounter < quicklaunchWrapCount)
			tmpCount = itemCounter;
		else
			tmpCount = quicklaunchWrapCount;

		if (iQuicklaunchMode == 1 || iQuicklaunchMode == -1)
		{
			actualwidth = leftBorder+tmpCount*quicklaunchItemSize+(tmpCount-1)*xSpacing+rightBorder;
			actualheight = nextYPosition+((itemCounter%quicklaunchWrapCount != 0) ? quicklaunchItemSize : 0)+bottomBorder;
		}
		else if (iQuicklaunchMode == 2 || iQuicklaunchMode == -2)
		{
			actualwidth = nextXPosition+((itemCounter%quicklaunchWrapCount != 0) ? quicklaunchItemSize : 0)+rightBorder;
			actualheight = topBorder+tmpCount*quicklaunchItemSize+(tmpCount-1)*ySpacing+bottomBorder;
		}
	}
	else
	{
		//Measure with Type 1 only check TitleFont Settings for maxNeededWidth
		int tmp = ComputeEntryWidth(captionname, true);
		int neededTitleWidth = leftBorder+tmp+rightBorder;

		//Check Max/Min EntryWidth
		maxNeededWidth = max(min(maxNeededWidth, maxEntryWidth), minEntryWidth);

		//Adapt EntryWidth based on Max/Min Width
		if ( leftBorder+(menuBreaks+1)*maxNeededWidth+menuBreaks*xSpacing+rightBorder >= maxWidth )
		{
			actualwidth = maxWidth;
			maxNeededWidth = (actualwidth-leftBorder-rightBorder-menuBreaks*xSpacing)/(menuBreaks+1);
		}
		else if (leftBorder+(menuBreaks+1)*maxNeededWidth+menuBreaks*xSpacing+rightBorder <= minWidth )
		{
			actualwidth = minWidth;
			maxNeededWidth = (actualwidth-leftBorder-rightBorder-menuBreaks*xSpacing)/(menuBreaks+1);
		}
		else
		{
			actualwidth = leftBorder+(menuBreaks+1)*maxNeededWidth+menuBreaks*xSpacing+rightBorder;
			actualwidth = min(max(neededTitleWidth, actualwidth), maxWidth);
		}

		if (hiddenPart != 0 || menuBreaks != 0)
			actualheight = adaptedmaxHeight;
		else
			actualheight = min(nextYPosition-ySpacing+bottomBorder, maxHeight);

		actualheight = max(actualheight, minHeight);
	}

	//Reset Helper Variables
	startentry = NULL;
	
	for (EntryListIterator eit = entrylist.begin(); eit != entrylist.end(); ++eit)
	{
		if (startentry == NULL && (*eit)->entryType <= 0)
			startentry = *eit;

		if (iQuicklaunchMode == 0)
		{
			if (menuBreaks == 0)
			{
				(*eit)->width = max(maxNeededWidth, actualwidth-leftBorder-rightBorder);
			}
			else
			{
				(*eit)->width = maxNeededWidth;
				
				(*eit)->x = leftBorder + (*eit)->multiPart*(maxNeededWidth + xSpacing);	
			}
		}

		if ( subFolderLevel <= iCacheLevel )
		{
			if ( (*eit)->entryType < 0 && (*eit)->clsid == 0 && (*eit)->hWnd == NULL )
				CreateSubMenu( ("sub_"+(*eit)->caption), (*eit)->action, (*eit), false, true );
		}
	}

	iMaxParts = hiddenPart;
	iMaxMenuBreaks = menuBreaks;

	reposition(2, 0, 0, actualwidth, actualheight);
	
	//Important UpdateWindowTransparent crashes, don't know why !??? :|
	if (!NotifyUpdate)
		repaint(true);
	else
		bResetBackground = true;
}

void Popup::ScanShellFolder(int clsid)
{
	for (EntryListIterator eit = entrylist.begin(); eit != entrylist.end(); ++eit)
	{	
		delete *eit;
		*eit = NULL;
	}

	entrylist.clear();

	//Reset old Pointers
	//-------------------------
	iCurrentPart = 0;

	startentry = NULL;
	lastactiveentry = NULL;
	hoverentry = NULL;
	//-------------------------

	//SHELL PART 1
	LPITEMIDLIST m_pidlFull = NULL;
	IShellFolder *m_pshf = NULL;
	SHGetSpecialFolderLocation(NULL, clsid, &m_pidlFull);

	if (m_pidlFull)
	{
		SHFILEINFO shfi;

		SHGetFileInfo((LPSTR) m_pidlFull, 0, &shfi, sizeof(SHFILEINFO), SHGFI_ATTRIBUTES | SHGFI_DISPLAYNAME |
		                                        SHGFI_PIDL | SHGFI_SYSICONINDEX | SHGFI_SMALLICON);

		if (shfi.dwAttributes & SFGAO_FOLDER && !(shfi.dwAttributes & SFGAO_REMOVABLE))
		{
			IShellFolder *pshfDesktop;
			SHGetDesktopFolder(&pshfDesktop);

			pshfDesktop->BindToObject(m_pidlFull, NULL, IID_IShellFolder,
			                          (void **) &m_pshf);

			pshfDesktop->Release();
		}
	}

	//---------------------------------------------------------------------------------------
	// START Scanning
	//---------------------------------------------------------------------------------------
	IEnumIDList *peidl;
	vector<LPITEMIDLIST> PIDLList;
	LPITEMIDLIST pidl;

	HRESULT hr = m_pshf->EnumObjects(NULL, SHCONTF_FOLDERS |
	                         SHCONTF_NONFOLDERS, &peidl);

	if ( SUCCEEDED(hr) )
	{
		while (peidl->Next(1, &pidl, NULL) == S_OK)
		{
			LPITEMIDLIST pidlFull = JoinIDLists(m_pidlFull, pidl);

			PIDLList.insert(PIDLList.end(), pidlFull);
		}

		peidl->Release();
	}
	//---------------------------------------------------------------------------------------
	// END Scanning
	//---------------------------------------------------------------------------------------

	//Sorting
	if (clsid != CSIDL_DRIVES && clsid != CSIDL_RECENT)
		SortShellFolder(PIDLList);

	//Initializing of Values
	int nextXPosition = leftBorder;
	int nextYPosition = topBorder;
	int hiddenPart = 0;
	int menuBreaks = 0;
	int maxNeededWidth = 0;
	int adaptedmaxHeight = 0;

	//Insert AutoSeparator if wished between Title and Entrys
	if (autoSeparator >= 2 && iQuicklaunchMode == 0)
	{
		ENTRY* newitem = new ENTRY;	
					
		newitem->entryType = 1;

		newitem->x = nextXPosition;
		if ( nextYPosition != topBorder )
		{
			newitem->y = nextYPosition+separatorySpacing;
			nextYPosition = newitem->y;
		}
		else
		{
			newitem->y = nextYPosition;
		}
					
		newitem->height = separatorHeight;
					
		if (nextYPosition+newitem->height+bottomBorder > maxHeight)
		{
			if (autoMenuBreak)
			{
				if (adaptedmaxHeight < nextYPosition-ySpacing+bottomBorder)
					adaptedmaxHeight = nextYPosition-ySpacing+bottomBorder;

				nextXPosition = nextXPosition+maxNeededWidth+xSpacing;
				nextYPosition = topBorder;

				newitem->x = nextXPosition;
				newitem->y = nextYPosition;

				menuBreaks++;

				nextYPosition = nextYPosition+newitem->height+ySpacing+separatorySpacing;
			}
			else
			{
				if (adaptedmaxHeight < nextYPosition-ySpacing+bottomBorder)
					adaptedmaxHeight = nextYPosition-ySpacing+bottomBorder;
							
				nextYPosition = topBorder;

				newitem->y = nextYPosition;

				hiddenPart++;
							
				nextYPosition = nextYPosition+newitem->height+ySpacing+separatorySpacing;
			}
		}
		else
		{
			nextYPosition = nextYPosition+newitem->height+ySpacing+separatorySpacing;
		}
						
		newitem->hiddenPart = hiddenPart;
		newitem->multiPart = menuBreaks;

		entrylist.insert(entrylist.end(), newitem);
	}

	for (vector<LPITEMIDLIST>::iterator pit = PIDLList.begin(); pit != PIDLList.end(); ++pit)
	{
		SHFILEINFO shfi;

		SHGetFileInfo((LPCSTR) (*pit), 0, &shfi, sizeof(SHFILEINFO),
										  SHGFI_ATTRIBUTES | SHGFI_DISPLAYNAME | SHGFI_PIDL );

		if (shfi.dwAttributes & SFGAO_FOLDER)
		{
			//Entry
			ENTRY* newitem = new ENTRY;

			newitem->m_pidl = (*pit);

			newitem->x = nextXPosition;
			newitem->y = nextYPosition;

			CString easyHelper = shfi.szDisplayName;
			easyHelper.Replace("&", "&&");

			newitem->caption = easyHelper;

			ScanKeyCodes(newitem, (string) easyHelper, true);

			newitem->height = entryHeight;
						
			if (nextYPosition+newitem->height+bottomBorder > maxHeight)
			{
				if (autoMenuBreak)
				{
					if (adaptedmaxHeight < nextYPosition-ySpacing+bottomBorder)
						adaptedmaxHeight = nextYPosition-ySpacing+bottomBorder;

					nextXPosition = nextXPosition+maxNeededWidth+xSpacing;
					nextYPosition = topBorder;

					newitem->x = nextXPosition;
					newitem->y = nextYPosition;

					menuBreaks++;

					nextYPosition = nextYPosition+newitem->height+ySpacing;
				}
				else
				{
					if (adaptedmaxHeight < nextYPosition-ySpacing+bottomBorder)
						adaptedmaxHeight = nextYPosition-ySpacing+bottomBorder;
							
					nextYPosition = topBorder;

					newitem->y = nextYPosition;

					hiddenPart++;
						
					nextYPosition = nextYPosition+newitem->height+ySpacing;
				}
			}
			else
			{
				nextYPosition = nextYPosition+newitem->height+ySpacing;
			}
								
			newitem->hiddenPart = hiddenPart;
			newitem->multiPart = menuBreaks;

			int tmp = ComputeEntryWidth(newitem->caption);

			if (maxNeededWidth < tmp)
				maxNeededWidth = tmp;

			entrylist.insert(entrylist.end(), newitem);
		}
		else
		{
			//Entry
			ENTRY* newitem = new ENTRY;

			newitem->m_pidl = (*pit);

			newitem->x = nextXPosition;
			newitem->y = nextYPosition;

			CString easyHelper = shfi.szDisplayName;
			easyHelper.Replace("&", "&&");

			newitem->caption = easyHelper;

			ScanKeyCodes(newitem, (string) easyHelper, true);
			
			newitem->height = entryHeight;
						
			if (nextYPosition+newitem->height+bottomBorder > maxHeight)
			{
				if (autoMenuBreak)
				{
					if (adaptedmaxHeight < nextYPosition-ySpacing+bottomBorder)
						adaptedmaxHeight = nextYPosition-ySpacing+bottomBorder;

					nextXPosition = nextXPosition+maxNeededWidth+xSpacing;
					nextYPosition = topBorder;

					newitem->x = nextXPosition;
					newitem->y = nextYPosition;

					menuBreaks++;

					nextYPosition = nextYPosition+newitem->height+ySpacing;
				}
				else
				{
					if (adaptedmaxHeight < nextYPosition-ySpacing+bottomBorder)
						adaptedmaxHeight = nextYPosition-ySpacing+bottomBorder;
							
					nextYPosition = topBorder;

					newitem->y = nextYPosition;

					hiddenPart++;
						
					nextYPosition = nextYPosition+newitem->height+ySpacing;
				}
			}
			else
			{
				nextYPosition = nextYPosition+newitem->height+ySpacing;
			}
								
			newitem->hiddenPart = hiddenPart;
			newitem->multiPart = menuBreaks;

			int tmp = ComputeEntryWidth(newitem->caption);

			if (maxNeededWidth < tmp)
				maxNeededWidth = tmp;

			entrylist.insert(entrylist.end(), newitem);
		}
	}

	//Insert AutoSeparator if wished at the Bottom
	if (autoSeparator == 3 && iQuicklaunchMode == 0)
	{
		ENTRY* newitem = new ENTRY;	
					
		newitem->entryType = 1;

		newitem->x = nextXPosition;
		if ( nextYPosition != topBorder )
		{
			newitem->y = nextYPosition+separatorySpacing;
			nextYPosition = newitem->y;
		}
		else
		{
			newitem->y = nextYPosition;
		}
					
		newitem->height = separatorHeight;
					
		if (nextYPosition+newitem->height+bottomBorder > maxHeight)
		{
			if (autoMenuBreak)
			{
				if (adaptedmaxHeight < nextYPosition-ySpacing+bottomBorder)
					adaptedmaxHeight = nextYPosition-ySpacing+bottomBorder;

				nextXPosition = nextXPosition+maxNeededWidth+xSpacing;
				nextYPosition = topBorder;

				newitem->x = nextXPosition;
				newitem->y = nextYPosition;

				menuBreaks++;

				nextYPosition = nextYPosition+newitem->height+ySpacing+separatorySpacing;
			}
			else
			{
				if (adaptedmaxHeight < nextYPosition-ySpacing+bottomBorder)
					adaptedmaxHeight = nextYPosition-ySpacing+bottomBorder;
							
				nextYPosition = topBorder;

				newitem->y = nextYPosition;

				hiddenPart++;
							
				nextYPosition = nextYPosition+newitem->height+ySpacing+separatorySpacing;
			}
		}
		else
		{
			nextYPosition = nextYPosition+newitem->height+ySpacing+separatorySpacing;
		}
						
		newitem->hiddenPart = hiddenPart;
		newitem->multiPart = menuBreaks;

		entrylist.insert(entrylist.end(), newitem);
	}

	//Check Max/Min EntryWidth
	maxNeededWidth = max(min(maxNeededWidth, maxEntryWidth), minEntryWidth);
		
	//Measure with Type 1 only check TitleFont Settings for maxNeededWidth
	int tmp = ComputeEntryWidth(captionname, true);
	int neededTitleWidth = leftBorder+tmp+rightBorder;

	//Adapt EntryWidth based on Max/Min Width
	if ( leftBorder+(menuBreaks+1)*maxNeededWidth+menuBreaks*xSpacing+rightBorder >= maxWidth )
	{
		actualwidth = maxWidth;
		maxNeededWidth = (actualwidth-leftBorder-rightBorder-menuBreaks*xSpacing)/(menuBreaks+1);
	}
	else if (leftBorder+(menuBreaks+1)*maxNeededWidth+menuBreaks*xSpacing+rightBorder <= minWidth )
	{
		actualwidth = minWidth;
		maxNeededWidth = (actualwidth-leftBorder-rightBorder-menuBreaks*xSpacing)/(menuBreaks+1);
	}
	else
	{
		actualwidth = leftBorder+(menuBreaks+1)*maxNeededWidth+menuBreaks*xSpacing+rightBorder;
		actualwidth = min(max(neededTitleWidth, actualwidth), maxWidth);
	}

	if (hiddenPart != 0 || menuBreaks != 0)
		actualheight = adaptedmaxHeight;
	else
		actualheight = min(nextYPosition-ySpacing+bottomBorder, maxHeight);
	actualheight = max(actualheight, minHeight);

	startentry = NULL;

	for (eit = entrylist.begin(); eit != entrylist.end(); ++eit)
	{
		if (startentry == NULL && (*eit)->entryType <= 0)
			startentry = *eit;

		if (menuBreaks == 0)
		{
			(*eit)->width = max(maxNeededWidth, actualwidth-leftBorder-rightBorder);
		}
		else
		{
			(*eit)->width = maxNeededWidth;
				
			(*eit)->x = leftBorder + (*eit)->multiPart*(maxNeededWidth + xSpacing);
		}
	}

	iMaxParts = hiddenPart;
	iMaxMenuBreaks = menuBreaks;

	reposition(2, 0, 0, actualwidth, actualheight);
	repaint(true);
}

void Popup::CreateLoadingPopup()
{
	if ( pParentPopup != NULL )
		pParentPopup->repaint();
	
	if (noLoadingPopup || alphaFade)
		return;
	
	//Create/Show "Loading" Popup
	ENTRY* newitem = new ENTRY;

	newitem->entryType = 2;

	newitem->x = leftBorder;
	newitem->y = topBorder;
	newitem->height = entryHeight;
	newitem->caption = "Loading ...";
	newitem->info = "Loading ...";

	newitem->hiddenPart = 0;
	newitem->multiPart = 0;

	int tmp = ComputeEntryWidth(newitem->caption, false, true);

	newitem->width = tmp;

	tmp = ComputeEntryWidth(captionname, true);

	tmp = max(newitem->width, tmp);

	newitem->width = tmp;

	entrylist.insert(entrylist.end(), newitem);

	actualwidth = leftBorder + tmp + rightBorder;
	actualheight = topBorder + newitem->height + bottomBorder;

	//Enable InfoEntry temporary
	hasInfoEntry = true;

	reposition(2, 0, 0, actualwidth, actualheight);
	repaint(true);
	
	show(false, true);
	
	//Disable InfoEntry 
	hasInfoEntry = false;
	//---------------------------
}

void Popup::ScanTasks()
{
	isTaskPopup = true;

	for (EntryListIterator eit = entrylist.begin(); eit != entrylist.end(); ++eit)
	{
		delete *eit;
		*eit = NULL;
	}

	entrylist.clear();
	
	//Reset old Pointers
	//-------------------------
	iCurrentPart = 0;

	startentry = NULL;
	lastactiveentry = NULL;
	hoverentry = NULL;
	//-------------------------

	//---------------------------------------------------------------------------------------
	// START Scanning
	//---------------------------------------------------------------------------------------
	tasklist.clear();

	EnumWindows(Popup::EnumWindowsProc, (LPARAM) this);
	//---------------------------------------------------------------------------------------
	// END Scanning
	//---------------------------------------------------------------------------------------

	//Sorting
	SortTasks(tasklist);

	//Initializing of Values
	int nextXPosition = leftBorder;
	int nextYPosition = topBorder;
	int hiddenPart = 0;
	int menuBreaks = 0;
	int maxNeededWidth = 0;
	int adaptedmaxHeight = 0;

	//Insert AutoSeparator if wished between Title and Entrys
	if (autoSeparator >= 2 && iQuicklaunchMode == 0)
	{
		ENTRY* newitem = new ENTRY;	
					
		newitem->entryType = 1;

		newitem->x = nextXPosition;
		if ( nextYPosition != topBorder )
		{
			newitem->y = nextYPosition+separatorySpacing;
			nextYPosition = newitem->y;
		}
		else
		{
			newitem->y = nextYPosition;
		}
					
		newitem->height = separatorHeight;
					
		if (nextYPosition+newitem->height+bottomBorder > maxHeight)
		{
			if (autoMenuBreak)
			{
				if (adaptedmaxHeight < nextYPosition-ySpacing+bottomBorder)
					adaptedmaxHeight = nextYPosition-ySpacing+bottomBorder;

				nextXPosition = nextXPosition+maxNeededWidth+xSpacing;
				nextYPosition = topBorder;

				newitem->x = nextXPosition;
				newitem->y = nextYPosition;

				menuBreaks++;

				nextYPosition = nextYPosition+newitem->height+ySpacing+separatorySpacing;
			}
			else
			{
				if (adaptedmaxHeight < nextYPosition-ySpacing+bottomBorder)
					adaptedmaxHeight = nextYPosition-ySpacing+bottomBorder;
							
				nextYPosition = topBorder;

				newitem->y = nextYPosition;

				hiddenPart++;
							
				nextYPosition = nextYPosition+newitem->height+ySpacing+separatorySpacing;
			}
		}
		else
		{
			nextYPosition = nextYPosition+newitem->height+ySpacing+separatorySpacing;
		}
						
		newitem->hiddenPart = hiddenPart;
		newitem->multiPart = menuBreaks;

		entrylist.insert(entrylist.end(), newitem);
	}

	for (hWndTaskListIterator tit = tasklist.begin(); tit != tasklist.end(); ++tit)
	{
		ENTRY* newitem = new ENTRY;

		newitem->x = nextXPosition;
		
		char windowCaption[MAX_LINE_LENGTH] = {0};
		GetWindowText( (*tit), windowCaption, MAX_LINE_LENGTH );

		CString easyHelper = windowCaption;
		easyHelper.Replace("&", "&&");
		
		newitem->caption = easyHelper;

		ScanKeyCodes(newitem, (string) easyHelper, true);

		newitem->y = nextYPosition;
		newitem->height = entryHeight;
				
		if (nextYPosition+newitem->height+bottomBorder > maxHeight)
		{
			if (autoMenuBreak)
			{
				if (adaptedmaxHeight < nextYPosition-ySpacing+bottomBorder)
					adaptedmaxHeight = nextYPosition-ySpacing+bottomBorder;

				nextXPosition = nextXPosition+maxNeededWidth+xSpacing;
				nextYPosition = topBorder;

				newitem->x = nextXPosition;
				newitem->y = nextYPosition;

				menuBreaks++;

				nextYPosition = nextYPosition+newitem->height+ySpacing;
			}
			else
			{
				if (adaptedmaxHeight < nextYPosition-ySpacing+bottomBorder)
					adaptedmaxHeight = nextYPosition-ySpacing+bottomBorder;
					
				nextYPosition = topBorder;

				newitem->y = nextYPosition;

				hiddenPart++;
				
				nextYPosition = nextYPosition+newitem->height+ySpacing;
			}
		}
		else
		{
			nextYPosition = nextYPosition+newitem->height+ySpacing;
		}
						
		newitem->hiddenPart = hiddenPart;
		newitem->multiPart = menuBreaks;

		newitem->hWnd = (*tit);

		int tmp = ComputeEntryWidth(newitem->caption);

		if (maxNeededWidth < tmp)
			maxNeededWidth = tmp;

		entrylist.insert(entrylist.end(), newitem);
	}

	//Insert AutoSeparator if wished at the Bottom
	if (autoSeparator == 3 && iQuicklaunchMode == 0)
	{
		ENTRY* newitem = new ENTRY;	
					
		newitem->entryType = 1;

		newitem->x = nextXPosition;
		if ( nextYPosition != topBorder )
		{
			newitem->y = nextYPosition+separatorySpacing;
			nextYPosition = newitem->y;
		}
		else
		{
			newitem->y = nextYPosition;
		}
					
		newitem->height = separatorHeight;
					
		if (nextYPosition+newitem->height+bottomBorder > maxHeight)
		{
			if (autoMenuBreak)
			{
				if (adaptedmaxHeight < nextYPosition-ySpacing+bottomBorder)
					adaptedmaxHeight = nextYPosition-ySpacing+bottomBorder;

				nextXPosition = nextXPosition+maxNeededWidth+xSpacing;
				nextYPosition = topBorder;

				newitem->x = nextXPosition;
				newitem->y = nextYPosition;

				menuBreaks++;

				nextYPosition = nextYPosition+newitem->height+ySpacing+separatorySpacing;
			}
			else
			{
				if (adaptedmaxHeight < nextYPosition-ySpacing+bottomBorder)
					adaptedmaxHeight = nextYPosition-ySpacing+bottomBorder;
							
				nextYPosition = topBorder;

				newitem->y = nextYPosition;

				hiddenPart++;
							
				nextYPosition = nextYPosition+newitem->height+ySpacing+separatorySpacing;
			}
		}
		else
		{
			nextYPosition = nextYPosition+newitem->height+ySpacing+separatorySpacing;
		}
						
		newitem->hiddenPart = hiddenPart;
		newitem->multiPart = menuBreaks;

		entrylist.insert(entrylist.end(), newitem);
	}

	//Check Max/Min EntryWidth
	maxNeededWidth = max(min(maxNeededWidth, maxEntryWidth), minEntryWidth);

	//Measure with Type 1 only check TitleFont Settings for maxNeededWidth
	int tmp = ComputeEntryWidth(captionname, true);
	int neededTitleWidth = leftBorder+tmp+rightBorder;

	//Adapt EntryWidth based on Max/Min Width
	if ( leftBorder+(menuBreaks+1)*maxNeededWidth+menuBreaks*xSpacing+rightBorder >= maxWidth )
	{
		actualwidth = maxWidth;
		maxNeededWidth = (actualwidth-leftBorder-rightBorder-menuBreaks*xSpacing)/(menuBreaks+1);
	}
	else if (leftBorder+(menuBreaks+1)*maxNeededWidth+menuBreaks*xSpacing+rightBorder <= minWidth )
	{
		actualwidth = minWidth;
		maxNeededWidth = (actualwidth-leftBorder-rightBorder-menuBreaks*xSpacing)/(menuBreaks+1);
	}
	else
	{
		actualwidth = leftBorder+(menuBreaks+1)*maxNeededWidth+menuBreaks*xSpacing+rightBorder;
		actualwidth = min(max(neededTitleWidth, actualwidth), maxWidth);
	}

	if (hiddenPart != 0 || menuBreaks != 0)
		actualheight = adaptedmaxHeight;
	else
		actualheight = min(nextYPosition-ySpacing+bottomBorder, maxHeight);
	actualheight = max(actualheight, minHeight);

	startentry = NULL;

	for (eit = entrylist.begin(); eit != entrylist.end(); ++eit)
	{
		if (startentry == NULL && (*eit)->entryType <= 0)
			startentry = *eit;

		if (menuBreaks == 0)
		{
			(*eit)->width = max(maxNeededWidth, actualwidth-leftBorder-rightBorder);
		}
		else
		{
			(*eit)->width = maxNeededWidth;
			
			(*eit)->x = leftBorder + (*eit)->multiPart*(maxNeededWidth + xSpacing);
		}
	}

	iMaxParts = hiddenPart;
	iMaxMenuBreaks = menuBreaks;

	reposition(2, 0, 0, actualwidth, actualheight);
	repaint(true);

	if ((isTaskPopup || hasDynamicContent) && !hasInfoEntry)
		SetTimer(hWnd, TIMER_UPDATE, tasksUpdate, NULL);
	else if (hasInfoEntry)
		SetTimer(hWnd, TIMER_UPDATE, 1000, NULL);
}

BOOL WINAPI Popup::EnumWindowsProc(HWND hWnd, LPARAM lParam)
{
	Popup *popup = (Popup *) lParam;
	if (popup && IsAppWindow(hWnd))
	{
		popup->tasklist.insert(popup->tasklist.end(), hWnd);
	}
	return TRUE;
}

void Popup::CreateSubMenu(string newsubname, string newrootpath, ENTRY* pNewParentEntry, bool jumpTo, bool noShow)
{
	CString help = newsubname.c_str();
	help.Replace("sub_", "");

	for(PopupListIterator plit = subpopupList.begin(); plit != subpopupList.end(); plit++)
	{
		if (_stricmp(help, (*plit)->captionname.c_str()) == NULL && _stricmp(newrootpath.c_str(), (*plit)->rootScanPath.c_str()) == NULL)
			return;
	}

	Popup *popup = NULL;

	if ( customSubFolderList.empty() && pNewParentEntry->layout.empty() )
		popup = new Popup(strName, newsubname, pPaintClasses, plistGhostPopup, plistGhostButton, subFolderLevel);
	else if ( !pNewParentEntry->layout.empty() )
		popup = new Popup(pNewParentEntry->layout, newsubname, NULL, NULL, NULL, subFolderLevel, (pNewParentEntry->defaultlayout.empty() ? strName : pNewParentEntry->defaultlayout));
	else if ( subFolderLevel < customSubFolderList.size() )
		popup = new Popup(customSubFolderList[subFolderLevel], newsubname, NULL, NULL, NULL, subFolderLevel, strName);
	else
		popup = new Popup(*(customSubFolderList.end()-1), newsubname, NULL, NULL, NULL, subFolderLevel, strName);
			
	subpopupList.insert(subpopupList.end(), popup);
	
	popup->pParentPopup = this;
	popup->pParentEntry = pNewParentEntry;
			
	popup->load(hInstance);

	if ( !noShow )
		popup->CreateLoadingPopup();
	
	if (!newrootpath.empty())
	{
		popup->rootScanPath = newrootpath;
		popup->rootUniversalAction = pNewParentEntry->universalaction;
		
		popup->ScanSubMenu(newrootpath, pNewParentEntry->entryType, pNewParentEntry->universalaction);

		if ( pNewParentEntry->entryType == -2 )
		{
			popup->dynamicPopup = true;

			bool newIteration = true;
			CString localworker1 = popup->rootScanPath.c_str();
			CString localworker2 = "";

			while (newIteration == true)
			{
				if (localworker1.Find("|") != -1)
				{
					if ( subFolderLevel <= 1 )
					{
						CNotifyDirCheck *checkDirectory = new CNotifyDirCheck();
					
						popup->checkdirectorylist.insert(popup->checkdirectorylist.end(), checkDirectory);
						
						localworker2 = localworker1.Left(localworker1.Find("|"));

						if ( localworker2.Find("*") == -1 )
							checkDirectory->SetDirectory( localworker2 );
						else
							checkDirectory->SetDirectory( localworker2.Left(localworker2.Find("*")) );

						//set your callback to work with each new event
						checkDirectory->SetPopup( popup );

						if ( subFolderLevel == 1 )
							checkDirectory->SetSubFolderChanges( TRUE );

						checkDirectory->Run();
					}

					localworker1 = localworker1.Right(localworker1.GetLength()-localworker1.Find("|")-1);
				}
				else
				{
					if ( subFolderLevel <= 1 )
					{
						CNotifyDirCheck *checkDirectory = new CNotifyDirCheck();
						
						popup->checkdirectorylist.insert(popup->checkdirectorylist.end(), checkDirectory);
							
						if ( localworker1.Find("*") == -1 )
							checkDirectory->SetDirectory( localworker1 );
						else
							checkDirectory->SetDirectory( localworker1.Left(localworker1.Find("*")) );

						//set your callback to work with each new event
						checkDirectory->SetPopup( popup );

						if ( subFolderLevel == 1 )
							checkDirectory->SetSubFolderChanges( TRUE );

						checkDirectory->Run();
					}

					newIteration = false;
				}
			}
		}
	}

	if ( !noShow )
	{	
		popup->show();

		if (jumpTo && popup->startentry != NULL)
			SetCursorPos(popup->actualx+popup->startentry->x+2, popup->actualy+popup->startentry->y+2);
	}
}

void Popup::CreateShellSubMenu( string newsubname, int newclsid, ENTRY* pNewParentEntry, bool jumpTo, bool noShow)
{
	CString help = newsubname.c_str();
	help.Replace("sub_", "");

	for(PopupListIterator plit = subpopupList.begin(); plit != subpopupList.end(); plit++)
	{
		if (_stricmp( help, (*plit)->captionname.c_str()) == NULL)
		{
			return;
		}
	}

	Popup *popup = NULL;
	
	if ( customSubFolderList.empty() && pNewParentEntry->layout.empty() )
		popup = new Popup(strName, newsubname, pPaintClasses, plistGhostPopup, plistGhostButton, subFolderLevel);
	else if ( !pNewParentEntry->layout.empty() )
		popup = new Popup(pNewParentEntry->layout, newsubname, NULL, NULL, NULL, subFolderLevel, (pNewParentEntry->defaultlayout.empty() ? strName : pNewParentEntry->defaultlayout));
	else if ( subFolderLevel < customSubFolderList.size() )
		popup = new Popup(customSubFolderList[subFolderLevel], newsubname, NULL, NULL, NULL, subFolderLevel, strName);
	else
		popup = new Popup(*(customSubFolderList.end()-1), newsubname, NULL, NULL, NULL, subFolderLevel, strName);
			
	subpopupList.insert(subpopupList.end(), popup);

	popup->pParentPopup = this;
	popup->pParentEntry = pNewParentEntry;
										
	popup->load(hInstance);

	if (!noShow)
		popup->CreateLoadingPopup();

	popup->ScanShellFolder(newclsid);

	if (!noShow)
	{
		popup->show();
	
		if (jumpTo && popup->startentry != NULL)
			SetCursorPos(popup->actualx+popup->startentry->x+2, popup->actualy+popup->startentry->y+2);
	}
}

void Popup::CreateTaskSubMenu( string newsubname, ENTRY* pNewParentEntry, bool jumpTo, bool noShow)
{
	Popup *popup = NULL;
	
	if ( customSubFolderList.empty() && pNewParentEntry->layout.empty() )
		popup = new Popup(strName, newsubname, pPaintClasses, plistGhostPopup, plistGhostButton, subFolderLevel);
	else if ( !pNewParentEntry->layout.empty() )
		popup = new Popup(pNewParentEntry->layout, newsubname, NULL, NULL, NULL, subFolderLevel, (pNewParentEntry->defaultlayout.empty() ? strName : pNewParentEntry->defaultlayout));
	else if ( subFolderLevel < customSubFolderList.size() )
		popup = new Popup(customSubFolderList[subFolderLevel], newsubname, NULL, NULL, NULL, subFolderLevel, strName);
	else
		popup = new Popup(*(customSubFolderList.end()-1), newsubname, NULL, NULL, NULL, subFolderLevel, strName);

	subpopupList.insert(subpopupList.end(), popup);

	popup->pParentPopup = this;
	popup->pParentEntry = pNewParentEntry;
										
	popup->load(hInstance);

	if (!noShow)
		popup->CreateLoadingPopup();

	popup->ScanTasks();

	if (!noShow)
	{
		popup->show();	
		
		if (jumpTo && popup->startentry != NULL)
			SetCursorPos(popup->actualx+popup->startentry->x+2, popup->actualy+popup->startentry->y+2);
	}
}

//////////////////////////////////////////////////////////////////////
// Window Related: Show/Hide/Reposition
//////////////////////////////////////////////////////////////////////

void Popup::reposition(int type, int newx, int newy, int newwidth, int newheight, int externCall)
{
	if ( externCall == 1 )
	{
		minWidth = newwidth;
		minHeight = newheight;
		maxWidth = newwidth;
		maxHeight = newheight;

		for(PopupListIterator plit = subpopupList.begin(); plit != subpopupList.end(); plit++)
		{
			(*plit)->hide(true/*noFade*/);
			delete *plit;
			*plit = NULL;
		}

		subpopupList.clear();

		for (EntryListIterator eit = entrylist.begin(); eit != entrylist.end(); ++eit)
		{
			delete *eit;
			*eit = NULL;
		}

		entrylist.clear();

		//Reset old Pointers
		//-------------------------
		iCurrentPart = 0;

		startentry = NULL;
		lastactiveentry = NULL;
		hoverentry = NULL;
		//-------------------------

		ParseStepRC(true);
	}
	else
	{
		if ( type == 1 )
		{
			SetWindowPos(hWnd, 0, newx, newy, 0, 0,
					SWP_NOACTIVATE | SWP_NOSIZE | SWP_NOZORDER);
		}
		else if ( type == 2 )
		{
			SetWindowPos(hWnd, 0, 0, 0, newwidth, newheight,
					SWP_NOACTIVATE | SWP_NOMOVE | SWP_NOZORDER);
		}
		else if ( type == 3 )
		{
			SetWindowPos(hWnd, 0, newx, newy, newwidth, newheight,
					SWP_NOACTIVATE | SWP_NOZORDER);
		}
		
		//Set new Rect for ToolTip Tracker
		if ( type != 1 && pTooltip && (iQuicklaunchMode != 0 || iTooltipMode != 0) )
			pTooltip->apply("", true);
	}
}

void Popup::show(bool setcursor, bool loading)
{
	if ( visible )
	{
		if ( hasDynamicContent )
			ParseStepRC(false);

		if ( pParentPopup != NULL && pParentEntry != NULL )
		{
			if (pParentEntry->hWnd != NULL && pParentPopup->hWnd == pParentEntry->hWnd)
				ScanTasks();
		}

		if ( setcursor && startentry != NULL )
		{
			SetCursorPos(actualx+startentry->x+2, actualy+startentry->y+2);
			hoverentry = startentry;
			repaint();
		}

		RepositionSubPopupMenus();

		if ( bPinnedPopupMode )
		{
			SetForegroundWindow(hWnd);
			SetActiveWindow(hWnd);
			SetFocus(hWnd);
		}

		return;
	}

	//Dynamic Popups
	if ( !loading )
	{
		if ( hasDynamicContent )
			ParseStepRC(false);

		if ( pParentPopup != NULL && pParentEntry != NULL )
		{
			if (pParentEntry->hWnd != NULL && pParentPopup->hWnd == pParentEntry->hWnd)
				ScanTasks();
		}
	}

	bool showit = false;

	//UnPin "SubPopupMenus", if they are called by Parent
	if ( bPinnedPopupMode && pParentPopup != NULL && pParentEntry != NULL )
	{
		bPinnedPopupMode = false;
		setAlwaysOnTop(true);

		showit = true;
	}

	//Check if Main Popup is visible
	Popup* tmpPointer = pParentPopup;
	if (tmpPointer != NULL && !showit)
	{
		while (tmpPointer->pParentPopup != NULL)
		{
			if ( tmpPointer->isVisible() && tmpPointer->bPinnedPopupMode )
			{
				showit = true;
				break;
			}
			tmpPointer = tmpPointer->pParentPopup;
		}
			
		if ( !tmpPointer->isVisible() )
			return;
		else
			showit = true;
	}
	else
	{
		showit = true;
	}

	if (showit)
	{
		RepositionSubPopupMenus();

		if ( hasInfoEntry )
		{
			for (EntryListIterator eit = entrylist.begin(); eit != entrylist.end(); ++eit)
			{
				if ( !(*eit)->info.empty() )
					(*eit)->caption = InfoUpdate( (*eit)->info );
			}

			repaint(false, true);
		}

		if (alphaFade)
		{
			UpdateAlpha = 0;
			UpdateWindowTransparent(hWnd, UpdateAlpha);
			ShowWindow(hWnd, SW_SHOW);
			RedrawWindow(hWnd, NULL, NULL, RDW_INVALIDATE | RDW_UPDATENOW);	
				
			KillTimer(hWnd, TIMER_ALPHAFADEOUT);
			SetTimer(hWnd, TIMER_ALPHAFADEIN, alphaFadeDelay, NULL);
		}
		else if (alphaMap || fakealphaMap)
		{
			ShowWindow(hWnd, SW_SHOW);
			RedrawWindow(hWnd, NULL, NULL, RDW_INVALIDATE | RDW_UPDATENOW );
			UpdateWindowTransparent(hWnd, alphaTransparency);
		} 
		else
		{
			ShowWindow(hWnd, SW_SHOW);
			RedrawWindow(hWnd, NULL, NULL, RDW_INVALIDATE | RDW_UPDATENOW );
		}

		//Only Main Popups
		if (pParentPopup == NULL)
			SetTimer(hWnd, TIMER_ACTIVE, hideDelay, NULL);

		visible = true;

		listVisiblePopups.insert(listVisiblePopups.end(), this);

		if ( setcursor && startentry != NULL)
			SetCursorPos(actualx+startentry->x+2, actualy+startentry->y+2);

		if ( !BangToCall.empty() )
		{
			if ( !openSound.empty() )
				ParseBangCommand(hWnd, "!playsound", openSound.c_str());

			if ( !openCommand.empty() )
			{
				CString result = openCommand.c_str();

				CString worker = BangToCall.c_str();
				worker.Replace("!", "");
				
				result.Replace("%[popupname]%", worker);
				result.Replace("%{popupname}%", worker);

				LSExecute(hWnd, result, SW_SHOWNORMAL);
			}

			SetForegroundWindow(hWnd);
			SetActiveWindow(hWnd);
			SetFocus(hWnd);
		}
	}

	if ( !bPinnedPopupMode || (bPinnedPopupMode && !bPinnedPopupNotOnTop) )
		SetWindowPos(hWnd, HWND_TOPMOST, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE);

	//Make sure Parent Popup Tooltips remain visible
	if ( pTooltip && !loading )
	{
		Popup* tmpPointer = pParentPopup;
		if ( tmpPointer != NULL )
		{
			if ( tmpPointer->hWndToolTip )			
				SetWindowPos( tmpPointer->hWndToolTip, HWND_TOPMOST, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE | SWP_NOACTIVATE ); 

			while (tmpPointer->pParentPopup != NULL)
			{
				tmpPointer = tmpPointer->pParentPopup;

				if ( tmpPointer->hWndToolTip )			
					SetWindowPos( tmpPointer->hWndToolTip, HWND_TOPMOST, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE | SWP_NOACTIVATE ); 
			}
		}
	}
}

void Popup::RepositionSubPopupMenus()
{
	//Reposition "SubPopupMenus"
	if (pParentPopup != NULL && pParentEntry != NULL)
	{
		int x = 0;
		int y = 0;

		//QuickLaunch Mode
		if (iQuicklaunchMode != 0)
		{
			if (iQuicklaunchMode == 1 || iQuicklaunchMode == -1)
			{
				x = pParentPopup->actualx + pParentEntry->x + overlapX;
					
				if (quicklaunchOpenTo == 1)
					y = pParentPopup->actualy - actualheight + overlapY;
				else if (quicklaunchOpenTo == 2)
					y = pParentPopup->actualy + pParentPopup->actualheight + overlapY;
			}
			else if (iQuicklaunchMode == 2 || iQuicklaunchMode == -2)
			{
				if (quicklaunchOpenTo == 1)
					x = pParentPopup->actualx + pParentPopup->actualwidth + overlapX;
				else if (quicklaunchOpenTo == 2)
					x = pParentPopup->actualx - actualwidth + overlapX;

				y = pParentPopup->actualy + pParentEntry->y + overlapY;
			}

			//InScreen Check
			if ( iRealMonitors > 1 )
			{
				HMONITOR hMonitor = MonitorFromWindow(pParentPopup->gethWnd(), MONITOR_DEFAULTTONEAREST);
				
				MONITORINFO mi = { sizeof(MONITORINFO) };
				GetMonitorInfo(hMonitor, &mi);

				if (x+actualwidth > mi.rcMonitor.right)
					x = pParentPopup->actualx-actualwidth-overlapX;

				if (y+actualheight > mi.rcMonitor.bottom)
					y = mi.rcMonitor.bottom-actualheight;

				//Minimum x, y is screenX/Y
				if ( x < mi.rcMonitor.left )
					x = mi.rcMonitor.left;
				if ( y < mi.rcMonitor.top )
					y = mi.rcMonitor.top;
			}
			else
			{
				if (x+actualwidth > iScreenWidth)
					x = pParentPopup->actualx-actualwidth-overlapX;

				if (y+actualheight > iScreenHeight)
					y = iScreenHeight-actualheight;

				//Minimum x, y is screenX/Y
				if ( x < iScreenX )
					x = iScreenX;
				if ( y < iScreenY )
					y = iScreenY;
			}

			reposition(1, x, y, 0, 0);
		}
		//Normal PopupMenus
		else
		{
			if ( bOpenToTheLeft )
			{
				if ( pParentEntry->multiPart == 0 )
					x = pParentPopup->actualx - actualwidth - overlapX;
				else
					x = pParentPopup->actualx + pParentEntry->x - actualwidth - pParentPopup->xSpacing/2 - overlapX;
							
				y = pParentPopup->actualy + pParentEntry->y + overlapY;

				//InScreen Check
				if ( iRealMonitors > 1 )
				{
					HMONITOR hMonitor = MonitorFromWindow(pParentPopup->gethWnd(), MONITOR_DEFAULTTONEAREST);
				
					MONITORINFO mi = { sizeof(MONITORINFO) };
					GetMonitorInfo(hMonitor, &mi);

					if ( x < mi.rcMonitor.right && pParentEntry->multiPart == 0 && pParentPopup->iMaxMenuBreaks != 0 )
						x = pParentPopup->actualx + pParentEntry->x + pParentEntry->width + pParentPopup->xSpacing/2 + overlapX;
					else if ( x < mi.rcMonitor.right )
						x = pParentPopup->actualx + pParentPopup->actualwidth + overlapX;

					if (y+actualheight > mi.rcMonitor.bottom)
						y = mi.rcMonitor.bottom-actualheight;

					//Maximum x is screenWidth, Minimum y is screenY
					if ( x > mi.rcMonitor.right )
						x = mi.rcMonitor.right-actualwidth;
					if ( y < mi.rcMonitor.top )
						y = mi.rcMonitor.top;
				}
				else
				{
					if ( x < iScreenX && pParentEntry->multiPart == 0 && pParentPopup->iMaxMenuBreaks != 0 )
						x = pParentPopup->actualx + pParentEntry->x + pParentEntry->width + pParentPopup->xSpacing/2 + overlapX;
					else if ( x < iScreenX )
						x = pParentPopup->actualx + pParentPopup->actualwidth + overlapX;

					if (y+actualheight > iScreenHeight)
						y = iScreenHeight-actualheight;

					//Maximum x is screenWidth, Minimum y is screenY
					if ( x > iScreenWidth )
						x = iScreenWidth-actualwidth;
					if ( y < iScreenY )
						y = iScreenY;
				}
			}
			else
			{
				if ( pParentEntry->multiPart != pParentPopup->iMaxMenuBreaks && pParentPopup->iMaxMenuBreaks != 0 )
					x = pParentPopup->actualx + pParentEntry->x + pParentEntry->width + pParentPopup->xSpacing/2 + overlapX;
				else
					x = pParentPopup->actualx + pParentPopup->actualwidth + overlapX;
			
				y = pParentPopup->actualy + pParentEntry->y + overlapY;

				//InScreen Check
				if ( iRealMonitors > 1 )
				{
					HMONITOR hMonitor = MonitorFromWindow(pParentPopup->gethWnd(), MONITOR_DEFAULTTONEAREST);
				
					MONITORINFO mi = { sizeof(MONITORINFO) };
					GetMonitorInfo(hMonitor, &mi);

					if ( x+actualwidth > mi.rcMonitor.right && pParentEntry->multiPart == 0 )
						x = pParentPopup->actualx - actualwidth - overlapX;
					else if ( x+actualwidth > mi.rcMonitor.right )
						x = pParentPopup->actualx + pParentEntry->x - actualwidth - pParentPopup->xSpacing/2 - overlapX;

					if (y+actualheight > mi.rcMonitor.bottom)
						y = mi.rcMonitor.bottom-actualheight;

					//Minimum x, y is screenX/Y
					if ( x < mi.rcMonitor.left )
						x = mi.rcMonitor.left;
					if ( y < mi.rcMonitor.top )
						y = mi.rcMonitor.top;
				}
				else
				{
					if ( x+actualwidth > iScreenWidth && pParentEntry->multiPart == 0 )
						x = pParentPopup->actualx - actualwidth - overlapX;
					else if ( x+actualwidth > iScreenWidth )
						x = pParentPopup->actualx + pParentEntry->x - actualwidth - pParentPopup->xSpacing/2 - overlapX;

					if (y+actualheight > iScreenHeight)
						y = iScreenHeight-actualheight;

					//Minimum x, y is screenX/Y
					if ( x < iScreenX )
						x = iScreenX;
					if ( y < iScreenY )
						y = iScreenY;
				}
			}

			reposition(1, x, y, 0, 0);
		}
	}
}

void Popup::hide(bool noFade)
{
	bool wasVisible = visible;

	if ( !bPinnedPopupMode && visible )
	{
		if (alphaFade && !noFade)
		{
			UpdateAlpha = alphaTransparency;

			KillTimer(hWnd, TIMER_ALPHAFADEIN);
			SetTimer(hWnd, TIMER_ALPHAFADEOUT, alphaFadeDelay, NULL);
		}
		else if ( !alphaFade || noFade )
		{
			KillTimer(hWnd, TIMER_ALPHAFADEIN);
			KillTimer(hWnd, TIMER_ALPHAFADEOUT);
			ShowWindow(hWnd, SW_HIDE);

			if ( wasVisible && !BangToCall.empty() && bExportHoverState )
			{
				PrefixedLSSetVariable(strName.c_str(), "ExportedHoverCaption", "");
				PrefixedLSSetVariable(strName.c_str(), "ExportedHoverAction", "");
			}
		
			lastactiveentry = NULL;
			hoverentry = NULL;
			lastkeyentry = -1;
			
			visible = false;
		}

		for(PopupListIterator plit = listVisiblePopups.begin(); plit != listVisiblePopups.end(); plit++)
		{
			if ( (*plit) == this )
			{
				listVisiblePopups.erase( plit );
				break;
			}
		}
		
		KillTimer(hWnd, TIMER_ACTIVE);
	
		for(plit = subpopupList.begin(); plit != subpopupList.end(); plit++)
		{
			if ( !(*plit)->bPinnedPopupMode && (*plit)->isVisible() )
				(*plit)->hide(noFade);
		}
		
		if ( !BangToCall.empty() )
		{	
			if ( !closeSound.empty() && !execSoundPlayed )
				ParseBangCommand(hWnd, "!playsound", closeSound.c_str());

			execSoundPlayed = false;

			if ( !closeCommand.empty() && wasVisible )
			{
				CString result = closeCommand.c_str();

				CString worker = BangToCall.c_str();
				worker.Replace("!", "");
				
				result.Replace("%[popupname]%", worker);
				result.Replace("%{popupname}%", worker);

				LSExecute(hWnd, result, SW_SHOWNORMAL);
			}
		}
	}
	else if ( bPinnedPopupMode )
	{
		for(PopupListIterator plit = subpopupList.begin(); plit != subpopupList.end(); plit++)
		{
			if ( !(*plit)->bPinnedPopupMode && (*plit)->isVisible() )
				(*plit)->hide(noFade);
		}
	}

	//Make sure Parent Popup Tooltips remain visible
	if ( pTooltip )
	{
		Popup* tmpPointer = pParentPopup;
		if ( tmpPointer != NULL )
		{
			if ( tmpPointer->hWndToolTip )			
				SetWindowPos( tmpPointer->hWndToolTip, HWND_TOPMOST, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE | SWP_NOACTIVATE ); 

			while (tmpPointer->pParentPopup != NULL)
			{
				tmpPointer = tmpPointer->pParentPopup;

				if ( tmpPointer->hWndToolTip )			
					SetWindowPos( tmpPointer->hWndToolTip, HWND_TOPMOST, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE | SWP_NOACTIVATE ); 
			}
		}
	}
}

void Popup::ToggleShade()
{
	if (shaded != -1)
	{
		if ( !BangToCall.empty() && !unshadeCommand.empty() )
		{
			CString result = unshadeCommand.c_str();

			CString worker = BangToCall.c_str();
			worker.Replace("!", "");
				
			result.Replace("%[popupname]%", worker);
			result.Replace("%{popupname}%", worker);

			LSExecute(hWnd, result, SW_SHOWNORMAL);
		}

		int tmpHeight = shaded;
		shaded = -1;
		reposition(2, 0, 0, actualwidth, tmpHeight);
		repaint(true);
	}
	else
	{
		if ( !BangToCall.empty() && !shadeCommand.empty() )
		{
			CString result = shadeCommand.c_str();

			CString worker = BangToCall.c_str();
			worker.Replace("!", "");
				
			result.Replace("%[popupname]%", worker);
			result.Replace("%{popupname}%", worker);

			LSExecute(hWnd, result, SW_SHOWNORMAL);
		}

		shaded = actualheight;
		reposition(2, 0, 0, actualwidth, max(/*max(*/topBorder+bottomBorder/*, popuptexture->getMinNeededHeight())*/, orgminHeight));
		repaint(true);
	}
}

//------------------------------------------------------------
//Mouse Related Actions (Clicks, Enter, Leave, Move, ...)
//------------------------------------------------------------

void Popup::ontemplateButtonDown(const int &x, const int &y, int titleActionType, int bottomActionType)
{
	if ( (titleActionType == 6 && MoveAllowed && y >= 0 && y <= topBorder)
			|| (bottomActionType == 6 && MoveAllowed && y >= actualheight-bottomBorder && y <= actualheight) )
	{
		mousePressed = true;
		GetCursorPos(&savedpt);
		ScreenToClient(hWnd, &savedpt);
		bMoveButtonDown = true;
		SetCapture(hWnd);
	}
	else
	{
		if ( dragndropEnabled )
			ActionAtPoint( x, y, false, false, 1 );

		SetCapture(hWnd);
	}
}

void Popup::ontemplateButtonUp(const int &x, const int &y, int titleActionType, int bottomActionType, int mousebutton)
{
	bool normalactions = true;
	bool somethingdone = false;
	if ( ((titleActionType == 1 || titleActionType == 2 || titleActionType == 7) && y >= 0 && y <= topBorder )
			|| ((bottomActionType == 1 || bottomActionType == 2 || bottomActionType == 7) && y >= actualheight-bottomBorder && y <= actualheight) )
	{
		normalactions = false;
		somethingdone = true;
		if ( bPinnedPopupMode && (titleActionType == 2 || titleActionType == 7) && y >= 0 && y <= topBorder )
		{
			bPinnedPopupMode = false;
			setAlwaysOnTop(true);
		}
		if ( bPinnedPopupMode && (bottomActionType == 2 || bottomActionType == 7) && y >= actualheight-bottomBorder && y <= actualheight )
		{
			bPinnedPopupMode = false;
			setAlwaysOnTop(true);
		}
		if (!bPinnedPopupMode && titleActionType == 1 && y >= 0 && y <= topBorder )
		{
			bPinnedPopupMode = true;
			if ( bPinnedPopupNotOnTop )
				setAlwaysOnTop(false);
			else
				setAlwaysOnTop(true);
		}
		if (!bPinnedPopupMode && bottomActionType == 1 && y >= actualheight-bottomBorder && y <= actualheight )
		{
			bPinnedPopupMode = true;
			if ( bPinnedPopupNotOnTop )
				setAlwaysOnTop(false);
			else
				setAlwaysOnTop(true);
		}
		if ( ((titleActionType == 2 || titleActionType == 7) && y >= 0 && y <= topBorder)
			|| ((bottomActionType == 2 || bottomActionType == 7) && y >= actualheight-bottomBorder && y <= actualheight) )
			hide();
	}
	else if ( (titleActionType == 3 && y >= 0 && y <= topBorder) 
				|| (bottomActionType == 3 && y >= actualheight-bottomBorder && y <= actualheight) )
	{
		normalactions = false;
		somethingdone = true;
		if ( bPinnedPopupMode )
		{
			bPinnedPopupMode = false;
			setAlwaysOnTop(true);
			hide();
		}
		else if ( !bPinnedPopupMode )
		{
			bPinnedPopupMode = true;
			if ( bPinnedPopupNotOnTop )
				setAlwaysOnTop(false);
			else
				setAlwaysOnTop(true);
		}
	}
	else if ( (titleActionType == 4 && y >= 0 && y <= topBorder) 
				|| (bottomActionType == 4 && y >= actualheight-bottomBorder && y <= actualheight) )
	{
		normalactions = false;
		if ( ShadeAllowed && !(0x8000 & GetKeyState(VK_CONTROL)) && !(0x8000 & GetKeyState(VK_SHIFT)) )
		{
			ToggleShade();
			somethingdone = true;
		}
	}
	else if ( (titleActionType == 5 && y >= 0 && y <= topBorder) 
				|| (bottomActionType == 5 && y >= actualheight-bottomBorder && y <= actualheight) )
	{
		normalactions = false;
		if ( bPinnedPopupMode )
		{
			somethingdone = true;
			setAlwaysOnTop(!alwaysOnTop);
		}
	}
	else if ( titleActionType == 6 || bottomActionType == 6 )
	{
		if ( bMoveButtonDown )
		{
			somethingdone = true;
			normalactions = false;
			if ( MoveAllowed && mousePressed && mouseMoved && pParentPopup != NULL )
			{
				if ( AutoPin && !bPinnedPopupMode )
				{
					bPinnedPopupMode = true;
					if (bPinnedPopupNotOnTop)
						setAlwaysOnTop(false);
					else
						setAlwaysOnTop(true);
					SetTimer(hWnd, TIMER_ACTIVE, hideDelay, NULL);
				}	
			}
			mouseMoved = false;

			//To be on the save side
			mousePressed = false;

			bMoveButtonDown = false;
		}
	}
	else if ( (titleActionType == 8 && y >= 0 && y <= topBorder) 
			|| (bottomActionType == 8 && y >= actualheight-bottomBorder && y <= actualheight) )
	{
		normalactions = false;
		if ( iMaxParts != 0)
		{
			iCurrentPart--;
			if (iCurrentPart < 0)
				iCurrentPart = iMaxParts;

			oldmovetracker.x = 0;

			repaint(true);
		}
		somethingdone = true;
	}
	else if ( (titleActionType == 9 && y >= 0 && y <= topBorder) 
			|| (bottomActionType == 9 && y >= actualheight-bottomBorder && y <= actualheight) )
	{
		normalactions = false;
		if ( iMaxParts != 0)
		{
			iCurrentPart++;
			if (iCurrentPart > iMaxParts)
				iCurrentPart = 0;

			oldmovetracker.x = 0;

			repaint(true);
		}
		somethingdone = true;
	}
	else if ( (titleActionType == 10 && y >= 0 && y <= topBorder) )
	{
		normalactions = false;

		if ( mousebutton == 1 )
			LSExecute(NULL, strtitleLClickAction.c_str(), NULL);
		else if ( mousebutton == 2 )
			LSExecute(NULL, strtitleRClickAction.c_str(), NULL);
		else if ( mousebutton == 3 )
			LSExecute(NULL, strtitleMClickAction.c_str(), NULL);
		
		somethingdone = true;
	}
	else if ( (bottomActionType == 10 && y >= actualheight-bottomBorder && y <= actualheight) )
	{
		normalactions = false;

		if ( mousebutton == 1 )
			LSExecute(NULL, strbottomLClickAction.c_str(), NULL);
		else if ( mousebutton == 2 )
			LSExecute(NULL, strbottomRClickAction.c_str(), NULL);
		else if ( mousebutton == 3 )
			LSExecute(NULL, strbottomMClickAction.c_str(), NULL);
		
		somethingdone = true;
	}

	if ( normalactions && shaded == -1 )
	{
		if ( mousebutton == 1 )
		{	
			if ( dragndropEnabled )
			{
				ActionAtPoint( x, y, false, false, 2 );
		
				if ( _stricmp(dragndropSource.c_str(), "-1") != NULL && !dragndropDestination.empty() )
				{
					string destination = "";
					destination.append(dragndropDestination);

					//Filter FolderClicks
					if ( _stricmp(dragndropSource.c_str(), destination.c_str()) != NULL )
					{	
						CString worker = dragndropSource.c_str();
						worker = worker.Right(worker.GetLength()-worker.ReverseFind('\\'));
						
						destination.append( worker );

						if ( _stricmp(dragndropSource.c_str(), destination.c_str()) != NULL && StrRStrI( destination.c_str(), NULL, dragndropSource.c_str() ) == NULL )
						{
							ReleaseCapture();

							if ( MessageBox(NULL, ("Are you sure you want to move the following file/folder?\n\n" + dragndropSource + "\nto:\n" + destination ).c_str(), "User Confirmation", MB_YESNO | MB_ICONQUESTION | MB_SETFOREGROUND) == IDYES )
								MoveFile( dragndropSource.c_str(), destination.c_str() );
						}
						else
						{
							//ActionAtPoint: First Bool is Normal Action, Second is ContextMenu
							ActionAtPoint(x, y, true);
						}
					}
					else
					{
						//ActionAtPoint: First Bool is Normal Action, Second is ContextMenu
						ActionAtPoint(x, y, true);
					}
				}
				else
				{
					//ActionAtPoint: First Bool is Normal Action, Second is ContextMenu
					ActionAtPoint(x, y, true);
				}
			}
			else
			{
				//ActionAtPoint: First Bool is Normal Action, Second is ContextMenu
				ActionAtPoint(x, y, true);
			}
		}
		else if ( mousebutton == 2 )
		{
			//ActionAtPoint: First Bool is Normal Action, Second is ContextMenu
			ActionAtPoint(x, y, false, true);
		}
	}
	if ( !somethingdone && shaded == -1 && mousebutton == 3 && (x >= 4 || y >= 4) )
	{
		//Bool Value: True = Upwards, False = downwards
		repaint(SelectNextEntry(false));
	}

	ReleaseCapture();
}

void Popup::onLButtonDblClk(int x, int y)
{
	if ( ((showCaption == 1 || showCaption == 0) && HideAllowed && y >= 0 && y <= topBorder)
		|| ((showCaption == 2 || showCaption == 0) && HideAllowed && y >= actualheight-bottomBorder && y <= actualheight) )
	{
		if (bPinnedPopupMode)
		{
			bPinnedPopupMode = false;
			setAlwaysOnTop(true);
			hide();
		}
		else
		{
			hide();
		}
	}
	else 
	{
		for (EntryListIterator eit = entrylist.begin(); eit != entrylist.end(); ++eit)
		{
			if (iCurrentPart != (*eit)->hiddenPart || (*eit)->entryType >= 0)
				continue;

			if ( x < ((*eit)->x + (*eit)->width) && x >= (*eit)->x &&
				 y < ((*eit)->y + (*eit)->height) && y >= (*eit)->y )
			{
				for(PopupListIterator plit = subpopupList.begin(); plit != subpopupList.end(); plit++)
				{
					if (_stricmp( (*eit)->caption.c_str(), (*plit)->captionname.c_str()) == NULL)
					{
						if ( !(*eit)->action.empty() )
						{
							if ((*eit)->pipedList.size() == 1)
							{
								Popup *tmpPointer = pParentPopup;
								if (tmpPointer != NULL)
								{
									while (tmpPointer->pParentPopup != NULL)
										tmpPointer = tmpPointer->pParentPopup;
									tmpPointer->hide(true);
								}
								else
								{
									hide(true);
								}
								
								LSExecute(NULL, (folderDoubleClickAction + " \"" + (*eit)->action + "\"").c_str(), NULL);
							}
							else
							{
								HMENU GroupMenu = CreatePopupMenu();

								int counter = 1;
								for(StringListIterator sit = (*eit)->pipedList.begin(); sit != (*eit)->pipedList.end(); ++sit)
								{
									AppendMenu( GroupMenu, MF_STRING, counter, (*sit).c_str() );
									counter++;
								}

								// display the menu
								POINT pt;
								GetCursorPos( &pt );

								int command = TrackPopupMenu( GroupMenu,
															  TPM_RETURNCMD | TPM_RIGHTBUTTON,
															  pt.x, pt.y, 0,
															  this->gethWnd(),
															  NULL );

								if ( command )
								{
									StringListIterator sit = (*eit)->pipedList.begin();
									counter = 1; 
									while (counter < command)
									{
										sit++;
										counter++;
									}

									if ( counter <= (*eit)->pipedList.size() )
									{
										Popup *tmpPointer = pParentPopup;
										if (tmpPointer != NULL)
										{
											while (tmpPointer->pParentPopup != NULL)
												tmpPointer = tmpPointer->pParentPopup;
											tmpPointer->hide(true);
										}
										else
											hide(true);

										LSExecute(NULL, (folderDoubleClickAction + " \"" + (*sit) + "\"").c_str(), NULL);
									}
								}
											
								if (GroupMenu)
									DestroyMenu(GroupMenu);
							}
						}
						break;
					}
				}
			}
		}
	}
}

void Popup::onWheelDown(int x, int y)
{
	if ( (0x8000 & GetKeyState(VK_CONTROL)) )
	{
		if ( iMaxParts != 0)
		{
			iCurrentPart++;
			if (iCurrentPart > iMaxParts)
				iCurrentPart = 0;

			oldmovetracker.x = 0;

			repaint(true);
		}
	}
	else
	{
		//Bool Value: True = Upwards, False = downwards
		repaint(SelectNextEntry(false));
	}
}

void Popup::onWheelUp(int x, int y)
{
	if ( (0x8000 & GetKeyState(VK_CONTROL)) )
	{
		if ( iMaxParts != 0)
		{
			iCurrentPart--;
			if (iCurrentPart < 0)
				iCurrentPart = iMaxParts;

			oldmovetracker.x = 0;

			repaint(true);
		}
	}
	else
	{
		//Bool Value: True = Upwards, False = downwards
		repaint(SelectNextEntry(true));
	}
}

void Popup::onKey(int nKey)
{
	switch (nKey)
	{
		case VK_F5:
		{
			if ( !BangToCall.empty() && pParentPopup == NULL )
			{
				ParseBangCommand(hWnd, "!popuprescan", "");
			}
			else if ( pParentPopup != NULL )
			{
				for(PopupListIterator plit = pParentPopup->subpopupList.begin(); plit != pParentPopup->subpopupList.end(); plit++)
				{
					if ( (*plit) == this )
					{
						pParentPopup->subpopupList.erase( plit );
						(*plit)->hide(true/*noFade*/);
						delete *plit;
						*plit = NULL;
						break;
					}
				}
			}
		}
		break;

		case VK_PRIOR:
		{
			if ( iMaxParts != 0)
			{
				iCurrentPart--;
				if (iCurrentPart < 0)
					iCurrentPart = iMaxParts;

				oldmovetracker.x = 0;

				repaint(true);
			}
		}
		break;

		case VK_NEXT:
		{
			if ( iMaxParts != 0)
			{
				iCurrentPart++;
				if (iCurrentPart > iMaxParts)
					iCurrentPart = 0;

				oldmovetracker.x = 0;

				repaint(true);
			}
		}
		break;

		case VK_RIGHT:
		{
			for (EntryListIterator eit = entrylist.begin(); eit != entrylist.end(); ++eit)
			{
				if ( (*eit) == hoverentry && (*eit)->entryType < 0 )
				{
					lastactiveentry = (*eit);
					bool alreadyCreated = false;
					for(PopupListIterator plit = subpopupList.begin(); plit != subpopupList.end(); plit++)
					{
						if (_stricmp((*eit)->caption.c_str(), (*plit)->captionname.c_str()) == NULL && _stricmp((*eit)->action.c_str(), (*plit)->rootScanPath.c_str()) == NULL)
						{
							(*plit)->hoverentry = (*plit)->startentry;
							(*plit)->show(true);
							alreadyCreated = true;
						}
						else
						{
							if ( !(*plit)->bPinnedPopupMode && (*plit)->isVisible() )
								(*plit)->hide();
						}
					}
					if (!alreadyCreated && (*eit)->clsid == 0 && (*eit)->hWnd == NULL)
						CreateSubMenu( ("sub_"+(*eit)->caption), (*eit)->action, (*eit), true);
					else if (!alreadyCreated && (*eit)->clsid != 0)
						CreateShellSubMenu( ("sub_"+(*eit)->caption), (*eit)->clsid, (*eit), true);
					else if (!alreadyCreated && (*eit)->hWnd == hWnd)
						CreateTaskSubMenu( ("sub_"+(*eit)->caption), (*eit), true);
				
					break;
				}
			}
		}
		break;

		case VK_LEFT:
		{
			if (pParentPopup != NULL)
			{
				if ( pParentPopup->lastactiveentry != NULL )
				{
					pParentPopup->hoverentry = pParentPopup->lastactiveentry;
					SetCursorPos(pParentPopup->actualx+pParentPopup->lastactiveentry->x+2, pParentPopup->actualy+pParentPopup->lastactiveentry->y+2);	
				}
				else if ( pParentPopup->startentry != NULL )
				{
					pParentPopup->hoverentry = pParentPopup->startentry;
					SetCursorPos(pParentPopup->actualx+pParentPopup->startentry->x+2, pParentPopup->actualy+pParentPopup->startentry->y+2);	
				}

				if ( !bPinnedPopupMode && visible )
					hide();

				pParentPopup->repaint();

				SetForegroundWindow(pParentPopup->gethWnd());
				SetActiveWindow(pParentPopup->gethWnd());
				SetFocus(pParentPopup->gethWnd());
			}
		}
		break;

		case VK_ESCAPE:
		{
			if ( HideAllowed )
			{
				//Search for Main Popup and Hide All SubPopups
				Popup* tmpPopupPointer = pParentPopup;
				if (tmpPopupPointer != NULL)
				{
					while (tmpPopupPointer->pParentPopup != NULL)
						tmpPopupPointer = tmpPopupPointer->pParentPopup;

					if ( !tmpPopupPointer->bPinnedPopupMode && tmpPopupPointer->isVisible() )
						tmpPopupPointer->hide();
				}
				else
				{
					if ( !bPinnedPopupMode && visible )
						hide();
				}
			}
		}
		break;

		case VK_UP:
		{
			for(PopupListIterator plit = subpopupList.begin(); plit != subpopupList.end(); plit++)
			{
				if ( !(*plit)->bPinnedPopupMode && (*plit)->isVisible() )
					(*plit)->hide();
			}

			//Bool Value: True = Upwards, False = downwards
			repaint(SelectNextEntry(true));
		}
		break;

		case VK_DOWN:
		{
			for(PopupListIterator plit = subpopupList.begin(); plit != subpopupList.end(); plit++)
			{
				if ( !(*plit)->bPinnedPopupMode && (*plit)->isVisible() )
					(*plit)->hide();
			}

			//Bool Value: True = Upwards, False = downwards
			repaint(SelectNextEntry(false));
		}
		break;

		case VK_RETURN:
		{
			POINT pt;
			GetCursorPos(&pt);
			ScreenToClient(hWnd, &pt);

			ActionAtPoint(pt.x, pt.y, true);
		}
		break;

		default:
		{
			int howmuchfolders = 0;
			int howmuchfiles = 0;
			int iWhatFound = 0;
			int entrycounter = 0;
			
			for (EntryListIterator eit = entrylist.begin(); eit != entrylist.end(); ++eit)
			{
				entrycounter++;
				if ( (*eit)->keyCode == nKey && (*eit)->entryType < 0 )
					howmuchfolders++;
				else if ( (*eit)->keyCode == nKey && (*eit)->entryType >= 0 )
					howmuchfiles++;

				if ( hoverentry == (*eit) && (*eit)->keyCode == nKey )
					lastkeyentry = entrycounter;
			}

			if ( howmuchfolders == 0 && howmuchfiles == 0 )
				break;
	
			entrycounter = 0;
			for (eit = entrylist.begin(); eit != entrylist.end(); ++eit)
			{
				entrycounter++;
				if ( ((*eit)->keyCode == nKey || (inChordingTimeOut && (*eit)->keyCode1 == nKey)) && entrycounter > lastkeyentry )
				{
					if ( (*eit)->entryType < 0 )
						lastactiveentry = (*eit);
					hoverentry = (*eit);
					lastkeyentry = entrycounter;
				
					if ( (*eit)->hiddenPart != iCurrentPart )
						iCurrentPart = (*eit)->hiddenPart;

					SetCursorPos(actualx+(*eit)->x+2, actualy+(*eit)->y+2);

					iWhatFound = ((*eit)->entryType < 0 ? 2 : 1);

					break;
				}
			}
			if (lastkeyentry != 0 && iWhatFound == 0)
			{
				entrycounter = 0;
				for (EntryListIterator eit = entrylist.begin(); eit != entrylist.end(); ++eit)
				{
					entrycounter++;
					if ( (*eit)->keyCode == nKey || (inChordingTimeOut && (*eit)->keyCode1 == nKey) )
					{
						if ( (*eit)->entryType < 0 )
							lastactiveentry = (*eit);
						hoverentry = (*eit);
						lastkeyentry = entrycounter;
					
						if ( (*eit)->hiddenPart != iCurrentPart )
							iCurrentPart = (*eit)->hiddenPart;

						SetCursorPos(actualx+(*eit)->x+2, actualy+(*eit)->y+2);

						iWhatFound = ((*eit)->entryType < 0 ? 2 : 1);

						break;
					}
				}
			}

			repaint(true);

			//One Folder ->Open It
			if (iWhatFound == 2 && howmuchfolders == 1 && howmuchfiles == 0 && lastactiveentry != NULL)
			{
				bool alreadyCreated = false;
				for(PopupListIterator plit = subpopupList.begin(); plit != subpopupList.end(); plit++)
				{
					if (_stricmp(lastactiveentry->caption.c_str(), (*plit)->captionname.c_str()) == NULL && _stricmp(lastactiveentry->action.c_str(), (*plit)->rootScanPath.c_str()) == NULL)
					{
						(*plit)->hoverentry = (*plit)->startentry;
						(*plit)->show(true);
						alreadyCreated = true;
					}
					else
					{
						if ( !(*plit)->bPinnedPopupMode && (*plit)->isVisible() )
							(*plit)->hide();
					}
				}
				if (!alreadyCreated && lastactiveentry->clsid == 0 && lastactiveentry->hWnd == NULL)
					CreateSubMenu( ("sub_"+lastactiveentry->caption), lastactiveentry->action, lastactiveentry, true);
				else if (!alreadyCreated && lastactiveentry->clsid != 0)
					CreateShellSubMenu( ("sub_"+lastactiveentry->caption), lastactiveentry->clsid, lastactiveentry,  true);
				else if (!alreadyCreated && lastactiveentry->hWnd == hWnd)
					CreateTaskSubMenu( ("sub_"+lastactiveentry->caption), lastactiveentry, true);
			}
			//One File -> Open It
			else if (iWhatFound == 1 && howmuchfolders == 0 && howmuchfiles == 1 && hoverentry != NULL)
			{
				//Normal Action
				if ( hoverentry->entryType >= 0 && (!hoverentry->action.empty() || !hoverentry->universalaction.empty()) && hoverentry->m_pidl == NULL )
				{
					ENTRY *tmpEntry = hoverentry;

					Popup *tmpPointer = pParentPopup;
					if (tmpPointer != NULL)
					{
						while (tmpPointer->pParentPopup != NULL)
							tmpPointer = tmpPointer->pParentPopup;

						if ( !execSound.empty() )
						{
							tmpPointer->execSoundPlayed = true;
							ParseBangCommand(hWnd, "!playsound", execSound.c_str());
						}
						tmpPointer->hide(true);
					}
					else
					{
						if ( !execSound.empty() )
						{
							execSoundPlayed = true;
							ParseBangCommand(hWnd, "!playsound", execSound.c_str());
						}
						hide(true);
					}

					if ( !tmpEntry->universalaction.empty() )
						LSExecute(NULL, tmpEntry->universalaction.c_str(), NULL);
					else
						LSExecute(NULL, tmpEntry->action.c_str(), NULL);
					
					break;
				}
				//System Action
				else if ( hoverentry->entryType >= 0 && hoverentry->m_pidl != NULL )
				{
					ENTRY *tmpEntry = hoverentry;

					Popup *tmpPointer = pParentPopup;
					if (tmpPointer != NULL)
					{
						while (tmpPointer->pParentPopup != NULL)
							tmpPointer = tmpPointer->pParentPopup;

						if ( !execSound.empty() )
						{
							tmpPointer->execSoundPlayed = true;
							ParseBangCommand(hWnd, "!playsound", execSound.c_str());
						}
						tmpPointer->hide(true);
					}
					else
					{
						if ( !execSound.empty() )
						{
							execSoundPlayed = true;
							ParseBangCommand(hWnd, "!playsound", execSound.c_str());
						}
						hide(true);
					}

					Invoke(tmpEntry->m_pidl);
					
					break;
				}
				//Tasks
				else if ( hoverentry->hWnd != NULL )
				{
					ENTRY *tmpEntry = hoverentry;

					Popup *tmpPointer = pParentPopup;
					if (tmpPointer != NULL)
					{
						while (tmpPointer->pParentPopup != NULL)
							tmpPointer = tmpPointer->pParentPopup;

						if ( !execSound.empty() )
						{
							tmpPointer->execSoundPlayed = true;
							ParseBangCommand(hWnd, "!playsound", execSound.c_str());
						}
						tmpPointer->hide(true);
					}
					else
					{
						if ( !execSound.empty() )
						{
							execSoundPlayed = true;
							ParseBangCommand(hWnd, "!playsound", execSound.c_str());
						}
						hide(true);
					}

					SetForegroundWindow( tmpEntry->hWnd );

					if ( IsIconic(tmpEntry->hWnd) )
					{
						//if ( pSwitchToThisWindow )
						//	pSwitchToThisWindow( tmpEntry->hWnd, true );
						//else
							ShowWindowAsync( tmpEntry->hWnd, SW_RESTORE );
							BringWindowToTop( tmpEntry->hWnd );
					}
					else
					{
						//if ( pSwitchToThisWindow )
						//	pSwitchToThisWindow( tmpEntry->hWnd, false );
						//else
							ShowWindowAsync( tmpEntry->hWnd, SW_SHOW );
							BringWindowToTop( tmpEntry->hWnd );
					}

					SendMessage(GetLitestepWnd(), LM_BRINGTOFRONT, 0, (LPARAM)tmpEntry->hWnd);
					break;
				}
			}
			else
			{
				inChordingTimeOut = true;
				SetTimer(hWnd, TIMER_CHORDING, 750, 0);
			}
		}
		break;
	}
}

void Popup::onMouseMove(int x, int y)
{
	if ( mousePressed )
	{
		KillTimer(hWnd, TIMER_MOUSETRACK);

		isMultiScrollTimerRunning = 0;
		KillTimer(hWnd, TIMER_MULTISCROLLUP);
		KillTimer(hWnd, TIMER_MULTISCROLLDOWN);

		POINT newpt, pt;
		GetCursorPos(&newpt);
		pt = newpt;
		
		newpt.x = newpt.x-savedpt.x;
		newpt.y = newpt.y-savedpt.y;

		if ( !mouseMoved && newpt.x != actualx && newpt.y != actualy )
			mouseMoved = true;

		manualMove = false;

		if ( mouseMoved )
			reposition(1, newpt.x, newpt.y, 0, 0);
	}
	else if ( mouseInside )
	{
		SetTimer(hWnd, TIMER_MOUSETRACK, trackDelay, NULL);

		manualMove = true;

		HoverActions(x, y);
	}

	if( !mouseInside )
	{
		mouseInside = true;
		SetTimer(hWnd, TIMER_MOUSETRACK, trackDelay, NULL);

		if ( listVisiblePopups.size() == 1 )
		{
			SetForegroundWindow(hWnd);
			SetActiveWindow(hWnd);
			SetFocus(hWnd);
			BringWindowToTop(hWnd);
		}

		if ( startentry != NULL )
			lastactiveentry = startentry;

		HoverActions(x, y);
	}
}

void Popup::repaint(bool invalidateCache, bool regionUpdate)
{
	if (invalidateCache)
		bResetBackground = true;
	
	if ( !alphaMap && !fakealphaMap )
	{
		RedrawWindow(hWnd, NULL, NULL, RDW_INVALIDATE | RDW_UPDATENOW );	
			
		if ( invalidateCache || regionUpdate )
			updateRegion();
	}
	else
	{
		UpdateWindowTransparent(hWnd, alphaTransparency);
	}
}

void Popup::ResetBackground()
{
	if ( defaultBitmap )
	{
		defaultBitmap = (HBITMAP) SelectObject(defaultDC, olddefaultBMP);
		DeleteObject(defaultBitmap);
	}
	if ( defaultDC )
		DeleteDC(defaultDC);
	if ( olddefaultBMP )
		DeleteObject(olddefaultBMP);

	defaultDC = NULL;
	defaultBitmap = NULL;
	olddefaultBMP = NULL;

	if(fadeBackupBitmap)
	{
		fadeBackupBitmap = (HBITMAP) SelectObject(fadeBackupDC, oldfadeBackupBMP);
		DeleteObject(fadeBackupBitmap);
	}
	if (fadeBackupDC)
		DeleteDC(fadeBackupDC);
	if ( oldfadeBackupBMP )
		DeleteObject(oldfadeBackupBMP);

	fadeBackupDC = NULL;
	fadeBackupBitmap = NULL;
	oldfadeBackupBMP = NULL;

	RedrawWindow( hWnd, NULL, NULL, RDW_INVALIDATE | RDW_UPDATENOW );
}

void Popup::updateRegion()
{
	HDC tmpDC = CreateCompatibleDC(NULL);
	HBITMAP tmpBmp = CreateBitmap(actualwidth, actualheight, 1, iBitsPerPixel, NULL);
	HBITMAP oldBMP = (HBITMAP) SelectObject(tmpDC, tmpBmp);

	onPaint(tmpDC);
						
	tmpBmp = (HBITMAP) SelectObject(tmpDC, oldBMP);

	HRGN region = BitmapToRegion(tmpBmp, RGB(255, 0, 255), 0, 0, 0);
	HRGN windowRgn = CreateRectRgn(0, 0, 0, 0);
	CombineRgn(windowRgn, region, NULL, RGN_COPY);

	SetWindowRgn(hWnd, windowRgn, true);	

	DeleteObject(windowRgn);
	DeleteObject(region);

	DeleteDC(tmpDC);
	DeleteObject(tmpBmp);
	DeleteObject(oldBMP);
}

void Popup::onPaint(HDC hDC)
{
	if ( bResetBackground )
	{
		bResetBackground = false;
		ResetBackground();
	}

	RECT rectPosition;
	GetWindowRect(hWnd, &rectPosition);

	int width = (rectPosition.right-rectPosition.left);
	int height = (rectPosition.bottom-rectPosition.top);
	
	if ( width <= 0 || height <= 0 || !pPaintClasses )
		return;

	// make a rendition of the background
	HDC backgroundDC = CreateCompatibleDC(hDC);
	HBITMAP backgroundBitmap = CreateCompatibleBitmap(hDC, width, height);
	HBITMAP oldbackgroundBMP = (HBITMAP) SelectObject(backgroundDC, backgroundBitmap);
		
	RECT rect = {0,0,width,height};
	::SetBkColor(backgroundDC, RGB(255, 0, 255));
	::ExtTextOut(backgroundDC, 0, 0, ETO_OPAQUE, &rect, NULL, 0, NULL);

	//Normal Background Painting
	pPaintClasses->pPopupTexture->apply(backgroundDC, 0, 0, width, height);

	OverlayPopupPainter(backgroundDC);

	// create & keep a cached rendition of the default popupmenu
	if (defaultBitmap == NULL && defaultDC == NULL)
	{
		defaultDC = CreateCompatibleDC(hDC);
		defaultBitmap = CreateCompatibleBitmap(hDC, width, height);
		olddefaultBMP = (HBITMAP) SelectObject(defaultDC, defaultBitmap);
		
		// blt background into double buffer
		BitBlt(defaultDC, 0, 0, width, height, backgroundDC, 0, 0, SRCCOPY);

		//Only Painted if NOT Shaded
		if ( shaded == -1 )
		{
			int state = 0;
			int arrowstate = 0;
			for (EntryListIterator eit = entrylist.begin(); eit != entrylist.end(); ++eit)
			{
				if ( iCurrentPart == (*eit)->hiddenPart )
				{
					//initialize
					state = 0;
					arrowstate = 0;

					if ( (*eit)->entryType == 0 )
					{
						state = 1;
						pPaintClasses->pEntryTexture->apply( defaultDC, (*eit)->x, (*eit)->y, (*eit)->width, (*eit)->height);
						OverlayButtonPainter( defaultDC, (*eit)->x, (*eit)->y, (*eit)->width, (*eit)->height, state );
			
						if (bShowIcon)
						{
							HICON hIcon = RetrieveIcon((*eit)->iconPath, (*eit)->hWnd, (*eit)->m_pidl, DefaultIcon, bBigEntryIcon);
							pPaintClasses->pEntryIcon->apply( defaultDC, (*eit)->x, (*eit)->y, (*eit)->width, (*eit)->height, hIcon);
							DestroyIcon(hIcon);
						}
						if (iQuicklaunchMode <= 0 && bShowText)
							pPaintClasses->pEntryText->apply( defaultDC, (*eit)->x, (*eit)->y, (*eit)->width, (*eit)->height, (*eit)->caption.c_str());
					}
					else if ( (*eit)->entryType < 0 )
					{
						state = 3;
						arrowstate = 31;
						pPaintClasses->pFolderTexture->apply( defaultDC, (*eit)->x, (*eit)->y, (*eit)->width, (*eit)->height);
						OverlayButtonPainter( defaultDC, (*eit)->x, (*eit)->y, (*eit)->width, (*eit)->height, state );
			
						if (bShowIcon)
						{
							HICON hIcon = RetrieveIcon((*eit)->iconPath, (*eit)->hWnd, (*eit)->m_pidl, DefaultFolderIcon, bBigFolderIcon);
							pPaintClasses->pFolderIcon->apply( defaultDC, (*eit)->x, (*eit)->y, (*eit)->width, (*eit)->height, hIcon);
							DestroyIcon(hIcon);
						}
						if (iQuicklaunchMode <= 0 && bShowText)
							pPaintClasses->pFolderText->apply( defaultDC, (*eit)->x, (*eit)->y, (*eit)->width, (*eit)->height, (*eit)->caption.c_str());
					}
					else if ( (*eit)->entryType == 1 )
					{
						state = 10;
						pPaintClasses->pSeparatorTexture->apply( defaultDC, (*eit)->x, (*eit)->y, (*eit)->width, (*eit)->height);
						OverlayButtonPainter( defaultDC, (*eit)->x, (*eit)->y, (*eit)->width, (*eit)->height, state );
					}
					else
					{
						continue;
					}
				
					//Paint Arrow
					if (arrowstate != 0)
					{
						POINT arrowSize = pPaintClasses->pArrowTexture->getImageSize();
						pPaintClasses->pArrowTexture->apply( defaultDC, (*eit)->x+(bArrowOnLeft ? iArrowSpacing : (*eit)->width-arrowSize.x-iArrowSpacing), (*eit)->y+((*eit)->height-arrowSize.y)/2, arrowSize.x, arrowSize.y);
								
						if ( pPaintClasses->pArrowText )
							pPaintClasses->pArrowText->apply( defaultDC, (*eit)->x, (*eit)->y, (*eit)->width, (*eit)->height, strArrowText.c_str());
					}
				}
			}
		}
	}

	// double buffer for flicker-free paint
	HDC bufferDC = CreateCompatibleDC(hDC);
	HBITMAP bufferBitmap = CreateCompatibleBitmap(hDC, width, height);
	HBITMAP oldbufferBMP = (HBITMAP) SelectObject(bufferDC, bufferBitmap);

	// blt default background into double buffer
	BitBlt(bufferDC, 0, 0, width, height, defaultDC, 0, 0, SRCCOPY);
		
	CString strHelpCaption = captionname.c_str();
	strHelpCaption.Replace("&&", "..--..");
	strHelpCaption.Replace("&", "");
	strHelpCaption.Replace("..--..", "&&");
	
	if ( iQuicklaunchMode == 0 )
	{
		if ( topBorder > 0 )
		{
			pPaintClasses->pTitleTexture->apply( bufferDC, 0, 0, actualwidth, topBorder );

			OverlayButtonPainter( bufferDC, 0, 0, actualwidth, topBorder, -1 );
		}
		if ( bottomBorder > 0 )
		{
			pPaintClasses->pBottomTexture->apply( bufferDC, 0, actualheight-bottomBorder, actualwidth, bottomBorder );

			OverlayButtonPainter( bufferDC, 0, actualheight-bottomBorder, actualwidth, bottomBorder, -2 );
		}
		if (showCaption == 1)
		{
			if ( pParentPopup != NULL && pParentEntry != NULL && pPaintClasses->pTitleIcon )
			{
				HICON hIcon = RetrieveIcon(pParentEntry->iconPath, pParentEntry->hWnd, pParentEntry->m_pidl, "", bBigTitleIcon);
				pPaintClasses->pTitleIcon->apply( bufferDC, 0, 0, actualwidth, topBorder, hIcon );
				DestroyIcon(hIcon);
			}
			if ( bPinnedPopupMode && pPaintClasses->pPinnedIcon )
			{		
				HICON hIcon = RetrieveIcon("", NULL, NULL, strPinnedIconPath, bBigPinnedIcon);
				pPaintClasses->pPinnedIcon->apply( bufferDC, 0, 0, actualwidth, topBorder, hIcon );
				DestroyIcon(hIcon);
			}

			pPaintClasses->pTitleText->apply( bufferDC, 0, 0, actualwidth, topBorder, strHelpCaption);
		}
		else if (showCaption == 2)
		{
			if ( pParentPopup != NULL && pParentEntry != NULL && pPaintClasses->pTitleIcon )
			{
				HICON hIcon = RetrieveIcon(pParentEntry->iconPath, pParentEntry->hWnd, pParentEntry->m_pidl, "", bBigTitleIcon);
				pPaintClasses->pTitleIcon->apply( bufferDC, 0, actualheight-bottomBorder, actualwidth, bottomBorder, hIcon );
				DestroyIcon(hIcon);
			}
			if ( bPinnedPopupMode && pPaintClasses->pPinnedIcon )
			{
				HICON hIcon = RetrieveIcon("", NULL, NULL, strPinnedIconPath, bBigPinnedIcon);
				pPaintClasses->pPinnedIcon->apply( bufferDC, 0, actualheight-bottomBorder, actualwidth, bottomBorder, hIcon );
				DestroyIcon(hIcon);
			}
			pPaintClasses->pTitleText->apply( bufferDC, 0, actualheight-bottomBorder, actualwidth, bottomBorder, strHelpCaption);
		}
		else if (showCaption == 3)
		{
			if ( pParentPopup != NULL && pParentEntry != NULL && pPaintClasses->pTitleIcon )
			{
				HICON hIcon = RetrieveIcon(pParentEntry->iconPath, pParentEntry->hWnd, pParentEntry->m_pidl, "", bBigTitleIcon);
				pPaintClasses->pTitleIcon->apply( bufferDC, 0, 0, actualwidth, actualheight, hIcon );
				DestroyIcon(hIcon);
			}
			if ( bPinnedPopupMode && pPaintClasses->pPinnedIcon )
			{
				HICON hIcon = RetrieveIcon("", NULL, NULL, strPinnedIconPath, bBigPinnedIcon);
				pPaintClasses->pPinnedIcon->apply( bufferDC, 0, 0, actualwidth, actualheight, hIcon );
				DestroyIcon(hIcon);
			}
			pPaintClasses->pTitleText->apply( bufferDC, 0, 0, actualwidth, actualheight, strHelpCaption);
		}
	}
	//Only Painted if NOT Shaded
	if ( shaded == -1 )
	{
		int state = 0;
		int arrowstate = 0;
		if ( iQuicklaunchMode == 0 && hasInfoEntry )
		{
			for (EntryListIterator eit = entrylist.begin(); eit != entrylist.end(); ++eit)
			{
				if ( iCurrentPart == (*eit)->hiddenPart && (*eit) != hoverentry && (*eit)->entryType == 2 )
				{
					state = 5;

					//Normal Background Painting behind ActiveEntry
					BitBlt(bufferDC, (*eit)->x, (*eit)->y, (*eit)->width, (*eit)->height, backgroundDC, (*eit)->x, (*eit)->y, SRCCOPY);

					//Paint Entry Texture
					pPaintClasses->pInfoTexture->apply( bufferDC, (*eit)->x, (*eit)->y, (*eit)->width, (*eit)->height);
					
					OverlayButtonPainter( bufferDC, (*eit)->x, (*eit)->y, (*eit)->width, (*eit)->height, state );
					
					HICON hIcon = RetrieveIcon((*eit)->iconPath, NULL, NULL, "", bBigInfoIcon);
					pPaintClasses->pInfoIcon->apply( bufferDC, (*eit)->x, (*eit)->y, (*eit)->width, (*eit)->height, hIcon);
					DestroyIcon(hIcon);
				
					pPaintClasses->pInfoText->apply( bufferDC, (*eit)->x, (*eit)->y, (*eit)->width, (*eit)->height, (*eit)->caption.c_str());
				}
				else if ( iCurrentPart == (*eit)->hiddenPart && (*eit) == hoverentry && (*eit)->entryType == 2 )
				{
					state = 6;

					//Normal Background Painting behind ActiveEntry
					BitBlt(bufferDC, (*eit)->x, (*eit)->y, (*eit)->width, (*eit)->height, backgroundDC, (*eit)->x, (*eit)->y, SRCCOPY);

					//Paint Entry Texture
					pPaintClasses->pActiveInfoTexture->apply( bufferDC, (*eit)->x, (*eit)->y, (*eit)->width, (*eit)->height);
				
					OverlayButtonPainter( bufferDC, (*eit)->x, (*eit)->y, (*eit)->width, (*eit)->height, state );
					
					HICON hIcon = RetrieveIcon((*eit)->iconPath, NULL, NULL, "", bBigActiveInfoIcon);
					pPaintClasses->pActiveInfoIcon->apply( bufferDC, (*eit)->x, (*eit)->y, (*eit)->width, (*eit)->height, hIcon);
					DestroyIcon(hIcon);
					
					pPaintClasses->pActiveInfoText->apply( bufferDC, (*eit)->x, (*eit)->y, (*eit)->width, (*eit)->height, (*eit)->caption.c_str());
				}
			}
		}
		for (EntryListIterator eit = entrylist.begin(); eit != entrylist.end(); ++eit)
		{
			//initialize
			state = 0;
			arrowstate = 0;
			
			if ( iCurrentPart == (*eit)->hiddenPart && (*eit) == hoverentry && (*eit) != NULL)
			{
				if ( (*eit)->entryType == 0 )
				{
					//Normal Background Painting behind ActiveEntry
					BitBlt(bufferDC, (*eit)->x, (*eit)->y, (*eit)->width, (*eit)->height, backgroundDC, (*eit)->x, (*eit)->y, SRCCOPY);

					state = 2;
					pPaintClasses->pActiveEntryTexture->apply( bufferDC, (*eit)->x, (*eit)->y, (*eit)->width, (*eit)->height);
					OverlayButtonPainter( bufferDC, (*eit)->x, (*eit)->y, (*eit)->width, (*eit)->height, state );
					//If No Info, Icon is valid
					if (bShowIcon)
					{
						HICON hIcon = RetrieveIcon((*eit)->iconPath, (*eit)->hWnd, (*eit)->m_pidl, DefaultIcon, bBigActiveEntryIcon);
						pPaintClasses->pActiveEntryIcon->apply( bufferDC, (*eit)->x, (*eit)->y, (*eit)->width, (*eit)->height, hIcon);
						DestroyIcon(hIcon);
					}
					if (iQuicklaunchMode <= 0 && bShowText)
						pPaintClasses->pActiveEntryText->apply( bufferDC, (*eit)->x, (*eit)->y, (*eit)->width, (*eit)->height, (*eit)->caption.c_str());
				}
				else if ( (*eit)->entryType < 0 )
				{
					//Normal Background Painting behind ActiveEntry
					BitBlt(bufferDC, (*eit)->x, (*eit)->y, (*eit)->width, (*eit)->height, backgroundDC, (*eit)->x, (*eit)->y, SRCCOPY);

					state = 4;
					arrowstate = 41;
					pPaintClasses->pActiveFolderTexture->apply( bufferDC, (*eit)->x, (*eit)->y, (*eit)->width, (*eit)->height);
					OverlayButtonPainter( bufferDC, (*eit)->x, (*eit)->y, (*eit)->width, (*eit)->height, state );
					//If No Info, Icon is valid
					if (bShowIcon)
					{
						HICON hIcon = RetrieveIcon((*eit)->iconPath, (*eit)->hWnd, (*eit)->m_pidl, DefaultFolderIcon, bBigActiveFolderIcon);
						pPaintClasses->pActiveFolderIcon->apply( bufferDC, (*eit)->x, (*eit)->y, (*eit)->width, (*eit)->height, hIcon);
						DestroyIcon(hIcon);
					}
					if (iQuicklaunchMode <= 0 && bShowText)	
						pPaintClasses->pActiveFolderText->apply( bufferDC, (*eit)->x, (*eit)->y, (*eit)->width, (*eit)->height, (*eit)->caption.c_str());
				}
				else
				{
					continue;
				}
			
				//Paint Arrow
				if (arrowstate != 0)
				{
					POINT arrowSize = pPaintClasses->pActiveArrowTexture->getImageSize();
					pPaintClasses->pActiveArrowTexture->apply( bufferDC, (*eit)->x+(bSelArrowOnLeft ? iSelArrowSpacing : (*eit)->width-arrowSize.x-iSelArrowSpacing), (*eit)->y+((*eit)->height-arrowSize.y)/2, arrowSize.x, arrowSize.y);
				
					if ( pPaintClasses->pActiveArrowText )
						pPaintClasses->pActiveArrowText->apply( bufferDC, (*eit)->x, (*eit)->y, (*eit)->width, (*eit)->height, strSelArrowText.c_str());
				}			
			}
		}
	}

	// blt the double buffer to the display
	BitBlt(hDC, 0, 0, width, height, bufferDC, 0, 0, SRCCOPY);

	backgroundBitmap = (HBITMAP) SelectObject(backgroundDC, oldbackgroundBMP);
	DeleteObject(backgroundBitmap);
	DeleteDC(backgroundDC);	
	DeleteObject(oldbackgroundBMP);

	bufferBitmap = (HBITMAP) SelectObject(bufferDC, oldbufferBMP);
	DeleteObject(bufferBitmap);
	DeleteDC(bufferDC);	
	DeleteObject(oldbufferBMP);
}

bool Popup::PopupActive()
{
	char tmpClass[MAX_LINE_LENGTH] = {0};
			
	POINT pt;
	GetCursorPos(&pt);
			
	GetClassName(WindowFromPoint(pt), tmpClass, MAX_LINE_LENGTH);

	if (_stricmp(tmpClass, "xpopup") == NULL	|| _stricmp(tmpClass, "xpopupxp") == NULL)
		return true;
	else if ( autoHide )
		return false;

	GetClassName(GetForegroundWindow(), tmpClass, MAX_LINE_LENGTH);

	if (_stricmp(tmpClass, "xpopup") == NULL	|| _stricmp(tmpClass, "xpopupxp") == NULL)
		return true;

	GetClassName(GetActiveWindow(), tmpClass, MAX_LINE_LENGTH);

	if (_stricmp(tmpClass, "xpopup") == NULL	|| _stricmp(tmpClass, "xpopupxp") == NULL)
		return true;

	GetClassName(GetFocus(), tmpClass, MAX_LINE_LENGTH);

	if (_stricmp(tmpClass, "xpopup") == NULL	|| _stricmp(tmpClass, "xpopupxp") == NULL)
		return true;

	return false;
}


void Popup::onTimer(int timerID)
{
	switch(timerID)
	{	
		case TIMER_MOUSETRACK:
		{
			POINT pt;
			GetCursorPos(&pt);

			if (oldmovetracker.x != pt.x || oldmovetracker.y != pt.y)
			{
				oldmovetracker = pt;
				RECT rc;
				GetWindowRect(hWnd, &rc);

				if(!PtInRect(&rc, pt) || !visible)
				{
					//To be on the save side
					ReleaseCapture();
					mousePressed = false;

					KillTimer(hWnd, TIMER_MOUSETRACK);
					
					isMultiScrollTimerRunning = 0;
					KillTimer(hWnd, TIMER_MULTISCROLLUP);
					KillTimer(hWnd, TIMER_MULTISCROLLDOWN);

					mouseInside = false;
					lastkeyentry = -1;

					if ( hoverentry != NULL ) 
					{
						if ( hoverentry->entryType >= 0 )
						{
							hoverentry = NULL;
							repaint();
						}
						else
						{
							for(PopupListIterator plit = subpopupList.begin(); plit != subpopupList.end(); plit++)
							{
								if (_stricmp( hoverentry->caption.c_str(), (*plit)->captionname.c_str()) == NULL)
								{
									if ( (*plit)->isVisible() )
										return;
								}
							}
							hoverentry = NULL;
							repaint();
						}
					}

					break;
				}
			}
		}
		break;

		case TIMER_ALPHAFADEIN:
		{
			if (UpdateAlpha <= alphaTransparency-alphaFadeStep)
			{
				isFading = true;

				UpdateAlpha += alphaFadeStep;
				UpdateWindowTransparent(hWnd, UpdateAlpha);
			}
			else
			{
				KillTimer(hWnd, TIMER_ALPHAFADEIN);
				if (UpdateAlpha != alphaTransparency)
					UpdateWindowTransparent(hWnd, alphaTransparency);

				isFading = false;

				if(fadeBackupBitmap)
				{
					fadeBackupBitmap = (HBITMAP) SelectObject(fadeBackupDC, oldfadeBackupBMP);
					DeleteObject(fadeBackupBitmap);
				}
				if (fadeBackupDC)
					DeleteDC(fadeBackupDC);
				if ( oldfadeBackupBMP )
					DeleteObject(oldfadeBackupBMP);

				fadeBackupBitmap = NULL;
				fadeBackupDC = NULL;
				oldfadeBackupBMP = NULL;
			}
		}
		break;

		case TIMER_ALPHAFADEOUT:
		{
			if (UpdateAlpha >= alphaFadeStep)
			{
				isFading = true;

				UpdateAlpha -= alphaFadeStep;
				UpdateWindowTransparent(hWnd, UpdateAlpha);		
			}
			else
			{
				KillTimer(hWnd, TIMER_ALPHAFADEOUT);
				UpdateAlpha = 0;
				UpdateWindowTransparent(hWnd, UpdateAlpha);
				ShowWindow(hWnd, SW_HIDE);

				lastactiveentry = NULL;
				hoverentry = NULL;
				lastkeyentry = -1;

				if ( !BangToCall.empty() && bExportHoverState )
				{
					PrefixedLSSetVariable(strName.c_str(), "ExportedHoverCaption", "");
					PrefixedLSSetVariable(strName.c_str(), "ExportedHoverAction", "");
				}
				
				visible = false;

				isFading = false;

				if(fadeBackupBitmap)
				{
					fadeBackupBitmap = (HBITMAP) SelectObject(fadeBackupDC, oldfadeBackupBMP);
					DeleteObject(fadeBackupBitmap);
				}
				if (fadeBackupDC)
					DeleteDC(fadeBackupDC);
				if ( oldfadeBackupBMP )
					DeleteObject(oldfadeBackupBMP);

				fadeBackupBitmap = NULL;
				fadeBackupDC = NULL;
				oldfadeBackupBMP = NULL;
			}
		}
		break;

		case TIMER_UPDATE:
		{
			if ( !visible )
				break;

			if (hasInfoEntry)
			{
				bool needRepaint = false;
				for (EntryListIterator eit = entrylist.begin(); eit != entrylist.end(); ++eit)
				{
					if ( !(*eit)->info.empty() )
					{
						string info = InfoUpdate( (*eit)->info );
						if (!needRepaint && _stricmp( (*eit)->caption.c_str() , info.c_str() ) != NULL )
							needRepaint = true;
						(*eit)->caption = info;
					}
				}
				if (needRepaint)
					repaint(false, true);
			}
			
			if (bPinnedPopupMode && isTaskPopup)
			{
				KillTimer(hWnd, TIMER_UPDATE);
				ScanTasks();

				//In Screen Check of Tasks Popup
				int x = actualx;
				int y = actualy;

				if ( iRealMonitors > 1 )
				{
					HMONITOR hMonitor = MonitorFromWindow(hWnd, MONITOR_DEFAULTTONEAREST);
					
					MONITORINFO mi = { sizeof(MONITORINFO) };
					GetMonitorInfo(hMonitor, &mi);

					if ( x+actualwidth > mi.rcMonitor.right || iSnappedCode == 20 || iSnappedCode == 21 || iSnappedCode == 23 )
						x = mi.rcMonitor.right - actualwidth;
					else if ( iSnappedCode == 40 || iSnappedCode == 41 || iSnappedCode == 43 )
						x = mi.rcMonitor.left;

					if ( y+actualheight > mi.rcMonitor.bottom || iSnappedCode == 3 || iSnappedCode == 23 || iSnappedCode == 43 )
						y = mi.rcMonitor.bottom - actualheight;
					else if ( iSnappedCode == 1 || iSnappedCode == 21 || iSnappedCode == 41 )
						y = mi.rcMonitor.top;

					//Minimum x, y is screenX/Y
					if ( x < mi.rcMonitor.left )
						x = mi.rcMonitor.left;
					if ( y < mi.rcMonitor.top )
						y = mi.rcMonitor.top;
				}
				else
				{
					if ( x+actualwidth > iScreenWidth || iSnappedCode == 20 || iSnappedCode == 21 || iSnappedCode == 23 )
						x = iScreenWidth - actualwidth;
					else if ( iSnappedCode == 40 || iSnappedCode == 41 || iSnappedCode == 43 )
						x = 0;

					if ( y+actualheight > iScreenHeight || iSnappedCode == 3 || iSnappedCode == 23 || iSnappedCode == 43 )
						y = iScreenHeight - actualheight;
					else if ( iSnappedCode == 1 || iSnappedCode == 21 || iSnappedCode == 41 )
						y = 0;

					//Minimum x, y is screenX/Y
					if ( x < iScreenX )
						x = iScreenX;
					if ( y < iScreenY )
						y = iScreenY;
				}

				reposition(1, x, y, 0, 0);
			}
			else if (bPinnedPopupMode && hasDynamicContent)
			{
				KillTimer(hWnd, TIMER_UPDATE);
				ParseStepRC(false);

				//In Screen Check of Tasks Popup
				int x = actualx;
				int y = actualy;

				if ( iRealMonitors > 1 )
				{
					HMONITOR hMonitor = MonitorFromWindow(hWnd, MONITOR_DEFAULTTONEAREST);
					
					MONITORINFO mi = { sizeof(MONITORINFO) };
					GetMonitorInfo(hMonitor, &mi);

					if ( x+actualwidth > mi.rcMonitor.right || iSnappedCode == 20 || iSnappedCode == 21 || iSnappedCode == 23 )
						x = mi.rcMonitor.right - actualwidth;
					else if ( iSnappedCode == 40 || iSnappedCode == 41 || iSnappedCode == 43 )
						x = mi.rcMonitor.left;

					if ( y+actualheight > mi.rcMonitor.bottom || iSnappedCode == 3 || iSnappedCode == 23 || iSnappedCode == 43 )
						y = mi.rcMonitor.bottom - actualheight;
					else if ( iSnappedCode == 1 || iSnappedCode == 21 || iSnappedCode == 41 )
						y = mi.rcMonitor.top;

					//Minimum x, y is screenX/Y
					if ( x < mi.rcMonitor.left )
						x = mi.rcMonitor.left;
					if ( y < mi.rcMonitor.top )
						y = mi.rcMonitor.top;
				}
				else
				{
					if ( x+actualwidth > iScreenWidth || iSnappedCode == 20 || iSnappedCode == 21 || iSnappedCode == 23 )
						x = iScreenWidth - actualwidth;
					else if ( iSnappedCode == 40 || iSnappedCode == 41 || iSnappedCode == 43 )
						x = 0;

					if ( y+actualheight > iScreenHeight || iSnappedCode == 3 || iSnappedCode == 23 || iSnappedCode == 43 )
						y = iScreenHeight - actualheight;
					else if ( iSnappedCode == 1 || iSnappedCode == 21 || iSnappedCode == 41 )
						y = 0;

					//Minimum x, y is screenX/Y
					if ( x < iScreenX )
						x = iScreenX;
					if ( y < iScreenY )
						y = iScreenY;
				}

				reposition(1, x, y, 0, 0);
			}
		}
		break;

		case TIMER_ACTIVE:
		{	
			if ( PopupActive() )
				break;

			if ( bPinnedPopupMode )
			{
				if (hoverentry != NULL)
					hoverentry = NULL;

				repaint();
			}
			
			hide();
		}
		break;

		case TIMER_HOVER:
		{
			KillTimer(hWnd, TIMER_HOVER);

			ENTRY *entry = NULL;
	
			POINT pt;
			GetCursorPos(&pt);
			ScreenToClient(hWnd, &pt);

			if ( lastactiveentry != NULL )
			{
				entry = lastactiveentry;
				
				if ( pt.x > (entry->x + entry->width) || pt.x < entry->x 
						|| pt.y > (entry->y + entry->height) || pt.y < entry->y 
						|| entry->hiddenPart != iCurrentPart )
				{	
					break;
				}
			}
			else
			{
				for (EntryListIterator eit = entrylist.begin(); eit != entrylist.end(); ++eit)
				{
					if (iCurrentPart != (*eit)->hiddenPart)
						continue;

					if ( pt.x < ((*eit)->x + (*eit)->width) && pt.x >= (*eit)->x )
					{
						if ( pt.y < ((*eit)->y + (*eit)->height) && pt.y >= (*eit)->y )
						{
							entry = (*eit);
							break;
						}
					}
				}
			}

			if (entry != NULL)
			{
				//Check if mouse still hovers over old entry

				bool alreadyCreated = false;
				for(PopupListIterator plit = subpopupList.begin(); plit != subpopupList.end(); plit++)
				{
					if ( !(*plit)->bPinnedPopupMode && (*plit)->isVisible() )
					{
						(*plit)->hide();
						if ( (*plit)->pParentPopup != NULL)
						{
							SetForegroundWindow((*plit)->pParentPopup->gethWnd());
							SetActiveWindow((*plit)->pParentPopup->gethWnd());
							SetFocus((*plit)->pParentPopup->gethWnd());
						}
					}
				}
				for(plit = subpopupList.begin(); plit != subpopupList.end(); plit++)
				{
					if (_stricmp(entry->caption.c_str(), (*plit)->captionname.c_str()) == NULL && _stricmp(entry->action.c_str(), (*plit)->rootScanPath.c_str()) == NULL)
					{
						(*plit)->show();

						if ( pTooltip && (iQuicklaunchMode != 0 || iTooltipMode != 0) )
						{	
							if ( entry->tooltip.empty() )
							{
								if ( iQuicklaunchMode != 0 || iTooltipMode == 1 )
								{
									CString worker = "";
									worker = TrimAndStripQuote(entry->action).c_str();
									worker = worker.Right(worker.GetLength()-worker.ReverseFind('\\')-1);
									if (_stricmp(worker.Right(4), ".lnk") == NULL)
										worker = worker.Left(worker.GetLength()-4);

									entry->tooltip = worker;
								}
							}

							pTooltip->apply(entry->tooltip.c_str());
						}

						alreadyCreated = true;
						break;
					}
				}
				if (!alreadyCreated && entry->clsid == 0 && entry->hWnd == NULL)
					CreateSubMenu( ("sub_"+entry->caption), entry->action, entry);
				else if (!alreadyCreated && entry->clsid != 0)
					CreateShellSubMenu( ("sub_"+entry->caption), entry->clsid, entry);
				else if (!alreadyCreated && entry->hWnd == hWnd)
					CreateTaskSubMenu( ("sub_"+entry->caption), entry);
			}
		}
		break;

		case TIMER_CHORDING:
		{
			KillTimer(hWnd, TIMER_CHORDING);
			inChordingTimeOut = false;
		}
		break;

		case TIMER_AUTOUNLOCK:
		{
			KillTimer(hWnd, TIMER_AUTOUNLOCK);
			bLockKeyboard = false;
		}
		break;

		case TIMER_MULTISCROLLDOWN:
		{
			//Next
			iCurrentPart++;
			if (iCurrentPart > iMaxParts)
				iCurrentPart = 0;

			oldmovetracker.x = 0;

			repaint(true);
		}
		break;

		case TIMER_MULTISCROLLUP:
		{
			//Prev
			iCurrentPart--;
			if (iCurrentPart < 0)
				iCurrentPart = iMaxParts;
			
			oldmovetracker.x = 0;

			repaint(true);
		}
		break;
	}
}

void Popup::onWindowPosChanging(WINDOWPOS *windowPos)
{
	if ( (windowPos->flags & SWP_NOMOVE) != 0 || iSnapTo == 0 )
		return;

	int result = 0;

	int snaphoriz = 0;
	int snapvert = 0;

	if ( iSnapTo == 1 || iSnapTo == 3 )
	{
		// left edge
		if ( windowPos->x < customWorkArea.left+iSnapDistance && (windowPos->x > customWorkArea.left-iSnapDistance || restrictToWorkArea) )
		{
			windowPos->x = customWorkArea.left;
			snaphoriz = 4;
			result = 1;
		}

		// top edge
		if ( windowPos->y < customWorkArea.top+iSnapDistance && (windowPos->y > customWorkArea.top-iSnapDistance || restrictToWorkArea) )
		{
			windowPos->y = customWorkArea.top;
			snapvert = 1;
			result = 1;
		}

		// bottom edge
		if ( (windowPos->y+actualheight < customWorkArea.bottom+iSnapDistance || restrictToWorkArea) && windowPos->y+actualheight > customWorkArea.bottom-iSnapDistance)
		{
			windowPos->y = customWorkArea.bottom - actualheight;
			snapvert = 3;
			result = 1;
		}

		// right edge
		if ( (windowPos->x+actualwidth < customWorkArea.right+iSnapDistance || restrictToWorkArea) && windowPos->x+actualwidth > customWorkArea.right-iSnapDistance)
		{
			windowPos->x = customWorkArea.right - actualwidth;
			snaphoriz = 2;
			result = 1;
		}
	}

	iSnappedCode = 10*snaphoriz + snapvert;

	if ( result != 0 || iSnapTo == 3 )
		return;
			
	POINT pos = {windowPos->x, windowPos->y};
	POINT size = {actualwidth, actualheight};

	for( PopupListIterator plit = listVisiblePopups.begin(); plit != listVisiblePopups.end(); plit++)
	{
		if ( (*plit) == this )
			continue;

		result = isCloseTo( pos, size, (*plit)->actualx, (*plit)->actualy, (*plit)->actualwidth, (*plit)->actualheight, iSnapDistance);
		
		if (result != 0)
			break;
	}

	if (result == 1)
	{
		windowPos->x = pos.x;
	}
	else if (result == 2)
	{
		windowPos->y = pos.y;
	}
	else if (result == 3)
	{
		windowPos->x = pos.x;
		windowPos->y = pos.y;
	}
}

void Popup::onMove(int x, int y)
{
	actualx = x;
	actualy = y;

	if ( !BangToCall.empty() )
	{
		CString help = BangToCall.c_str();
		help.Replace("!", "");

		PrefixedLSSetVariable((string)help, "CurrentX", IntToString( x ));
		PrefixedLSSetVariable((string)help, "CurrentY", IntToString( y ));
	}
}

void Popup::onSize(int width, int height)
{
	actualwidth = width;
	actualheight = height;

	if (alphaMap || fakealphaMap)
		SetWindowRgn(hWnd, NULL, true);
	
	if ( !BangToCall.empty() )
	{
		CString help = BangToCall.c_str();
		help.Replace("!", "");

		PrefixedLSSetVariable((string)help, "CurrentWidth", IntToString( width ));
		PrefixedLSSetVariable((string)help, "CurrentHeight", IntToString( height ));
	}

	repaint(true, true);

	if ( pTooltip && (iQuicklaunchMode != 0 || iTooltipMode != 0) )
		pTooltip->apply("", true);
}

BOOL Popup::onWindowMessage( UINT nMessage, WPARAM wParam, LPARAM lParam, LRESULT &lResult )
{
	lResult = 0;

	if ((0x8000 & GetKeyState(VK_CONTROL)) && (0x8000 & GetKeyState(VK_SHIFT)) )
	{
		if (HideAllowed && bPinnedPopupMode && (0x8000 & GetKeyState(VK_UP)))
		{
			bPinnedPopupMode = false;
			setAlwaysOnTop(true);

			return TRUE;
		}
		else if (!bPinnedPopupMode && (0x8000 & GetKeyState(VK_DOWN)))
		{
			bPinnedPopupMode = true;
			if (bPinnedPopupNotOnTop)
				setAlwaysOnTop(false);
			else 
				setAlwaysOnTop(true);
			SetTimer(hWnd, TIMER_ACTIVE, hideDelay, NULL);

			return TRUE;
		}
	}

	switch ( nMessage )
	{
		case LM_HIDEPOPUP:
		{
			hide();
			return false;
		}

		case WM_CLOSE:
		{
			lResult = 0;
			return true;
		}

		case WM_LBUTTONDBLCLK:
		{
			onLButtonDblClk(GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam));
			return true;
		}

		case WM_LBUTTONDOWN:
		{
			ontemplateButtonDown(GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam), titleLClickAction, bottomLClickAction);
			return true;
		}

		case WM_LBUTTONUP:
		{
			ontemplateButtonUp(GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam), titleLClickAction, bottomLClickAction, 1);
			return true;
		}

		case WM_MBUTTONDOWN:
		{
			ontemplateButtonDown(GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam), titleMClickAction, bottomMClickAction);
			return true;
		}

		case WM_MBUTTONUP:
		{
			ontemplateButtonUp(GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam), titleMClickAction, bottomMClickAction, 3);
			return true;
		}

		case WM_RBUTTONDOWN:
		{
			ontemplateButtonDown(GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam), titleRClickAction, bottomRClickAction);
			return true;
		}

		case WM_RBUTTONUP:
		{
			ontemplateButtonUp(GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam), titleRClickAction, bottomRClickAction, 2);
			return true;
		}

		case WM_MOUSEWHEEL:
		{
			POINT pt;
			RECT childPosition;

			GetCursorPos(&pt);

			char tmpClass[MAX_LINE_LENGTH] = {0};				
			
			for(HWND hWndChild = ::GetTopWindow(hWnd); hWndChild != NULL; hWndChild = ::GetNextWindow(hWndChild, GW_HWNDNEXT))
			{
				GetWindowRect(hWndChild, &childPosition);
				if ( PtInRect(&childPosition, pt) )
				{
					GetClassName(hWndChild, tmpClass, MAX_LINE_LENGTH);
					if (_stricmp(tmpClass, "xlabel") == NULL	|| _stricmp(tmpClass, "xlabelxp") == NULL || _stricmp(tmpClass, "xtray") == NULL || _stricmp(tmpClass, "xtaskbar") == NULL)
					{
						ScreenToClient(hWndChild, &pt);
						::PostMessage(hWndChild, WM_MOUSEWHEEL, wParam, MAKELPARAM((short) pt.x, (short) pt.y));
								
						return true;
					}

					return true;
				}
			}

			ScreenToClient(hWnd, &pt);
			if ((short)(HIWORD(wParam)) < 0)
			{
				onWheelDown(pt.x, pt.y);

				return true;
			}
			else
			{
				onWheelUp(pt.x, pt.y);

				return true;
			}
		}

		case WM_XBUTTONDOWN:
		{
			return true;
		}

		case WM_XBUTTONUP:
		{
			char tmpClass[MAX_LINE_LENGTH] = {0};
			
			POINT pt;
			GetCursorPos(&pt);
			HWND window = WindowFromPoint(pt);

			if ( window != hWnd )
			{
				SendMessage(window, nMessage, wParam, lParam);
				return true;
			}
			else
			{
				if (HIWORD(wParam) == XBUTTON1)
				{
					if (pParentPopup != NULL)
					{
						if (pParentPopup->lastactiveentry != NULL)
						{
							pParentPopup->hoverentry = pParentPopup->lastactiveentry;
							SetCursorPos(pParentPopup->actualx+pParentPopup->lastactiveentry->x+2, pParentPopup->actualy+pParentPopup->lastactiveentry->y+2);	
						}
						else if ( pParentPopup->startentry != NULL )
						{
							pParentPopup->hoverentry = pParentPopup->startentry;
							SetCursorPos(pParentPopup->actualx+pParentPopup->startentry->x+2, pParentPopup->actualy+pParentPopup->startentry->y+2);	
						}

						if ( !bPinnedPopupMode && visible )
							hide();

						pParentPopup->repaint();

						SetForegroundWindow(pParentPopup->gethWnd());
						SetActiveWindow(pParentPopup->gethWnd());
						SetFocus(pParentPopup->gethWnd());
					}
				}
				else if (HIWORD(wParam) == XBUTTON2)
				{
					for (EntryListIterator eit = entrylist.begin(); eit != entrylist.end(); ++eit)
					{
						if ( (*eit) == hoverentry && (*eit)->entryType < 0 )
						{
							lastactiveentry = (*eit);
							bool alreadyCreated = false;
							for(PopupListIterator plit = subpopupList.begin(); plit != subpopupList.end(); plit++)
							{
								if (_stricmp((*eit)->caption.c_str(), (*plit)->captionname.c_str()) == NULL && _stricmp((*eit)->action.c_str(), (*plit)->rootScanPath.c_str()) == NULL)
								{
									(*plit)->hoverentry = (*plit)->startentry;
									(*plit)->show(true);
									alreadyCreated = true;
								}
								else
								{
									if ( !(*plit)->bPinnedPopupMode && (*plit)->isVisible() )
										(*plit)->hide();
								}
							}
							if (!alreadyCreated && (*eit)->clsid == 0 && (*eit)->hWnd == NULL)
								CreateSubMenu( ("sub_"+(*eit)->caption), (*eit)->action, (*eit), true);
							else if (!alreadyCreated && (*eit)->clsid != 0)
								CreateShellSubMenu( ("sub_"+(*eit)->caption), (*eit)->clsid, (*eit), true);
							else if (!alreadyCreated && (*eit)->hWnd == hWnd)
								CreateTaskSubMenu( ("sub_"+(*eit)->caption), (*eit), true);
						
							break;
						}
					}
				}
			}
			
			return true;
		}
	
		case WM_MOUSEMOVE:
		{
			if ( hWndToolTip )
			{
				MSG hintMessage = {hWnd, nMessage, wParam, lParam};
				SendMessage( hWndToolTip, (UINT) TTM_RELAYEVENT, 0, (LPARAM) (LPMSG) &hintMessage ); 
			}

			onMouseMove( GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam) );
			return true;
		}
	
		case WM_PAINT:
		{
			if ( alphaMap || fakealphaMap )
				break;

			PAINTSTRUCT ps;

			HDC hDC;

			hDC = BeginPaint( hWnd, &ps );

			onPaint( hDC );

			EndPaint( hWnd, &ps );
			
			return true;
		}

		case WM_WINDOWPOSCHANGING:
		{
			if ( !manualMove || restrictToWorkArea )
				onWindowPosChanging((WINDOWPOS *) lParam);
			
			return true;
		}

		case WM_MOVE:
		{
			onMove( GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam) );
			return true;
		}

		case WM_SIZE:
		{
			onSize( GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam) );
			return true;
		}

		case WM_TIMER:
		{
			onTimer((int) wParam);
			return true;
		}

		case WM_KEYDOWN:	// catches small letters and swedish keys..
		{
			if ( bLockKeyboard )
				return true;

			char tmpClass[MAX_LINE_LENGTH] = {0};
			
			POINT pt;
			GetCursorPos(&pt);
			HWND window = WindowFromPoint(pt);

			GetClassName(window, tmpClass, MAX_LINE_LENGTH);

			if ( window != hWnd && (_stricmp(tmpClass, "xpopup") == NULL || _stricmp(tmpClass, "xpopupxp") == NULL) )
			{
				SendMessage(window, nMessage, wParam, lParam);
				return true;
			}
			else
			{
				bLockKeyboard = true;
				SetTimer(hWnd, TIMER_AUTOUNLOCK, 100, NULL);
				onKey(wParam);
				return true;
			}
		}

		case WM_KEYUP:
		{
			KillTimer(hWnd, TIMER_AUTOUNLOCK);
			bLockKeyboard = false;
			return true;
		}
		
		case WM_DROPFILES:
		{
			if(dropCommand.empty())	// we shouldn't get here... but somehow we did :p
				break;

			POINT pt;
			GetCursorPos(&pt);
			ScreenToClient(hWnd, &pt);
			for (EntryListIterator eit = entrylist.begin(); eit != entrylist.end(); ++eit)
			{
				if (iCurrentPart != (*eit)->hiddenPart)
					continue;

				if ( pt.x < ((*eit)->x + (*eit)->width) && pt.x >= (*eit)->x )
				{
					if ( pt.y < ((*eit)->y + (*eit)->height) && pt.y >= (*eit)->y )
					{
						hoverentry = *eit;
						break;
					}
				}
			}

			if ( hoverentry == NULL )
			{
				break;
			}
			else if ( hoverentry != NULL )
			{
				if ( hoverentry->entryType >= 0 || hoverentry->clsid != 0 )
					break;
			}

			int intFileAction = 0;
			char szTarget[MAX_LINE_LENGTH] = {0};

			strcpy(szTarget, TrimAndStripQuote(hoverentry->action).c_str());

			if (_stricmp(dropCommand.c_str(), "move") == 0)
				intFileAction = 1;
			else if (_stricmp(dropCommand.c_str(), "copy") == 0)
				intFileAction = 2;
			else if (_stricmp(dropCommand.c_str(), "shortcut") == 0)
				intFileAction = 3;

			string check = szTarget;
			if (check.substr(1, check.length()-1) == "\\")
				check.append( "*" );
			else
				check.append( "\\*" );
				
			//Test Path for Existence
			WIN32_FIND_DATA FindData;
			HANDLE hFindHandle = FindFirstFile(check.c_str(), &FindData);

			if (intFileAction > 0 && intFileAction < 3)
			{
				char *szSourceFiles = NULL;
				char buffer[MAX_LINE_LENGTH];
				int numFiles;
				UINT len,scursor = 0;

				numFiles = DragQueryFile((HDROP) wParam, 0xFFFFFFFF, NULL, NULL);

				szSourceFiles = new char[MAX_LINE_LENGTH*numFiles+1];
				scursor = 0;

				for (int i=0;i<numFiles;i++)
				{
					len = DragQueryFile((HDROP) wParam, i, buffer, MAX_LINE_LENGTH);
					
					strcpy(szSourceFiles+scursor,buffer);
					scursor += len + 1;
				}
					
				DragFinish((HDROP) wParam);

				if( hFindHandle != INVALID_HANDLE_VALUE || intFileAction == 0 )
				{
					szSourceFiles[scursor] = 0;
					strcpy(buffer, szTarget);
					buffer[strlen(buffer)+1] = 0;
					
					SHFILEOPSTRUCT shfop;
					shfop.hwnd = NULL;
					if (intFileAction == 1)
						shfop.wFunc = FO_MOVE;
					else if (intFileAction == 2)
						shfop.wFunc = FO_COPY;
					shfop.pFrom = szSourceFiles;
					shfop.pTo = buffer;
					shfop.fFlags = FOF_NOCONFIRMMKDIR | FOF_ALLOWUNDO;
					SHFileOperation(&shfop);
				}
				
				delete[] szSourceFiles;
			}
			else if (hFindHandle != INVALID_HANDLE_VALUE && intFileAction == 3)
			{
				CString path = szTarget;
				if (path.Right(1) != "\\")
					path += "\\";
				int numDropped;
				char szFile[MAX_LINE_LENGTH];
				
				numDropped = DragQueryFile((HDROP) wParam, 0xFFFFFFFF, NULL, 0);
				for (int i = 0; i < numDropped; i++)
				{
					DragQueryFile((HDROP) wParam, i, (char *)&szFile, MAX_LINE_LENGTH);

					HRESULT hres; 
					IShellLink* psl; 

					hres = CoCreateInstance(CLSID_ShellLink, NULL,
					CLSCTX_INPROC_SERVER, IID_IShellLink, (LPVOID *) &psl); 
					if (SUCCEEDED(hres))
					{
						IPersistFile* ppf; 
						psl->SetPath(szFile); 
						psl->SetDescription("");
 
						hres = psl->QueryInterface(IID_IPersistFile, (LPVOID*)&ppf); 
						if (SUCCEEDED(hres))
						{
							CString work = szFile;
							work = work.Right(work.GetLength()-work.ReverseFind('\\')-1);
							work = work.Left(work.Find("."));
							work += ".lnk";

							WCHAR wsz[MAX_LINE_LENGTH]; 
							MultiByteToWideChar(CP_ACP, 0, (path + work), -1, wsz, MAX_LINE_LENGTH);  
							hres = ppf->Save(wsz, TRUE); 
							ppf->Release(); 
						}
						psl->Release(); 
					}
				}
				DragFinish((HDROP) wParam);	
			}
			FindClose(hFindHandle);
			
			return 0;
		}
	}

	return false;
}

LRESULT WINAPI Popup::windowProcedure(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	Popup *popup = (Popup *) GetWindowLong(hWnd, 0);

	if (message == WM_NCCREATE)
	{
		LPCREATESTRUCT pcs = (LPCREATESTRUCT) lParam;
		popup = (Popup *) pcs->lpCreateParams;

		popup->hWnd = hWnd;
		SetWindowLong(hWnd, 0, (LONG) popup);
	}

	if (popup)
	{
		LRESULT lResult;

		if (popup->onWindowMessage(message, wParam, lParam, lResult))
			return lResult;
	}

	return DefWindowProc(hWnd, message, wParam, lParam);
}

ENTRY *Popup::ActionAtPoint( int xP, int yP, bool execute, bool context, int filepath)
{
	if ( filepath == 1 )
		dragndropSource = "";
	else if ( filepath == 2 )
		dragndropDestination = "";

	for (EntryListIterator eit = entrylist.begin(); eit != entrylist.end(); ++eit)
	{
		if (iCurrentPart != (*eit)->hiddenPart)
			continue;

		if ( yP < ((*eit)->y + (*eit)->height) && yP >= (*eit)->y )
		{
			if ( xP < ((*eit)->x + (*eit)->width) && xP >= (*eit)->x )
			{
				if ( (*eit)->entryType > 0 )
					return NULL;
				//-------------------------------------------------------------------
				//LEFTCLICK NORMAL EXECUTE AND FOLDEROPENONCLICK (IF ACTIVE)
				//-------------------------------------------------------------------
				if ( execute && !context )
				{
					if ( (*eit)->entryType == 0 )
					{
						if ( !(*eit)->action.empty() || !(*eit)->universalaction.empty() )
						{
							if ( (closeAfterAction && !(0x8000 & GetKeyState(VK_CONTROL))) || (!closeAfterAction && (0x8000 & GetKeyState(VK_CONTROL))))
							{
								Popup *tmpPointer = pParentPopup;
								if (tmpPointer != NULL)
								{
									while (tmpPointer->pParentPopup != NULL)
										tmpPointer = tmpPointer->pParentPopup;

									if ( !execSound.empty() )
									{
										tmpPointer->execSoundPlayed = true;
										ParseBangCommand(hWnd, "!playsound", execSound.c_str());
									}
									tmpPointer->hide(true);
								}
								else
								{	
									if ( !execSound.empty() )
									{
										execSoundPlayed = true;
										ParseBangCommand(hWnd, "!playsound", execSound.c_str());
									}
									hide(true);
								}
							}
							
							if ( !(*eit)->universalaction.empty() )
								LSExecute(NULL, (*eit)->universalaction.c_str(), NULL);
							else
								LSExecute(NULL, (*eit)->action.c_str(), NULL);
						}
						else if ( (*eit)->m_pidl != NULL )
						{
							if ( (closeAfterAction && !(0x8000 & GetKeyState(VK_CONTROL))) || (!closeAfterAction && (0x8000 & GetKeyState(VK_CONTROL))))
							{
								Popup *tmpPointer = pParentPopup;
								if (tmpPointer != NULL)
								{
									while (tmpPointer->pParentPopup != NULL)
										tmpPointer = tmpPointer->pParentPopup;

									if ( !execSound.empty() )
									{
										tmpPointer->execSoundPlayed = true;
										ParseBangCommand(hWnd, "!playsound", execSound.c_str());
									}
									tmpPointer->hide(true);
								}
								else
								{	
									if ( !execSound.empty() )
									{
										execSoundPlayed = true;
										ParseBangCommand(hWnd, "!playsound", execSound.c_str());
									}
									hide(true);
								}
							}

							Invoke((*eit)->m_pidl);
						}
						else if ( (*eit)->hWnd != NULL )
						{
							if ( (closeAfterAction && !(0x8000 & GetKeyState(VK_CONTROL))) || (!closeAfterAction && (0x8000 & GetKeyState(VK_CONTROL))))
							{
								Popup *tmpPointer = pParentPopup;
								if (tmpPointer != NULL)
								{
									while (tmpPointer->pParentPopup != NULL)
										tmpPointer = tmpPointer->pParentPopup;

									if ( !execSound.empty() )
									{
										tmpPointer->execSoundPlayed = true;
										ParseBangCommand(hWnd, "!playsound", execSound.c_str());
									}
									tmpPointer->hide(true);
								}
								else
								{
									if ( !execSound.empty() )
									{
										execSoundPlayed = true;
										ParseBangCommand(hWnd, "!playsound", execSound.c_str());
									}
									hide(true);
								}
							}

							SetForegroundWindow( (*eit)->hWnd );

							if ( IsIconic((*eit)->hWnd) )
							{
								//if ( pSwitchToThisWindow )
								//	pSwitchToThisWindow( (*eit)->hWnd, true );
								//else
									ShowWindowAsync( (*eit)->hWnd, SW_RESTORE );
									BringWindowToTop( (*eit)->hWnd );
							}
							else
							{
								//if ( pSwitchToThisWindow )
								//	pSwitchToThisWindow( (*eit)->hWnd, false );
								//else
									ShowWindowAsync( (*eit)->hWnd, SW_SHOW );
									BringWindowToTop( (*eit)->hWnd );
							}

							SendMessage(GetLitestepWnd(), LM_BRINGTOFRONT, 0, (LPARAM)(*eit)->hWnd);
						}

						hoverentry = *eit;
					}
					//Jump Cursor into Folder
					else if ( (*eit)->entryType < 0 )
					{
						bool alreadyCreated = false;
						for(PopupListIterator plit = subpopupList.begin(); plit != subpopupList.end(); plit++)
						{
							if (_stricmp((*eit)->caption.c_str(), (*plit)->captionname.c_str()) == NULL && _stricmp((*eit)->action.c_str(), (*plit)->rootScanPath.c_str()) == NULL)
							{
								if ( (0x8000 & GetKeyState(VK_CONTROL)) )
								{
									if ( !(*eit)->action.empty() )
									{
										if ((*eit)->pipedList.size() == 1)
										{
											LSExecute(NULL, (folderDoubleClickAction + " \"" + (*eit)->action + "\"").c_str(), NULL);
										}
										else
										{
											HMENU GroupMenu = CreatePopupMenu();

											int counter = 1;
											for(StringListIterator sit = (*eit)->pipedList.begin(); sit != (*eit)->pipedList.end(); ++sit)
											{
												AppendMenu( GroupMenu, MF_STRING, counter, (*sit).c_str() );
												counter++;
											}

											// display the menu
											POINT pt;
											GetCursorPos( &pt );

											int command = TrackPopupMenu( GroupMenu,
																		  TPM_RETURNCMD | TPM_RIGHTBUTTON,
																		  pt.x, pt.y, 0,
																		  this->gethWnd(),
																		  NULL );

											if ( command )
											{
												StringListIterator sit = (*eit)->pipedList.begin();
												counter = 1; 
												while (counter < command)
												{
													sit++;
													counter++;
												}

												if ( counter <= (*eit)->pipedList.size() )
												{
													LSExecute(NULL, (folderDoubleClickAction + " \"" + (*sit) + "\"").c_str(), NULL);
												}
											}
											
											if (GroupMenu)
												DestroyMenu(GroupMenu);
										}
									}
								}
								else
								{
									(*plit)->show();
								}
								alreadyCreated = true;
							}
							else
							{
								if ( !(*plit)->bPinnedPopupMode && (*plit)->isVisible() )
									(*plit)->hide();
							}
						}
						if (!alreadyCreated && (*eit)->clsid == 0 && (*eit)->hWnd == NULL)
							CreateSubMenu( ("sub_"+(*eit)->caption), (*eit)->action, (*eit));
						else if (!alreadyCreated && (*eit)->clsid != 0)
							CreateShellSubMenu( ("sub_"+(*eit)->caption), (*eit)->clsid, (*eit));
						else if (!alreadyCreated && (*eit)->hWnd == hWnd)
							CreateTaskSubMenu( ("sub_"+(*eit)->caption), (*eit));

						hoverentry = *eit;
					}
				}
				//-------------------------------------------------------------------
				//RIGHTCLICK CONTEXT EXECUTE
				//-------------------------------------------------------------------
				//Normal Context Menu
				else if ( !execute && context )
				{
					if ( !(*eit)->action.empty() )
					{
						CString easyWorker = TrimAndStripQuote((*eit)->action).c_str();

						if ( (((*eit)->pipedList.size() == 1 && (*eit)->entryType < 0) || ((*eit)->pipedList.size() <= 1 && (*eit)->entryType >= 0)) && easyWorker.Left(1) != "!" )
						{
							ContextMenu(easyWorker);
						}
						else if ( easyWorker.Left(1) != "!" )
						{
							HMENU GroupMenu = CreatePopupMenu();

							int counter = 1;
							for(StringListIterator sit = (*eit)->pipedList.begin(); sit != (*eit)->pipedList.end(); ++sit)
							{
								AppendMenu( GroupMenu, MF_STRING, counter, (*sit).c_str() );
								counter++;
							}

							// display the menu
							POINT pt;
							GetCursorPos( &pt );

							int command = TrackPopupMenu( GroupMenu,
														  TPM_RETURNCMD | TPM_RIGHTBUTTON,
														  pt.x, pt.y, 0,
														  this->gethWnd(),
														  NULL );

							if ( command )
							{
								StringListIterator sit = (*eit)->pipedList.begin();
								counter = 1; 
								while (counter < command)
								{
									sit++;
									counter++;
								}

								if ( counter <= (*eit)->pipedList.size() )
								{
									easyWorker = TrimAndStripQuote((*sit).c_str()).c_str();
									ContextMenu(easyWorker);
								}
							}
							
							if (GroupMenu)
								DestroyMenu(GroupMenu);
						}
					}
					//ShellFolder Context Menu
					else if ( (*eit)->m_pidl != NULL )
					{
						RInvoke((*eit)->m_pidl);
					}
					//Tasks ContextMenu
					else if ( (*eit)->hWnd != NULL )
					{
						HMENU systemMenu;

						systemMenu = GetSystemMenu( (*eit)->hWnd, FALSE );

						WINDOWPLACEMENT wp;
						wp.length = sizeof(WINDOWPLACEMENT);
						GetWindowPlacement( (*eit)->hWnd, &wp );

						// restore is enabled only if minimized or maximized (not normal)
						EnableMenuItem( systemMenu, SC_RESTORE, MF_BYCOMMAND | (wp.showCmd != SW_SHOWNORMAL ? MF_ENABLED : MF_GRAYED) );

						// move is enabled only if normal
						EnableMenuItem( systemMenu, SC_MOVE, MF_BYCOMMAND | (wp.showCmd == SW_SHOWNORMAL ? MF_ENABLED : MF_GRAYED) );

						// size is enabled only if normal
						EnableMenuItem( systemMenu, SC_SIZE, MF_BYCOMMAND | (wp.showCmd == SW_SHOWNORMAL ? MF_ENABLED : MF_GRAYED) );

						// minimize is enabled only if not minimized
						EnableMenuItem( systemMenu, SC_MOVE, MF_BYCOMMAND | (wp.showCmd != SW_SHOWMINIMIZED ? MF_ENABLED : MF_GRAYED) );

						// maximize is enabled only if not maximized
						EnableMenuItem( systemMenu, SC_MOVE, MF_BYCOMMAND | (wp.showCmd != SW_SHOWMAXIMIZED ? MF_ENABLED : MF_GRAYED) );

						// let application modify menu
						PostMessage( (*eit)->hWnd, WM_INITMENUPOPUP, (WPARAM) systemMenu, MAKELPARAM( 0, TRUE ) );
						PostMessage( (*eit)->hWnd, WM_INITMENU, (WPARAM) systemMenu, 0 );

						// display the menu
						POINT pt;
						GetCursorPos( &pt );

						int command = TrackPopupMenu( systemMenu,
													  TPM_RETURNCMD | TPM_RIGHTBUTTON,
													  pt.x, pt.y, 0,
													  hWnd,
													  NULL );

						if ( command )
						{
							if ( command != SC_MINIMIZE && command != SC_CLOSE )
							{
								SetForegroundWindow( (*eit)->hWnd );
								
								if ( IsIconic((*eit)->hWnd) )
								{
									//if ( pSwitchToThisWindow )
									//	pSwitchToThisWindow( (*eit)->hWnd, true );
									//else
										ShowWindowAsync( (*eit)->hWnd, SW_RESTORE );
										BringWindowToTop( (*eit)->hWnd );
								}
								else
								{
									//if ( pSwitchToThisWindow )
									//	pSwitchToThisWindow( (*eit)->hWnd, false );
									//else
										ShowWindowAsync( (*eit)->hWnd, SW_SHOW );
										BringWindowToTop( (*eit)->hWnd );
								}

								SendMessage(GetLitestepWnd(), LM_BRINGTOFRONT, 0, (LPARAM)(*eit)->hWnd);
							}

							PostMessage( (*eit)->hWnd, WM_SYSCOMMAND, (WPARAM) command, MAKELPARAM( pt.x, pt.y ) );
						}
					}
				}
				else if (!execute && !context && filepath != 0 && !(*eit)->action.empty() && (*eit)->m_pidl == NULL && (*eit)->hWnd == NULL)
				{	
					if ( filepath == 1 )
					{
						dragndropSource.append(	TrimAndStripQuote((*eit)->action, true) );
						if ( !PathFileExists( dragndropSource.c_str() ) )
							dragndropSource = "-1";
					}
					else if ( filepath == 2 )
					{
						dragndropDestination.append( TrimAndStripQuote((*eit)->action, true) );
						
						if ( !PathFileExists( dragndropDestination.c_str() ) || !PathIsDirectory( dragndropDestination.c_str() ) )
							dragndropDestination = "";
						
						if ( dragndropDestination.empty() && !rootScanPath.empty() )
						{
							dragndropDestination.append( TrimAndStripQuote(rootScanPath, true) );
							if ( !PathFileExists( dragndropDestination.c_str() ) || !PathIsDirectory( dragndropDestination.c_str() ) )
								dragndropDestination = "";
						}
					}

					return *eit;
				}
	
				break;
				//-------------------------------------------------------------------
			}
		}
	}

	return NULL;
}

void Popup::HoverActions( int xP, int yP )
{
	if (hoverentry != NULL)
	{
		if ( yP < (hoverentry->y + hoverentry->height) && yP >= hoverentry->y
				&&  xP < (hoverentry->x + hoverentry->width) && xP >= hoverentry->x 
				&& hoverentry->hiddenPart == iCurrentPart )
		{
			return;
		}
	}

	//Check TopBorder and BottomBorder form MultiPartScrolling
	if ( iMaxParts != 0 )
	{
		if ( yP < topBorder )
		{
			if ( isMultiScrollTimerRunning != 1 )
			{
				if ( isMultiScrollTimerRunning == 2 )
					KillTimer(hWnd, TIMER_MULTISCROLLDOWN);

				isMultiScrollTimerRunning = 1;
				SetTimer(hWnd, TIMER_MULTISCROLLUP, multiscrollDelay, NULL);
			}
		}
		else
		{
			KillTimer(hWnd, TIMER_MULTISCROLLUP);
			
			if ( yP >= actualheight-bottomBorder )
			{
				if ( isMultiScrollTimerRunning != 2 )
				{
					isMultiScrollTimerRunning = 2;
					SetTimer(hWnd, TIMER_MULTISCROLLDOWN, multiscrollDelay, NULL);
				}
			}
			else
			{
				isMultiScrollTimerRunning = 0;
				KillTimer(hWnd, TIMER_MULTISCROLLDOWN);
			}
		}
	}

	bool quickskip = false;
	for (EntryListIterator eit = entrylist.begin(); eit != entrylist.end(); ++eit)
	{
		if (iCurrentPart != (*eit)->hiddenPart)
			continue;

		if ( yP < ((*eit)->y + (*eit)->height) && yP >= (*eit)->y && !quickskip )
		{
			if ( xP < ((*eit)->x + (*eit)->width) && xP >= (*eit)->x )
			{
				//Separator and InfoEntry cannot be "Active" (Skip)
				if ( (*eit)->entryType == 1 /*|| (*eit)->entryType == 2*/ )
					continue;

				lastactiveentry = NULL;

				if ( pTooltip && (iQuicklaunchMode != 0 || iTooltipMode != 0) )
				{
					if ( (*eit)->tooltip.empty() )
					{
						if ( iQuicklaunchMode != 0 || iTooltipMode == 1 )
						{
							CString worker = "";
							worker = TrimAndStripQuote((*eit)->action).c_str();
							worker = worker.Right(worker.GetLength()-worker.ReverseFind('\\')-1);
							if (_stricmp(worker.Right(4), ".lnk") == NULL)
								worker = worker.Left(worker.GetLength()-4);

							(*eit)->tooltip = worker;
						}
					}

					pTooltip->apply((*eit)->tooltip.c_str());
				}
				
				if ( !subSound.empty() )
					ParseBangCommand(hWnd, "!playsound", subSound.c_str());
				
				hoverentry = (*eit);

				if ( bExportHoverState )
				{
					CString strHelper = (*eit)->caption.c_str();
					strHelper.Replace("&&", "..--..");
					strHelper.Replace("&", "");
					strHelper.Replace("..--..", "&");

					PrefixedLSSetVariable(strName.c_str(), "ExportedHoverCaption", (string)strHelper);
					
					strHelper = TrimAndStripQuote((*eit)->action).c_str();
					strHelper = ("\"" + strHelper + "\"");

					PrefixedLSSetVariable(strName.c_str(), "ExportedHoverAction", (string)strHelper);
				}

				if ( (*eit)->entryType >= 0 && !(*eit)->action.empty() )
				{
					KillTimer(hWnd, TIMER_HOVER);

					for(PopupListIterator plit = subpopupList.begin(); plit != subpopupList.end(); plit++)
					{
						if ( !(*plit)->bPinnedPopupMode && (*plit)->isVisible() )
							(*plit)->hide();
					}
				}
				else if ( (*eit)->entryType < 0 && !folderOpenOnClick )
				{
					//Hack to avoid AutoOpen on Manual Open (Keyboard, Click)
					if ( xP < (*eit)->x+1 || xP > (*eit)->x+3 || yP < (*eit)->y+1 || yP > (*eit)->y+3 )
					{
						KillTimer(hWnd, TIMER_HOVER);

						lastactiveentry = (*eit);

						SetTimer(hWnd, TIMER_HOVER, openDelay, NULL);
					}
					else
					{
						hoverentry = NULL;
					}
				}

				quickskip = true;

				break;
				//-------------------------------------------------------------------
			}
		}
	}

	if (hoverentry != NULL)
	{
		if (!quickskip && hoverentry->entryType >= 0)
			hoverentry = NULL;
	}

	if ( hoverentry == NULL && bExportHoverState )
	{
		PrefixedLSSetVariable(strName.c_str(), "ExportedHoverCaption", "");
		PrefixedLSSetVariable(strName.c_str(), "ExportedHoverAction", "");
	}
	
	repaint();
}

int Popup::isCloseTo(POINT &pos, const POINT size, const int x, const int y, const int cx, const int cy, const int span)
{
	bool internalX = false;
	bool internalY = false;

	if ((pos.x > x-span) && (pos.x < x+span))
	{
		if ( ((pos.y > y-span) && (pos.y < y+span)) 
			|| ((pos.y+size.y > y-span) && (pos.y+size.y < y+span)) 
			|| ((pos.y > y+cy-span ) && (pos.y < y+cy+span)) 
			|| ((pos.y+size.y > y+cy-span ) && (pos.y+size.y < y+cy+span)) )
		{
			pos.x = x;
			internalX = true;
		}
		else if ( (pos.y > y) && (pos.y < y+cy) 
				&& (pos.y+size.y > y) && (pos.y+size.y < y+cy) )
		{
			pos.x = x;
			internalX = true;
		}
	}
	else if ((pos.x > x+cx-span ) && (pos.x < x+cx+span))
	{
		if ( ((pos.y > y-span) && (pos.y < y+span)) 
			|| ((pos.y+size.y > y-span) && (pos.y+size.y < y+span)) 
			|| ((pos.y > y+cy-span ) && (pos.y < y+cy+span)) 
			|| ((pos.y+size.y > y+cy-span ) && (pos.y+size.y < y+cy+span)) )
		{
			pos.x = x+cx;
			internalX = true;
		}
		else if ( (pos.y > y) && (pos.y < y+cy) 
				&& (pos.y+size.y > y) && (pos.y+size.y < y+cy) )
		{
			pos.x = x+cx;
			internalX = true;
		}
	}
	else if ((pos.x+size.x > x-span ) && (pos.x+size.x < x+span))
	{
		if ( ((pos.y > y-span) && (pos.y < y+span)) 
			|| ((pos.y+size.y > y-span) && (pos.y+size.y < y+span)) 
			|| ((pos.y > y+cy-span ) && (pos.y < y+cy+span)) 
			|| ((pos.y+size.y > y+cy-span ) && (pos.y+size.y < y+cy+span)) )
		{
			pos.x = x-size.x;
			internalX = true;
		}
		else if ( (pos.y > y) && (pos.y < y+cy) 
				&& (pos.y+size.y > y) && (pos.y+size.y < y+cy) )
		{
			pos.x = x-size.x;
			internalX = true;
		}
	}
	else if ((pos.x+size.x > x+cx-span ) && (pos.x+size.x < x+cx+span))
	{
		if ( ((pos.y > y-span) && (pos.y < y+span)) 
			|| ((pos.y+size.y > y-span) && (pos.y+size.y < y+span)) 
			|| ((pos.y > y+cy-span ) && (pos.y < y+cy+span)) 
			|| ((pos.y+size.y > y+cy-span ) && (pos.y+size.y < y+cy+span)) )
		{
			pos.x = x+cx-size.x;
			internalX = true;
		}
		else if ( (pos.y > y) && (pos.y < y+cy) 
				&& (pos.y+size.y > y) && (pos.y+size.y < y+cy) )
		{
			pos.x = x+cx-size.x;
			internalX = true;
		}
	}

	if ((pos.y > y-span) && (pos.y < y+span))
	{
		if ( ((pos.x > x-span) && (pos.x < x+span)) 
			|| ((pos.x+size.x > x-span) && (pos.x+size.x < x+span)) 
			|| ((pos.x > x+cx-span ) && (pos.x < x+cx+span)) 
			|| ((pos.x+size.x > x+cx-span ) && (pos.x+size.x < x+cx+span)) )
		{
			pos.y = y;
			internalY = true;
		}
		else if ( (pos.x > x) && (pos.x < x+cx) 
				&& (pos.x+size.x > x) && (pos.x+size.x < x+cx) )
		{
			pos.y = y;
			internalY = true;
		}
	}
	else if ((pos.y > y+cy-span ) && (pos.y < y+cy+span))
	{
		if ( ((pos.x > x-span) && (pos.x < x+span)) 
			|| ((pos.x+size.x > x-span) && (pos.x+size.x < x+span)) 
			|| ((pos.x > x+cx-span ) && (pos.x < x+cx+span)) 
			|| ((pos.x+size.x > x+cx-span ) && (pos.x+size.x < x+cx+span)) )
		{
			pos.y = y+cy;
			internalY = true;
		}
		else if ( (pos.x > x) && (pos.x < x+cx) 
				&& (pos.x+size.x > x) && (pos.x+size.x < x+cx) )
		{
			pos.y = y+cy;
			internalY = true;
		}
	}
	else if ((pos.y+size.y > y-span ) && (pos.y+size.y < y+span))
	{
		if ( ((pos.x > x-span) && (pos.x < x+span)) 
			|| ((pos.x+size.x > x-span) && (pos.x+size.x < x+span)) 
			|| ((pos.x > x+cx-span ) && (pos.x < x+cx+span)) 
			|| ((pos.x+size.x > x+cx-span ) && (pos.x+size.x < x+cx+span)) )
		{
			pos.y = y-size.y;
			internalY = true;
		}
		else if ( (pos.x > x) && (pos.x < x+cx) 
				&& (pos.x+size.x > x) && (pos.x+size.x < x+cx) )
		{
			pos.y = y-size.y;
			internalY = true;
		}
	}
	else if ((pos.y+size.y > y+cy-span ) && (pos.y+size.y < y+cy+span))
	{
		if ( ((pos.x > x-span) && (pos.x < x+span)) 
			|| ((pos.x+size.x > x-span) && (pos.x+size.x < x+span)) 
			|| ((pos.x > x+cx-span ) && (pos.x < x+cx+span)) 
			|| ((pos.x+size.x > x+cx-span ) && (pos.x+size.x < x+cx+span)) )
		{
			pos.y = y+cy-size.y;
			internalY = true;
		}
		else if ( (pos.x > x) && (pos.x < x+cx) 
				&& (pos.x+size.x > x) && (pos.x+size.x < x+cx) )
		{
			pos.y = y+cy-size.y;
			internalY = true;
		}
	}

	if (!internalX && !internalY)
		return 0;
	else if (internalX && !internalY)
		return 1;
	else if (!internalX && internalY)
		return 2;
	else
		return 3;
}

void Popup::Sort(vector<string> &inputList)
{
	std::sort( inputList.begin(), inputList.end(), Compare );
}

void Popup::SortTasks(hWndTaskList &inputList)
{
	std::sort( inputList.begin(), inputList.end(), CompareTasks );
}

void Popup::SortShellFolder(vector<LPITEMIDLIST> &inputList)
{
	std::sort( inputList.begin(), inputList.end(), CompareShellFolders );
}

string Popup::InfoUpdate(string strInfoCFG)
{
	if ( m_pProcessStatsRequest )
	{
		char tmp[4096] = {0};
		m_pProcessStatsRequest( tmp, TrimAndStripQuote(strInfoCFG).c_str() );

		return (string) tmp;
	}
	else
	{
		return TrimAndStripQuote(strInfoCFG);
	}
}

bool Popup::SelectNextEntry(bool bGoUpwards, bool bNoCursorMove)
{
	if ( entrylist.size() <= 1 )
		return false;

	bool bTrueRepaint = false;
	bool bActiveOneFound = false;

	for (EntryListIterator eit = entrylist.begin(); eit != entrylist.end(); ++eit)
	{
		if ( (*eit) == hoverentry )
		{
			if (bGoUpwards)
			{
				if (eit != entrylist.begin())
				{
					eit--;
				}
				else
				{
					eit = entrylist.end();
					eit--;
				}
			}
			else
			{
				if (eit != entrylist.end()-1)
					eit++;
				else
					eit = entrylist.begin();
			}

			//Separator and InfoEntry cannot be "Active" (Skip)
			while ((*eit)->entryType == 1 /*|| (*eit)->entryType == 2*/)
			{
				if (bGoUpwards)
				{
					if (eit != entrylist.begin())
					{
						eit--;
					}
					else
					{
						eit = entrylist.end();
						eit--;
					}
				}
				else
				{
					if (eit != entrylist.end()-1)
						eit++;
					else
						eit = entrylist.begin();
				}
			}

			hoverentry = (*eit);

			if (!bNoCursorMove)
				SetCursorPos(actualx+(*eit)->x+2, actualy+(*eit)->y+2);	

			if ( iCurrentPart != (*eit)->hiddenPart)
			{
				iCurrentPart--;
				if (iCurrentPart < 0)
					iCurrentPart = iMaxParts;

				bTrueRepaint = true;
			}

			bActiveOneFound = true;

			break;
		}
	}

	if (!bActiveOneFound)
	{
		if ( bGoUpwards )
		{
			if ( startentry != NULL )
				hoverentry = startentry;

			bTrueRepaint = SelectNextEntry(true);
		}
		else
		{
			if ( startentry != NULL )
			{
				hoverentry = startentry;

				if (!bNoCursorMove)
					SetCursorPos(actualx+startentry->x+2, actualy+startentry->y+2);	

				if ( iCurrentPart != startentry->hiddenPart)
				{
					iCurrentPart++;
					if (iCurrentPart > iMaxParts)
						iCurrentPart = 0;

					bTrueRepaint = true;
				}
			}
		}
	}

	return bTrueRepaint;
}

//-----------------------------------------------------------------------
// ShellFolder Support
//-----------------------------------------------------------------------

//Normal Action
void Popup::Invoke(LPITEMIDLIST m_pidlFull)
{
	SHELLEXECUTEINFO sei;

	sei.cbSize = sizeof(SHELLEXECUTEINFO);
	sei.fMask = SEE_MASK_INVOKEIDLIST | SEE_MASK_IDLIST | SEE_MASK_FLAG_NO_UI;
	sei.hwnd = NULL;
	sei.lpVerb = NULL;
	sei.lpFile = NULL;
	sei.lpParameters = NULL;
	sei.lpDirectory = NULL;
	sei.nShow = SW_SHOW;
	sei.lpIDList = m_pidlFull;

	ShellExecuteEx(&sei);
}

//Context Menu
void Popup::RInvoke(LPITEMIDLIST m_pidlFull)
{
	LPMALLOC pMalloc;
	LPSHELLFOLDER psfFolder, psfNextFolder;
	LPITEMIDLIST pidlItem, pidlNextItem, *ppidl;
	LPCONTEXTMENU pContextMenu;
	CMINVOKECOMMANDINFO ici;
	UINT nCount, nCmd;
	HMENU hMenu;

	// Get current cursor position & window
	POINT point;
	HWND hwnd;

	GetCursorPos(&point);
	hwnd = WindowFromPoint(point);

	//
	// Get pointers to the shell's IMalloc interface and the desktop's
	// IShellFolder interface.
	//

	if (SUCCEEDED (SHGetMalloc (&pMalloc)))
	{

		if (!SUCCEEDED (SHGetDesktopFolder (&psfFolder)))
		{
			pMalloc->Release();
		}
		else
		{
			//
			// Convert the path name into a pointer to an item ID list (pidl).
			//
			if (m_pidlFull != NULL)
			{

				if (nCount = GetItemCount (m_pidlFull))
				{ // nCount must be > 0
					//
					// Initialize psfFolder with a pointer to the IShellFolder
					// interface of the folder that contains the item whose context
					// menu we're after, and initialize pidlItem with a pointer to
					// the item's item ID. If nCount > 1, this requires us to walk
					// the list of item IDs stored in m_pidlFull and bind to each
					// subfolder referenced in the list.
					//
					pidlItem = m_pidlFull;

					while (--nCount)
					{
						//
						// Create a 1-item item ID list for the next item in m_pidlFull.
						//

						pidlNextItem = DuplicateItem (pMalloc, pidlItem);

						if (pidlNextItem == NULL)
						{
							psfFolder->Release();
							pMalloc->Release();
							return ;
						}

						//
						// Bind to the folder specified in the new item ID list.
						//
						if (!SUCCEEDED (psfFolder->BindToObject (pidlNextItem, NULL, IID_IShellFolder, (void**)&psfNextFolder)))
						{
							pMalloc->Free(pidlNextItem);
							psfFolder->Release();
							pMalloc->Release();
							return ;
						}

						//
						// Release the IShellFolder pointer to the parent folder
						// and set psfFolder equal to the IShellFolder pointer for
						// the current folder.
						//
						psfFolder->Release();
						psfFolder = psfNextFolder;

						//
						// Release the storage for the 1-item item ID list we created
						// just a moment ago and initialize pidlItem so that it points
						// to the next item in m_pidlFull.
						//
						pMalloc->Free(pidlNextItem);
						pidlItem = GetNextItem (pidlItem);
					}

					//
					// Get a pointer to the item's IContextMenu interface and call
					// IContextMenu::QueryContextMenu to initialize a context menu.
					//
					ppidl = &pidlItem;
					if (SUCCEEDED (psfFolder->GetUIObjectOf (hwnd, 1, (const ITEMIDLIST**)ppidl, (const GUID)IID_IContextMenu, NULL, (void**)&pContextMenu)))
					{

						hMenu = CreatePopupMenu ();
						if (SUCCEEDED (pContextMenu->QueryContextMenu (
						                   hMenu, 0, 1, 0x7FFF, CMF_NORMAL)))
						{

							//
							// Display the context menu.
							//
							nCmd = TrackPopupMenu (hMenu, TPM_LEFTALIGN |
							                       TPM_LEFTBUTTON | TPM_RIGHTBUTTON | TPM_RETURNCMD,
							                       point.x, point.y, 0, hwnd, NULL);

							//
							// If a command was selected from the menu, execute it.
							//
							if (nCmd)
							{
								ici.cbSize = sizeof (CMINVOKECOMMANDINFO);
								ici.fMask = SEE_MASK_FLAG_NO_UI;
								ici.hwnd = hwnd;
								ici.lpVerb = MAKEINTRESOURCE (nCmd - 1);
								ici.lpParameters = NULL;
								ici.lpDirectory = NULL;
								ici.nShow = SW_SHOWNORMAL;
								ici.dwHotKey = 0;
								ici.hIcon = NULL;

								pContextMenu->InvokeCommand(&ici);


							}
							DestroyMenu (hMenu);
							pContextMenu->Release();
						}
					}
				}
			}
			//
			// Clean up and return.
			//

			psfFolder->Release();
			pMalloc->Release();
		}
	}
}

//Helpers
LPITEMIDLIST CreateIDList(int cb)
{
	IMalloc *pMalloc = NULL;
	LPITEMIDLIST pidl = NULL;

	if (cb <= 0)
		return NULL;

	if (SUCCEEDED(SHGetMalloc(&pMalloc)))
	{
		pidl = (LPITEMIDLIST) pMalloc->Alloc(cb);
		pMalloc->Release();
	}

	if (pidl)
		memset(pidl, 0, cb);

	return pidl;
}

LPITEMIDLIST DuplicateIDList(LPCITEMIDLIST pidl)
{
	LPITEMIDLIST pidlNew = NULL;;
	int cb = GetIDListSize(pidl);

	if (!pidl)
		return NULL;

	pidlNew = CreateIDList(cb);

	if (pidlNew)
		memcpy(pidlNew, pidl, cb);

	return pidlNew;
}

LPITEMIDLIST JoinIDLists(LPCITEMIDLIST pidlA, LPCITEMIDLIST pidlB)
{
	LPITEMIDLIST pidl = NULL;;
	int cbA, cbB;

	cbA = pidlA ? GetIDListSize(pidlA) - sizeof(pidlA->mkid.cb) : 0;
	cbB = GetIDListSize(pidlB);

	pidl = CreateIDList(cbA + cbB);

	if (pidl)
	{
		if (pidlA)
			memcpy(pidl, pidlA, cbA);

		memcpy((LPBYTE) pidl + cbA, pidlB, cbB);
	}

	return pidl;
}

int GetIDListSize(LPCITEMIDLIST pidl)
{
	if (!pidl)
		return 0;

	int cb = sizeof(pidl->mkid.cb);

	while (pidl->mkid.cb)
	{
		cb = cb + pidl->mkid.cb;
		pidl = NextID(pidl);
	}

	return cb;
}

LPCITEMIDLIST NextID(LPCITEMIDLIST pidl)
{
	return (LPCITEMIDLIST) ((CONST BYTE *) pidl + pidl->mkid.cb);
}

void FreeIDList(LPITEMIDLIST pidl)
{
	IMalloc *pMalloc = NULL;

	if (SUCCEEDED(SHGetMalloc(&pMalloc)))
	{
		pMalloc->Free(pidl);
		pMalloc->Release();
	}
}


